//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "Base.h"
#include <Config.h>
#include <constants.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <errno.h>

// from soccerserver
namespace Soccer {
#include <param.h>
#include <types.h>
}

#define FORCE_CORE

//#define DEBUG

#include "FunsselStrings.h"
#include "Socket.h"

using namespace std;

// cwz-build-command: gmake FUNSSEL

// time in microsecs to wait after every 20 messages in playback
const int MSG_PAUSE = 1000;


int Base::strncmp(const string &s, const char *str, int len)
{
  return ::strncmp(s.data(), str, len);
}

//////////////////////////////////////////////////////////////

Base::Base(const string &server_host, const string &_logfile, bool _ro)
  throw (Exception) :
  fsocket(Config::getValueAsInt("proxy_socket"),
	  server_host.c_str(),
	  Config::getValueAsInt("server_socket"),
	  MAX_SERVER_MSG_SIZE),
  fsocket_ol(Config::getValueAsInt("proxy_olcoach_socket"),
	     server_host.c_str(),
	     Config::getValueAsInt("server_olcoach_socket"),
	     MAX_SERVER_MSG_SIZE),
  fsocket_cc(Config::getValueAsInt("proxy_coach_socket"),
	     server_host.c_str(),
	     Config::getValueAsInt("server_coach_socket"),
	     MAX_SERVER_MSG_SIZE),
  mode(Playback),
  srv_clock(-1),
  srv_time(-2),
  max_run(32000),
  max_delay(Config::getValueAsInt("max_delay")),
  filter(false),
  sync_mode(true),
  messages(_logfile),
  finished(false),
  readonly(_ro),
  log_board(!Config::getValueAsInt("filter_board_log")),
  msg_board(!Config::getValueAsInt("filter_board_audio")),
  log_commands(!Config::getValueAsInt("filter_commands"))
{
  clients.erase(clients.begin(), clients.end());
  for (int i = 0; i < NumberSActors; i++)
    clientIDOf[i] = -1;
}

Base::~Base()
{
  if (mode != Playback)
    stopGame();
}

void Base::stopGame() throw (Exception)
{
  if (mode != Playback)
    server.kill();
  dropNonFunsselMon();
  messages.flush();
}

bool Base::allPlayersCheckedIn(bool ignoreCoaches)
{
  for (int i =  PlayerL_1; i <= CoachR; i++) {
    if (clientIDOf[i] == -1) {
      if (ignoreCoaches && (i == CoachL || i == CoachR))
        continue;
#ifdef DEBUG
      cerr << "waiting for player " << i << endl;
#endif
      return false;
    }
  }
  return true;
}

///////////////////////////////////////////////////// sockets

Base::SoccerClient *Base::getMessage(FilterMessage &msg) throw (Exception)
{
  SocketDesc desc = S_None;
  msg = fsocket.getNextMessage();
  if (msg.isValid())
    desc = S_Primary;
  else {
    msg = fsocket_ol.getNextMessage();
    if (msg.isValid())
      desc = S_OnlineCoach;
    else {
      msg = fsocket_cc.getNextMessage();
      if (msg.isValid())
	desc = S_OfflineCoach;
      else
	// no message
	return NULL;
    }
  }

  SoccerClient &clnt = findClient(msg.client);
  clnt.port = desc;
  return &clnt;
}

void Base::sendMessage(FilterMessage &msg) throw (Exception)
{
  vector<SoccerClient>::iterator it;
  for (it = clients.begin(); it != clients.end(); it++)
    if (it->clientID == msg.client)
      break;
  if (it == clients.end())
    throw Exception("base: send to unknown client");
  switch (it->port) {
  case S_OnlineCoach:
    fsocket_ol.send(msg);
    break;
  case S_OfflineCoach:
    fsocket_cc.send(msg);
    break;
  default:
    fsocket.send(msg);
  }
}

void Base::dropNonFunsselMon() throw (Exception)
{
  // first temporary throw out all clients that are not
  // FunsselMonitors. They would not survive a server
  // restart anyway
  vector<SoccerClient>::iterator it;
  for (it = clients.begin(); it != clients.end(); it++) {
    if (it->actor != FunsselMonitor) {
      switch (it->port) {
      case S_Primary:
	fsocket.drop(it->clientID);
	break;
      case S_OnlineCoach:
	fsocket_ol.drop(it->clientID);
	break;
      case S_OfflineCoach:
	fsocket_cc.drop(it->clientID);
	break;
      default:;
      }
      // and declare unknown
      it->actor = NoOne;
    }
    else {
      // reset FunsselMonitor cut through connection
      fsocket.resetSendPort(it->clientID);
    }
  }

  for (int i = 0; i < NumberSActors; i++) {
    // funssel monitor must not be thrown out
    if (i != FunsselMonitor)
      clientIDOf[i] = -1;
  }
}

//////////////////////////////// connect socket
int Base::connect(SoccerActor actor)
{
  int id = fsocket.connect();

  // forced identification
  SoccerClient tmp_clnt;
  tmp_clnt.clientID = id;
  tmp_clnt.actor    = actor;
  tmp_clnt.port     = S_Primary;
  clients.push_back(tmp_clnt);
  if (clientIDOf[actor] != -1)
    clientIDOf[actor] = id;
  return id;
}

void Base::sendFMsg(FilterMessage &msg)
{
#ifdef DEBUG
  if (msg.message.length() < 64)
    cerr << "msg for Soccerserver: " << msg.message << endl;
  else
    cerr << "msg for Soccerserver: long crap of size " << msg.message.length() << endl;
#endif
  msg.fromSvr = FilterMessage::ToServer;
  fsocket.send(msg);
}

///////////////////////////////////////// act

bool Base::act()
{
  FilterMessage msg("", -1, FilterMessage::FromServer);
  SoccerClient *clntP = getMessage(msg);
  if (clntP == NULL) {
    return false;
  }
  SoccerClient &clnt(*clntP);
  if (clnt.actor == NoOne) {
    // try to determine client type
    switch (clnt.port) {
    case S_OfflineCoach:
      // offline coach is easy to determine: by socket
      identify(clnt, OfflineCoach);
      break;
    case S_OnlineCoach:
      if (msg.fromServer()) {
	SoccerActor actor = coachSide(msg.message);
	if (actor != NoOne)
	  identify(clnt, actor);
	else
	  cerr << "FUNSSEL warning: cannot determine coach side (a message was lost)"
	       << endl;
      }
      break;
    case S_Primary:
      if (msg.fromServer()) {
	//player numbers are declared by server
	SoccerActor actor = playerSide(msg.message);
	if (actor != NoOne)
	  identify(clnt, actor);	  
      }
      else {
	// monitors declare themselves
	if (!strncmp(msg.message, "(disp", 5))
	  identify(clnt, SoccerMonitor);
	else if (!strncmp(msg.message, FILTER_STRING " (dispinit))",
			  FILTER_STRING_LEN + 12))
	  identify(clnt, FunsselMonitor);
      }
      break;
    default:
      throw Exception("socket id illegal");
    }
  }

  // convert into SoccerMessage
  SoccerMessage s_msg(0, "", NoOne, NoOne);

  if (msg.fromServer()) {
    // messages from the server are no problem
    if (mode == Playback) {
      cerr << "server messages in logplay mode?" << endl;
      return true;
    }
    adjustTime(msg.message);

    // see wether the message shall be stored
    bool store = log_commands;
    
    Soccer::dispinfo_t2 *d = (Soccer::dispinfo_t2 *)(msg.message.data());
    switch (ntohs(d->mode)) {
    case MSG_MODE:
      switch (ntohs(d->body.msg.board)) {
      case MSG_BOARD:
	store = msg_board;
	break;
      case LOG_BOARD:
	store = log_board;
	break;
      default: ;
      }
      break;
    case SHOW_MODE:
    case PT_MODE:
    case PPARAM_MODE:
      store = true;
      break;
    case DRAW_MODE:
    case BLANK_MODE:
    case PM_MODE:
    case TEAM_MODE:
    case PARAM_MODE:
    default:
      store = false;
      break;
    }
#ifdef DEBUG
    if (store)
      cerr << "storing req " << ntohs(d->mode) << " data" << endl;
    else
      cerr << "ignoring req " << ntohs(d->mode) << " data" << endl;
#endif

    if (store) {
      s_msg.sender = SoccerServer;
      s_msg.recver = clnt.actor;
      s_msg.time   = srv_clock;
      s_msg.durtn  = 1;
      s_msg.message = msg.message;
    }

    // forward original message, even if it is not stored
    if (clnt.actor != FunsselMonitor)
      sendMessage(msg);
  }
  else {
    // now it gets tricky, thanks to the proxy protocol
    if (!strncmp(msg.message, FILTER_STRING, FILTER_STRING_LEN)) {
      // message has proxy form

      // filter message are only forwarded from known clients
      if (clnt.actor == NoOne)
	return true;

      const char *c_msg = msg.message.c_str() + FILTER_STRING_LEN + 1;
      int length = msg.message.length() - FILTER_STRING_LEN - 1;
      if (length <= 1) {
	sendFMsg(clnt.actor, "error illegal proxy command");
	return true;
      }
      if (*c_msg == '(') {
	// a funssel special command
	if (!strncmp(c_msg,"(dispmsg", 8)) {
	  c_msg  += 8;
	  length -= 8;
	  // a protected message - unpack it
	  int len, n, i_recver, i_sender;
	  if (sscanf(c_msg + 8, "%d %d %d %d %d %n", &s_msg.time,
		     &s_msg.durtn, &i_sender, &i_recver,
		     &len, &n) != 6) {
	    sendFMsg(clnt.actor, "error illegal dispmsg");
	    return true;
	  }
	  c_msg  += n;
	  length -= n;
	  if (length < len) {
	    sendFMsg(clnt.actor, "error message too short");
	    return true;
	  }
	  if (s_msg.time == -2)
	    s_msg.time = srv_clock;
	  s_msg.sender = (SoccerActor) i_sender;
	  s_msg.recver = (SoccerActor) i_recver;
	  s_msg.message.assign(c_msg, len);
	  // continue with the unpacked message
	  // forward if not FunsselMonitor (filtering/and send on demand)
	  // and is not for an ancient time step
	  if ((s_msg.recver != FunsselMonitor) && (s_msg.time == srv_time)) {
	    FilterMessage f_msg("", -1, FilterMessage::FromServer);
	    if (s_msg.recver == SoccerServer) {
	      f_msg.client  =  clientIDOf[s_msg.sender];
	      f_msg.fromSvr = true;
	    }
	    else {
	      f_msg.client  = clientIDOf[s_msg.recver];
	      f_msg.fromSvr = false;
	    }
	    f_msg.message = s_msg.message;
	    sendMessage(msg);
	  }
	}
	else {
	  parseSpecialCommand(clnt.actor, c_msg);
	  return true;
	}
      }
      else {
	// a normal proxy command - redirect to FunsselMonitor
	int len;
	s_msg.sender = clnt.actor;
	s_msg.recver = FunsselMonitor;
	s_msg.time   = srv_clock;
	if (sscanf(c_msg, "%d %n", &s_msg.durtn, &len) != 1) {
	  sendFMsg(clnt.actor, "error illegal command");
	  return true;
	}
	s_msg.message = msg.message;
	// these messages are not directly forwarded
	// because the funssel uses winding
      }
    }
    else {
      // a normal message for the server
      if (log_commands) {
	s_msg.sender = clnt.actor;
	s_msg.recver = SoccerServer;
	s_msg.time   = srv_clock;
	s_msg.durtn  = 1;
	s_msg.message = msg.message;
      }
#ifdef DEBUG
      cerr << "message for server " << s_msg.message << endl;
#endif

      // and forward (original message)
      sendMessage(msg);
    }
  }

  // ignore messages for unknown clients (or invalided msgs)
  if ((mode != Playback) && (s_msg.sender != NoOne) &&
      (s_msg.recver != NoOne)) {
    // synchronous pipe through to monitor in game
    // send only if not proxy or not filtered
    if (sync_mode && (strncmp(msg.message, FILTER_STRING, FILTER_STRING_LEN) || !filter)) {
      if (messages.getCurrentFetchTime() != s_msg.time) {
	char buf[256];
	// close timestep sent so far
	sendFMsg(FunsselMonitor, "dispeol");
	messages.fetchTime(s_msg.time);
	sprintf(buf, "datastream %d", messages.getCurrentFetchTime());
	sendFMsg(FunsselMonitor, buf);
      }
      sendFMon(s_msg);
    }
    if (!messages.insertMsg(s_msg))
      sendFMsg(FunsselMonitor, "(insert failed)");
  }
#ifdef DEBUG
  else
    cerr << "ignoring message " << s_msg.sender << " " << s_msg.recver << endl;
#endif
  return true;
}

void Base::parseSpecialCommand(SoccerActor actor, const char *com)
{
#ifdef DEBUG
  cerr << "special command " << com << endl;
#endif

  if (actor != FunsselMonitor) {
    char buf[32];
    // currently this proxy does not know anything about
    // proxy commands from players. Forward them to monitor
    sprintf(buf, "forward %d ", int(actor));
    sendFMsg(FunsselMonitor, string(buf) + com);
  }
  else if (!strncmp(com, "(dispinit)", 10))
    sendFMsg(FunsselMonitor, "welcome");
  else if (!strncmp(com, "(quit)", 6))
    finished = true;
  else if (!strncmp(com, "(server start)", 14)) {
    if (readonly)
      sendFMsg(FunsselMonitor, "error logfile is readonly");
    else {
      startServer();
      sendFMsg(FunsselMonitor, "server started");
    }
  }
  else if (!strncmp(com, "(server stop", 9))
    stopGame();
  else if (!strncmp(com, "(max_run", 8)) {
    if (strlen(com) <= 9)
      sendFMsg(FunsselMonitor, "error illegal max_run");
    max_run = atol(com + 9);
    sendFMsg(FunsselMonitor, "max_run ok");
  }
  else if (!strncmp(com, "(filter on", 10))
    filter = true;
  else if (!strncmp(com, "(filter off", 11))
    filter = false;
  else if (!strncmp(com, "(sync_mode", 10))
    sync_mode = true;
  else if (!strncmp(com, "(fetch", 6)) {
    if (strlen(com) <= 8)
      sendFMsg(FunsselMonitor, "error illegal fetch");
    else {
      int sign = 0;
      if (com[7] == 'b')
	sign = -1;
      if (com[7] == 'f')
	sign =  1;
      int time = atol(com + ((sign != 0) ? 8 : 7));
      if (sign > 0)
	fetch(messages.getCurrentFetchTime() + time);
      else if (sign < 0)
	fetch(messages.getCurrentFetchTime() - time);
      else
	fetch(time);
    }
  }
  else
    sendFMsg(actor, "error illegal command");

  if ((mode == Stopped) &&
      (srv_clock - messages.getCurrentFetchTime() <= max_delay) &&
      (srv_clock < max_run))
    contServer();

  return;
}

void Base::sendFMsg(SoccerActor actor, const string &msg) throw (Exception)
{
  if (clientIDOf[actor] < 0)
    return;

  char header[64];
  sprintf(header, FILTER_STRING " %d (", srv_time);
  string tmp_msg = header + msg + "))";
  FilterMessage f_msg(tmp_msg, clientIDOf[actor], FilterMessage::FromServer);
#ifdef DEBUG
  if (actor == FunsselMonitor) {
    if (tmp_msg.length() < 64)
      cerr << "sent to FMon: " << tmp_msg << endl;
    else
      cerr << "sent to FMon long crap" << endl;
  }
#endif
  sendMessage(f_msg);
}

void Base::sendFMon(const SoccerMessage &s_msg) throw (Exception)
{
  char buf[256];
  sprintf(buf, "dispmsg %d %d %d %d %d ",
	  s_msg.time,   s_msg.durtn,
	  s_msg.sender, s_msg.recver,
	  s_msg.getLength());
  string msg(buf);
  msg += s_msg.message;
  sendFMsg(FunsselMonitor, msg);
}

void Base::startServer() throw (Exception)
{
  dropNonFunsselMon();

  mode = Running;

  const char *argv[] = {
    Config::getValue("server_command"),
    Config::getValue("server_socket"),
    Config::getValue("server_olcoach_socket"),
    Config::getValue("server_coach_socket"),
    NULL
  };

  srv_time = -1;
  srv_clock = 0;

  server.restart(argv, Config::getValue("server_dir"));
  messages.openWrite();
#ifdef DEBUG
  cerr << "server started" << endl;
#endif

  // give server time to start
  int termtime = time(NULL) + Config::getValueAsInt("serverstarttime");
  while (time(NULL) < termtime)
    sleep(1);
#ifdef DEBUG
  cerr << "block finished" << endl;
#endif
}

void Base::stopServer() throw (Exception)
{
  server.stop();
  mode = Stopped;
}

void Base::contServer() throw (Exception)
{
  server.cont();
  mode = Running;
}

Base::SoccerClient &Base::findClient(int id)
{
  vector<SoccerClient>::iterator it;
  for (it = clients.begin(); it != clients.end(); it++)
    if (it->clientID == id)
      break;

  if (it == clients.end()) {
    // client's first connect
#ifdef DEBUG
    cerr << "got new client " << id << endl;
#endif
    SoccerClient tmp_clnt;
    tmp_clnt.clientID = id;
    tmp_clnt.actor    = NoOne;
    tmp_clnt.port     = S_None;
    clients.push_back(tmp_clnt);
    it = clients.end() - 1;
  } 
  return *it;
}

void Base::identify(SoccerClient &clnt, SoccerActor actor)
{
  clnt.actor = actor;
#ifdef DEBUG
  cout << "identify " << clnt.clientID << " as " << actor << endl;
#endif
  // first one is the only one FUNSSEL accepts
  if (clientIDOf[actor] == -1)
    clientIDOf[actor] = clnt.clientID;
}

void Base::fetch(int t) throw (Exception)
{
  char buf[256];
  sync_mode = false;
  SoccerMessage msg;

  messages.fetchTime(t);

  sprintf(buf, "databurst %d", messages.getCurrentFetchTime());
  sendFMsg(FunsselMonitor, buf);

  int cnt = 0;
  while ((msg = messages.getNextMessage()).isValid()) {
    if ((msg.recver == FunsselMonitor) &&
	(strncmp(msg.message, FILTER_STRING, FILTER_STRING_LEN) || !filter)) {
      sendFMon(msg);
      if (++cnt >= 10) {
	usleep(MSG_PAUSE);
	cnt = 0;
      }
    }
  }

  sendFMsg(FunsselMonitor, "dispeol");
}

void Base::adjustTime(const string &msg)
{
  // test binary message
  const Soccer::dispinfo_t2 *d =
    (const Soccer::dispinfo_t2 *)msg.data();
  int tmp_time;

  if (d->mode == htons(SHOW_MODE)) {
    tmp_time = int((short)ntohs(d->body.show.time));
    srv_clock++;
  }
  else {
    // check if text message
    int timeInd = -1;
    if (!strncmp(msg, "(see ", 5))
      timeInd = 5;
    else if (!strncmp(msg, "(sense_body ", 12))
      timeInd = 12;
    else if (!strncmp(msg, "(hear ", 6))
      timeInd = 6;
    else if (!strncmp(msg, "(ok look ", 9))
      timeInd = 9;
    else
      // no time in here (e.g. error)
      return;
    tmp_time = atol(msg.substr(timeInd).c_str());
  }

  if (tmp_time != srv_time) {
#ifdef DEBUG
    cout << "detected new time step " << tmp_time << endl;
#endif
    srv_time = tmp_time;
    sendFMsg(FunsselMonitor, "server time changed");
  }

  if ((srv_clock - messages.getCurrentFetchTime() > max_delay) ||
      (srv_clock >= max_run))
    stopServer();
}

SoccerActor Base::coachSide(const string &msg)
{
  if (strncmp(msg, "(init ", 6))
    return NoOne;

#ifdef DEBUG
  cerr << "coach got init msg" << endl;
#endif
  
  bool left = (msg[6] == 'l');
  
#ifdef DEBUG
  if (!left && (msg[6] != 'r'))
    cout << "left/right key failed:" << msg << " <" << msg[6] << ">" << endl;
#endif

#ifdef DEBUG
  cerr << "got olcoach " << left << endl;
#endif
  
  return (left) ? CoachL : CoachR;
}

SoccerActor Base::playerSide(const string &msg)
{
  if (strncmp(msg, "(init ", 6))
    return NoOne;

#ifdef DEBUG
  cerr << "player got init msg" << endl;
#endif
  
  bool left = (msg[6] == 'l');
  
#ifdef DEBUG
  if (!left && (msg[6] != 'r'))
    cout << "left/right key failed:" << msg << " <" << msg[6] << ">" << endl;
#endif

  int actor = atol(msg.substr(8).c_str());
  if (!left)
    actor += PlayerR_1 - 1;

#ifdef DEBUG
  cerr << "got player " << left << " " << actor << endl;
#endif
  return (SoccerActor)actor;
}

void usage(const char *name)
{
  cerr << "usage: " << name << " [-r] [-i] [-b] [-l logfile] [-s]" << endl;
}

int main(int argc, char **argv)
{
  Config::readFile("funsselrc");

  string logfile  = "soccerlog";
  bool batchmode = false;
  bool ignoreCoaches = false;
  bool print = false;
  bool readonly = false;

  char c;
  while ((c = getopt(argc, argv, "ribsl:")) != EOF) {
    switch (c) {
    case 'r':
      readonly = true;
      break;
    case 'i':
      ignoreCoaches = true;
      break;
    case 'b':
      batchmode = true;
      break;
    case 'l':
      logfile = optarg;
      break;
    case 's':
      print = true;
      break;
    default:
      usage(argv[0]);
    }
  }

  if (batchmode && readonly) {
	cerr << "batchmode makes no sense without writing a logfile" << endl;
	exit(-1);
  }

  if (print) {
    try {
      MsgStore store(logfile);
      store.show();
    }
    catch (Exception e) {
      cerr << "an exception occured: " << e << endl;
    }
    catch (...) {
      cerr << "an unexpected exception occured" << endl;
    }
    return 0;
  }

  FilterSocket::enableInts();

#ifdef FORCE_CORE
  struct rlimit rlim;
  getrlimit(RLIMIT_CORE, &rlim);
  rlim.rlim_cur = rlim.rlim_max;
  cerr << "core dump enabled." << endl;  
  if (setrlimit(RLIMIT_CORE, &rlim) == -1)
    cerr << errno << endl;
#endif
  
  try {
    int client = -1;
    Base base("localhost", logfile, readonly);

    if (batchmode) {
      // no delay based server stopping
      base.setMaxDelay(1000000);

      base.startServer();
      // connect and identify as monitor
      client = base.connect(FunsselMonitor);
      FilterMessage fmsg("(dispinit version 2)", client, FilterMessage::ToServer);
      base.sendFMsg(fmsg);
    }

    int maxKickoff = Config::getValueAsInt("maxkickoffcount");
    int kickoffcount = 0;
    int ctime = time(NULL);
    int laststime = -1;
    while (!base.termReq()) {
      if (!base.act())
	usleep(100000);
      if (batchmode) {
	// time cycle logging
	if ((base.getSrvTime() % 100) == 0 &&
	    laststime != base.getSrvTime()) {
	  cerr << "time " << base.getSrvTime() << "\r";
	  laststime = base.getSrvTime();
	}
	// every second check if kickoff necessary
	if (ctime != time(NULL)) {
	  ctime = time(NULL);
	  // see if we have to start the game
	  if ((base.getSrvTime() % 3000) == 0) {
	    // always terminate after 6000 cycles
	    // no extension time
	    if (base.getSrvTime() >= 6000)
	      break;
	    if (base.allPlayersCheckedIn(ignoreCoaches)) {
	      FilterMessage fmsg("(dispstart)", client, FilterMessage::ToServer);
	      base.sendFMsg(fmsg);
	      cerr << "sent kickoff" << endl;
	      kickoffcount = 0;
	    }
	    else {
	      if (kickoffcount++ > maxKickoff) {
		cerr << "timeout for kickoff" << endl;
		exit(-1);
	      }
	    }
	  }
	}
      }
    }
  }
  catch (Exception e) {
    cerr << "an exception occured: " << e << endl;
    exit(-1);
  }
  catch (...) {
    cerr << "an unexpected exception occured" << endl;
    exit(-1);
  }
  return 0;
}
