//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "Ensure.h"
// cwz-build-command: gmake all

double AdvSkills::goToThresh(double error, double tdist)
{
  // check if turn is needed (distance depended)
  // visual angle of target/error circle
  double gamma = atan2(error, tdist);
  double split = 1/(1 + goToThreshStretch*(tdist - error));
  return (1 - split)*M_PI/2 + split*gamma;
}

#ifdef MRBDEBUG
//#define RUNLOG(out) Comm::sendMCmd(MLog(out))
#endif

#ifndef RUNLOG
#define RUNLOG(out) 
#endif

AdvSkills::TurtleCommand
AdvSkills::runInDir( double rightdir,
		     double leftdir,
		     double prefAngle,
		     double stamina)
{
  TurtleCommand com;
  com.state = TONWAY;
  // time information not valid
  com.time = -1;
  Angle tangle;

  if (rightdir < 0)
    rightdir += 2 * M_PI;
 
  if (leftdir < 0)
    leftdir += 2 * M_PI;

  double bodyDir = world->my.bodyDir;
  if (bodyDir < 0)
    bodyDir += 2 *M_PI;
 

  if ( prefAngle < - 3* M_PI )
    {
      RUNLOG("Calculating tangle");
      // no angle set, calculate a turnangle
      tangle = ( rightdir + leftdir ) /2;
      if ( rightdir < leftdir )
	{
	  double dummy=rightdir;
	  rightdir = leftdir;
	  leftdir = dummy;
	  tangle += PI;
	}
    }
  else
    tangle = prefAngle;

  // rightdir -= tangle;
  // leftdir -= tangle;

#ifdef MRBDEBUG
  {
    ostringstream t;
    t << " bodyDir: " << world->my.bodyDir
      << " tangle: " << tangle
      << " rightDir: " << rightdir
      << " leftDir: " << leftdir
      << " turnMinAngle: " << turnMinAngle
      << " 1: " << (double) world->my.bodyDir - rightdir - turnMinAngle
      << " 2: " << leftdir - (double) world->my.bodyDir - turnMinAngle;
    RUNLOG(t.str());
  }
#endif

  // check if turn is needed (distance depended)
  if ( ( (double) bodyDir - rightdir - turnMinAngle ) > 0 ||
       ( leftdir - (double) bodyDir - turnMinAngle ) > 0 )
    {
      RUNLOG("Turn");
      com.v       = world->my.v;
      TurnCommand tmptcm = turn(tangle);
      com.bodyDir = tmptcm.bodyDir;
      com.turnMoment = tmptcm.turnMoment;
      com.command = TurtleCommand::TURN;
      return com;
    }

  RUNLOG("DASH");
  double minacc, maxacc;
  calcAcc(stamina, minacc, maxacc);

  double tgtDash = maxacc;

  com.v       = world->my.v + radAng2Vector2d(tgtDash, world->my.bodyDir);
  com.bodyDir = world->my.bodyDir;
  com.command = TurtleCommand::DASH;
  return com;
}

AdvSkills::TurtleCommand
AdvSkills::goToPosition(const Vector2d &target,
			double error, double stamina,
			bool minAngle, Angle prefAngle)
{
  TurtleCommand com;
  com.state = TONWAY;
  // time information not valid
  com.time = -1;
  // simple turn/dash algorithm. Will not work for small errors, I guess.
  // But it is not for that at all;

  com.v = target - world->my.pos; 
  Angle tangle = com.v.getAngle();
  double tdist = norm(com.v);

  if (tdist < error) {
    // we still calculate a command. But TFINISHED will be its flag
    com.time  = 0;
    com.state = TFINISHED;
  }

  // angle minimization if wanted
  bool reverse = false;
  if (minAngle && (fabs(tangle - prefAngle) > M_PI/2)) {
    reverse = true;
    tangle += PI;
    if (com.state != TFINISHED)
      com.state = TGOINGBACKWARDS;
  }

  // check if turn is needed (distance depended)
  Angle diff = tangle - world->my.bodyDir;
  if (fabs(diff) > goToThresh(error, tdist)) {
    com.v       = world->my.v;
    TurnCommand tmptcm = turn(tangle);
    com.bodyDir = tmptcm.bodyDir;
    com.turnMoment = tmptcm.turnMoment;
    com.command = TurtleCommand::TURN;
    return com;
  }

  double minacc, maxacc;
  calcAcc(stamina, minacc, maxacc);

  double l1, l2, tgtDash = 0;
  // overkill since this is very likely more exact than error implies
  switch (intersectLineCircle(radAng2Vector2d(1, world->my.bodyDir),
			      com.v - world->my.v, error, l1, l2)) {
  case None: // no panic. We didn't turn since we are far away.
    // dash at full speed (haha)
    if (reverse)
      tgtDash = minacc;
    else
      tgtDash = maxacc;
    
    break;
  case Valid:
    if (l1 > maxacc)
      // dash forward
      tgtDash = maxacc;
    else if (l2 < minacc)
      // dash backwards
      tgtDash = minacc;
    else
      tgtDash = (max(l1, minacc) + min(l2, maxacc))/2;
    break;
  case Instable: // impossible, norm(rad...) = 1
  default: ; // R1larger/R2larger only for Circle/Circle
  }

  com.v       = world->my.v + radAng2Vector2d(tgtDash, world->my.bodyDir);
  com.bodyDir = world->my.bodyDir;
  com.command = TurtleCommand::DASH;
  return com;
}

double AdvSkills::reachRadius(double dlen, int numDashes, double decay)
{
  return dlen/(1 - decay)*
    (numDashes - decay*(1 - power(decay, numDashes))/(1 - decay));
}

AdvSkills::TurtleCommand
AdvSkills::goToPositionTimed(const Vector2d &target,
			     double stamina, double error)
{
  WMovedObject tgt(0, 0, 0, target, Vector2d(0, 0));
  WMovedObject tmp;
  return interceptTimed(world->my, tgt, stamina, error, tmp); 
}

AdvSkills::TurtleCommand
AdvSkills::controlBall(double stamina)
{
  WBall tmp;
  return interceptTimed(world->my, world->ball, stamina, world->my.kickRange(), tmp); 
}

AdvSkills::TurtleCommand
AdvSkills::controlBall(double stamina,
		       double error, WBall &meet)
{
  return interceptTimed(world->my, world->ball, stamina, error, meet);
}

AdvSkills::TurtleCommand
AdvSkills::interceptTimed(const WPlayer &my,
			  const WMovedObject &target,
			  double stamina, double error,
			  WMovedObject &meet)
{
  double err;

#ifdef MRBDEBUG
  if (GETCONF()->draw) {
    ostringstream tmp;
    tmp << "interceptTimed: stamina input = " << stamina;
    const char *str = tmp.str().c_str();
    LEVELTRC(SKILLS_LEVEL, str);
  }
#endif
  TurtleCommand com;
  com.state = TONWAY;

  WPlayer simMy(my);
  double bdist = simMy.distTo(target);
  // the positional error we must allow
  double position_err = error + 0.5 * bdist * ( exp( QUANTIZE_STEP ) - 1 );
  const Vector2d mybvec = radAng2Vector2d(1, my.bodyDir);

  meet = target;

  if (bdist < error) {
    // we still calculate a command. But TFINISHED will be its flag
    com.time  = 0;
    if (com.state != TFINISHED)
      com.state = TFINISHED;
  }

  // we need at least one dash
  meet.simulate();
  simMy.simulate();

  ///////////////////////////////// single dash ?
  // also dashes backwards
  Vector2d relpos = meet.pos - simMy.pos;
  double l1, l2;

  // last dash ALWAYS at full power
  double maxacc = simMy.dashPowerRate * MAXPOWER, minacc = simMy.dashPowerRate * MINPOWER;
  double collRange = BALL_SIZE + world->my.size + M_SQRT2*(norm(world->my.v)*PLAYER_RAND + norm(world->ball.v)*BALL_RAND);

  // check for dash on innerSafeCircle, or if not possible, into error
  // so we are as close as still safe to the target
  if (intersectLineCircle(radAng2Vector2d(1, simMy.bodyDir),
			  relpos, collRange, l1, l2) == Valid) {
    double tgtDash = 0;
    bool use = false;
    bool l1ok = (l1 >= minacc) && (l1 <= maxacc);
    bool l2ok = (l2 >= minacc) && (l2 <= maxacc);

    if (l1ok && !l2ok) {
      tgtDash = l1;
      use = true;
    }
    else if (l2ok && !l1ok) {
      tgtDash = l2;
      use = true;
    }
    else if (l1ok && l2ok) {
      // use lower dash (dashing backwards costs double stamina)
      double l1stamina = (l1 > 0) ? l1 : 2*l1;
      double l2stamina = (l2 > 0) ? l2 : 2*l2;
      tgtDash = (fabs(l1stamina) < fabs(l2stamina)) ? l1 : l2;
      use = true;
    }
    if (use) {
      Vector2d dashVec = radAng2Vector2d(tgtDash, simMy.bodyDir);
      if (tgtDash < 0 && com.state != TFINISHED) {
	com.state = TGOINGBACKWARDS;
	LEVELTRC(SKILLS_LEVEL, "intercept: going backwards");
      }
      if (com.state != TFINISHED)
	com.time    = 1;
      com.v       = my.v + dashVec;
      com.bodyDir = my.bodyDir;
      com.command = TurtleCommand::DASH;

      NAMEDLINE("my v", my.pos, my.v);
      NAMEDLINE("dash only safe circle", simMy.pos, dashVec);
      LEVELDRW(SKILLS_LEVEL, MDCircle(meet.pos, collRange));
      LEVELTRC(SKILLS_LEVEL, "intercept: single close dash");
      return com;
    }
  }

  // safeCircle didn't do it, try given error
  // because of the above intersection, we know that
  // our result can't end up with a collision
  if (intersectLineCircle(radAng2Vector2d(1, simMy.bodyDir),
			  relpos, error,
			  l1, l2) == Valid) {
    if ((l1 <= maxacc) && (l2 >= minacc)) {
      double wacc = (l1 + l2)/2;
      if (wacc > maxacc)
	wacc = maxacc;
      if (wacc < minacc)
	wacc = minacc;
      Vector2d dashVec = radAng2Vector2d(wacc, simMy.bodyDir);      
      if (wacc < 0 && com.state != TFINISHED) {
	com.state = TGOINGBACKWARDS;
	LEVELTRC(SKILLS_LEVEL, "intercept: going backwards");
      }
      if (com.state != TFINISHED)
	com.time    = 1;
      com.v       = my.v + dashVec;
      com.bodyDir = my.bodyDir;
      com.command = TurtleCommand::DASH;

      NAMEDLINE("my v", my.pos, my.v);
      NAMEDLINE("dash only kick range", simMy.pos, dashVec);
      LEVELDRW(SKILLS_LEVEL, MDCircle(meet.pos, error));
      LEVELTRC(SKILLS_LEVEL, "intercept: single error dash");
      return com;
    }
  }

  // now we need at least two steps
  meet.simulate();
  simMy.simulate();

  ///////////////////////////////// turn-dash ?
  relpos = meet.pos - simMy.pos;
  Vector2d vl1, vl2, dashvec(0, 0);
  bool tdgood = false;
  switch(intersectCircleCircle(maxacc, relpos, error, vl1, vl2)) {
  case Valid:
    tdgood = true;
    // the direction is ok, although we can get nearer of course
    // but with that we deal after the turn
    dashvec = 0.5*(vl1 + vl2);
    break;
  case R1larger:
    tdgood = true;
    dashvec = relpos;
    break;
  case R2larger: // hmm, we already got the ball (???)
    tdgood = true; // do essentially nothing (since we turn first)
    dashvec = radAng2Vector2d(0.0001, relpos.getAngle());
  case None:
  default:;
  }

  if (tdgood) {
    NAMEDLINE("turn dash", simMy.pos, dashvec);
    if (com.state != TFINISHED)
      com.time    = 2;
    com.v       = my.v;
    com.command = TurtleCommand::TURN;

    TurnCommand tmptcm = turn(dashvec.getAngle()); 
    com.bodyDir = tmptcm.bodyDir;
    com.turnMoment = tmptcm.turnMoment;
    if (tmptcm.state != RANGLE) {
      LEVELTRC(SKILLS_LEVEL, "intercept: one turn perfect");
      return com;
    }
    // check if proposed turn still is enough
    double l1, l2;
    if (intersectLineCircle(radAng2Vector2d(1, tmptcm.bodyDir),
			    relpos, error, l1, l2) == Valid) {
      // there IS a solution with that angle
      if ((l1 <= maxacc) && (l2 >= minacc)) {
	LEVELDRW(SKILLS_LEVEL, MDLine(simMy.pos + l1*dashvec, simMy.pos + l2*dashvec));
	LEVELTRC(SKILLS_LEVEL, "intercept: one turn ok");
	return com;
      }
    }

    // it's not as easy. Maybe turning into the opposite dir is ok (for dashing backwards)
    // same process again
    tmptcm = turn(PI + dashvec.getAngle()); 
    com.bodyDir = tmptcm.bodyDir;
    com.turnMoment = tmptcm.turnMoment;
    if (tmptcm.state != RANGLE) {
      // HA!!! THAT works (sorry for the stamina)
      LEVELTRC(SKILLS_LEVEL, "intercept: one turn perfect for backwards dash");
      return com;
    }
    // check if proposed turn still is enough
    if (intersectLineCircle(radAng2Vector2d(1, tmptcm.bodyDir),
			    relpos, error, l1, l2) == Valid) {
      // there IS a solution with that angle
      if ((l1 <= maxacc) && (l2 >= minacc)) {
	LEVELDRW(SKILLS_LEVEL, MDLine(simMy.pos + l1*dashvec, simMy.pos + l2*dashvec));
	LEVELTRC(SKILLS_LEVEL, "intercept: one turn ok for backwards dash");
	return com;
      }
    }
  }

  ///////////////////////////////// dash-dash ?
  double dpower = simMy.dashPowerRate * min(MAXPOWER, (simMy.staminaIncMax + stamina/2));
  double len = reachRadius(dpower, 2, simMy.decay);

  if ((intersectLineCircle(mybvec, meet.pos - simMy.pos, position_err, l1, l2) == Valid) &&
      (l1 < len) && (l1 > 0)) {
    // I can get there without turn
    // l1 is the needed dash for a one cycle hit (impossible)
    // cut down if more dashes are needed
    if (l1 > dpower)
      l1 = dpower;
    if (com.state != TFINISHED)
      com.time    = 2;
    com.v       = my.v + radAng2Vector2d(l1, simMy.bodyDir);
    com.bodyDir = my.bodyDir;
    com.command = TurtleCommand::DASH;

#ifdef MRBDEBUG
    if (GETCONF()->draw) {
      char buf[256];
      sprintf(buf, "intercept: dashing straight, length=%f, time=2", len);
      LEVELTRC(SKILLS_LEVEL, buf);
      LEVELDRW(SKILLS_LEVEL, MDCircle(meet.pos, position_err));
    }
#endif
    return com;
  }

  ///////////////////////////////// simulation, t >= 3
  com.state = TONWAY;

  // after this loop, tgt is where we head
  // initialize to infinity position
  Vector2d tgt = target.pos + (1/(1 - target.decay))*target.v;
  dpower       = simMy.dashPowerRate * simMy.staminaIncMax;

  long time = 2; // this we know from before
  // hopefully not an endless loop
  // preserve stamina for last turn with full power
  stamina = max(0.0, stamina - 2*MAXPOWER);

  err = position_err;
  for (;;) {
    time++;
    meet.simulate();
    simMy.simulate();

    bdist = simMy.distTo( meet );
    // worst case error if the random is added maximal
    // both from the player and the ball
    err += norm(simMy.v)*PLAYER_RAND + norm(meet.v)*BALL_RAND;
    
    // first check if dashing alone would do it
    dpower = simMy.dashPowerRate * min(MAXPOWER, (simMy.staminaIncMax + stamina/time));
    len = reachRadius(dpower, time, simMy.decay);

    double l1, l2;
    if ((intersectLineCircle(mybvec, meet.pos - simMy.pos, err, l1, l2) == Valid) &&
	(l1 < len) && (l1 > 0)) {
      // I can get there without turn
      // l1 is the needed dash for a one cycle hit (impossible)
      // cut down if more dashes are needed
      if (l1 > dpower)
	l1 = dpower;
      com.time    = time;
      com.v       = my.v + radAng2Vector2d(l1, my.bodyDir);
      com.bodyDir = my.bodyDir;
      com.command = TurtleCommand::DASH;

#ifdef MRBDEBUG
      if (GETCONF()->draw) {
	char buf[256];
	sprintf(buf, "intercept: dashing straight, length=%f, time=%d", l1, (int)time);
	LEVELTRC(SKILLS_LEVEL, buf);
	LEVELDRW(SKILLS_LEVEL, MDCircle(meet.pos, err));
      }
#endif
      return com;
    }

    // now check one turn, if possible
    dpower = simMy.dashPowerRate * min(MAXPOWER, (simMy.staminaIncMax + stamina/(time - 1)));
    len = reachRadius(dpower, time - 1, simMy.decay);

    if (norm(meet.pos - simMy.pos) < len) {
      tgt = meet.pos;

      // check for angle not too large
      Angle dashdir = (tgt - simMy.pos).getAngle();
      TurnCommand tcm = turn(dashdir);
      // works, go to tgt
      if (tcm.state != RANGLE)
	break;
    }

    // check two turns
    dpower = simMy.dashPowerRate * min(MAXPOWER,
				       (simMy.staminaIncMax + stamina/(time - 2)));
    len = reachRadius(dpower, time - 2, simMy.decay);

    if (norm(meet.pos - simMy.pos) < len) {
      tgt = meet.pos;

      // ok, with two turns we get the ball, go!
      break;
    }
  }

  // if we fall here, tgt is where we should TURN
  // dashing straight is handled above
  // determine whether we should turn. This is a crude approximation

  Angle dashdir = (tgt - simMy.pos).getAngle();
  TurnCommand tcm = turn(dashdir);
  com.turnMoment = tcm.turnMoment;
  com.bodyDir = tcm.bodyDir;
  com.v       = my.v;
  if (com.state != TFINISHED)
    com.time    = time;
  com.command = TurtleCommand::TURN;

#ifdef MRBDEBUG
  if (GETCONF()->draw) {
    char buf[256];
    sprintf(buf, "intercept: turning, time=%d", (int)time);
    LEVELTRC(SKILLS_LEVEL, buf);
    LEVELDRW(SKILLS_LEVEL, MDLine(simMy.pos, tgt));
  }
#endif
  return com;
}

void AdvSkills::calcAcc(double stamina,
			double &minacc, double &maxacc)
{
  // ok, dash
  if (stamina < 0)
    stamina = 0;

  maxacc = min(stamina, MAXPOWER) * world->my.dashPowerRate;
  minacc = max(-0.5*stamina, MINPOWER) * world->my.dashPowerRate;

  double l1, l2;
  // max accelerations not exceeding player's maximum velocity
  switch (intersectLineCircle(radAng2Vector2d(1, world->my.bodyDir),
			      -world->my.v, world->my.vmax, l1, l2)) {
  case Valid:
    if ((l1 > minacc) && (l1 <= maxacc))
      minacc = l1;
    if ((l2 < maxacc) && (l2 >= minacc))
      maxacc = l2;
    break;
  case None: // Currently too fast (???)
  case Instable: // 1 == 0 (????)
  default:;
  }
}

void AdvSkills::exec(const AdvSkills::TurtleCommand &com) {
  if (com.state == TFINISHED)
    cerr << world->time  << " (" << comm->getMyTeam() 
	 << " No. " << world->my.number << ")" 
	 << " AdvSkills::exec(TurtleCommand) called when finished; minimal correction step." << endl;

  switch (com.command) {
  case TurtleCommand::TURN:
    basicTurn(com.turnMoment);
    
    break;
  case TurtleCommand::DASH: {
    // determine sign of direction
    double pow = norm(com.v - world->my.v)/world->my.dashPowerRate;
    if (com.state == TGOINGBACKWARDS)
      pow = -pow;
    basicDash(pow);
    break;
  }
  default:
    cerr << world->time  << " (" << comm->getMyTeam() 
	 << " No. " << world->my.number << ")" 
	 << " AdvSkills::exec(TurtleCommand): called without valid command"  
	 << " com  = " << (int) com.command 
	 << " comv = " << com.v << endl; 
    ensure(0);
    noAction();
  }
}
