//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// The Trace class implements a possiblity to send debug traces from
// each agent to FUNSSEL.
// Each module has to define a preprocessor variable to be able to use
// this class. You can then use the preprocessor maros
//
// DECISION(str)
// CONDITION(str)
// VALUE(str)
// LEVELTRC(level,str)
// LEVELDRW(level,MCommand)
// MCMD(MCommand)
//
// to send strings or MCommands to FUNSSEL. The LEVEL* macros can be
// used to define your own commands with different levels than the three
// predefined. If you want a trace to be always on, i.e. even when drawing
// is switched off, you can use the MCMD.
// The macros are defined using strstream, so you can do something like
// DECISION("Me here was at " << time << ". Yoda")
//
// The variables for each module are
// MOD_MC     for the MasterControl
// MOD_WM     for the Worldmodel
// MOD_TACT   for the Tactical Module
// MOD_HAND   for the Ballhandling Module
// MOD_PASS   for the Pass Module
// MOD_GOAL   for the Goalshot Module
// MOD_POS    for the Position Module
// MOD_STDSIT for the Standard Situation Module
// MOD_SKILLS for the Skills
//
// The predefined levels are
// DECISION    1
// CONDITION 101
// VALUE     201
//////////////////////////////////////////////////////////////////////

#ifndef TRACE_H
#define TRACE_H

#include "mcommand.h"
#include <string>
#include <sstream>
#include <fstream>

#ifdef MRBDEBUG

// to make sure sprintf is known
#include <string.h>

#define TRACE__(module,level,str) {ostringstream ostr;\
                                   ostr << str;\
                                   char lineNo[6];\
                                   sprintf(lineNo, "%5d", __LINE__);\
                                   theTrace->write(module,\
                                                   level,\
                                                   __FILE__,\
                                                   __PRETTY_FUNCTION__,\
                                                   lineNo,\
                                                   ostr);\
                                  }
#define TRACE__LVL_MCMD(module,level,mcmd) theTrace->write(module,level,mcmd)
#define TRACE__MCMD(module,mcmd)           theTrace->write(module,mcmd)
#define TRACE__GETCONF(module)             (theTrace->getconf(module))
#define TRACE__DEC(module,str)   TRACE__(module,Trace::DECISION,str)
#define TRACE__COND(module,str)  TRACE__(module,Trace::CONDITION,str)
#define TRACE__VALUE(module,str) TRACE__(module,Trace::VALUE,str)

#undef MOD_STR
#undef MOD_CODE

#ifdef MOD_MC
#define MOD_STR  MC
#define MOD_CODE Trace::MC
#undef MOD_MC
#endif // ifdef MOD_MC

#ifdef MOD_WM
#define MOD_STR  WM
#define MOD_CODE  Trace::WM
#undef MOD_WM
#endif // ifdef MOD_WM

#ifdef MOD_TACT
#define MOD_STR  TACT
#define MOD_CODE  Trace::TACT
#undef MOD_TACT
#endif // ifdef MOD_TACT

#ifdef MOD_HAND
#define MOD_STR  HAND
#define MOD_CODE  Trace::HAND
#undef MOD_HAND
#endif // ifdef MOD_HAND

#ifdef MOD_PASS
#define MOD_STR  PASS
#define MOD_CODE  Trace::PASS
#undef MOD_PASS
#endif // ifdef MOD_PASS

#ifdef MOD_GOAL
#define MOD_STR  GOAL
#define MOD_CODE  Trace::GOAL
#undef MOD_GOAL
#endif // ifdef MOD_GOAL

#ifdef MOD_POS
#define MOD_STR  POS
#define MOD_CODE  Trace::POS
#undef MOD_POS
#endif // ifdef MOD_POS

#ifdef MOD_STDSIT
#define MOD_STR  STDSIT
#define MOD_CODE  Trace::STDSIT
#undef MOD_STDSIT
#endif // ifdef MOD_STDSIT

#ifdef MOD_SKILLS
#define MOD_STR  SKILLS
#define MOD_CODE  Trace::SKILLS
#undef MOD_SKILLS
#endif // ifdef MOD_SKILLS

#define DECISION(str)        TRACE__DEC(MOD_CODE,str)
#define CONDITION(str)       TRACE__COND(MOD_CODE,str)
#define VALUE(str)           TRACE__VALUE(MOD_CODE,str)
#define LEVELTRC(level,str)  TRACE__(MOD_CODE,level,str)
#define LEVELDRW(level,mcmd) TRACE__LVL_MCMD(MOD_CODE,level,mcmd)
#define MCMD(mcmd)           TRACE__MCMD(MOD_CODE,mcmd)
#define GETCONF()	     TRACE__GETCONF(MOD_CODE)
#else // ifdef MRBDEBUG
#define DECISION(str)
#define CONDITION(str)
#define VALUE(str)
#define LEVELTRC(level,str)
#define LEVELDRW(level,mcmd)
#define MCMD(mcmd)
#define GETCONF()
#endif // ifdef MRBDEBUG

class Trace {
public:

  enum Module {
    MC = 0,
    WM,
    TACT,
    HAND,
    PASS,
    GOAL,
    POS,
    STDSIT,
    SKILLS,
    NUM_MODULES
  };

  enum Level {
    DECISION  =   1,
    CONDITION = 101,
    VALUE     = 201
  };

  struct TraceConfig {
    unsigned int players;
    unsigned int level;
    unsigned int min_level;
    bool draw;
  };

  static struct ModuleStrings_t {
    Module module;
    const char * playersStr;
    const char * levelStr;
    const char * min_levelStr;
    const char * drawStr;
    const char * name;
  } moduleStrings[NUM_MODULES];

  Trace( bool FUNSSELMode, const char * configFile );
  ~Trace( void ) { if (writeToFile) of.close(); };

  void setPlayer( int playerNo, int side );

  void write( Module, unsigned int level, 
	      const char * file, const char * function, const char * line,
	      std::ostringstream & );
  void write( Module, unsigned int level, const MCommand & );
  void write( Module, const MCommand & );
  void write( const Worldmodel & );
  TraceConfig *getconf(Module m) { return &modules[m]; };

protected:
  bool writeToFile;
  std::string traceFile;
  std::ofstream of;

  bool FUNSSELMode;
  bool isActive;

  std::string configFile;
  int player;
  TraceConfig modules[NUM_MODULES];

private:
  bool boolStringValue( const char * );

  void setModulePlayerInfo( const char *, Module );
  int  getPlayers( const char * & );

  void setModuleLevelInfo( const char *, unsigned int & level );
};

void theTrace_init( bool FUNSSELMode, const char * configFile );
void theTrace_stop( void );

extern Trace * theTrace;

#endif
