//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef _COACH_H_
#define _COACH_H_

#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sstream>

#include "constants.h"
#include "Comm.h"
#include "worldmodel.h"
#include "audiomgr.h"

using namespace std;

// filename of worldmodel log files
#define WORLD_LOG_FILENAME "world_"
#define WORLD_LOG_EXTENSION ".log"


class Coach {

public:


  // Constructor / Destructor
  ///////////////////////////////////////////////////////////////////////////

  // Generates and logs in a training coach client at the given soccerserver.
  // After the constructor has finished all properties of the agent are set.
  // 
  // Call this constructor first in your derived classes constructors before
  // you implement any further initialization of the agent.

  Coach (const char*  team_name,    // name of own team
	 playing_side side,         // side on which we are playing
	 const char*  host          = DEFAULT_HOST,
	 int          port          = COACH_PORT,   
	 const char*  configfile    = "agent.conf",
	 bool         logging_world = false,   // logs worldmodel for all timesteps
	 bool         FUNSSELmode   = false);  // tells if we can send mcommands


  // Starts execution of the Coach client. 
  // After calling this method, the agent will repeatedly call act()
  // once every timestep to perform the coaching.
  //
  // You cannot override this method.
  void coach();
  
  
  // Called once every timestep if game timer is running, otherwise once every
  // SIMULATOR_STEP ms.
  //
  // Override this method virtual in your derived classes to implement 
  // coach behaviour.
  virtual void act() = 0;
  
  
  // Cleans up and logs out
  //
  // You can put own code in your derived destructors. As the final action call
  // this destructor so the agent will leave the simulation properly 
  // and release its resources (files, sockets, windows, etc).
  virtual ~Coach();  
  
  
  // general information about the client himself and the server state
  /////////////////////////////////////////////////////////////////////

  // obvious
  int          getMyNumber()  {return comm->getMyNumber();} // is 0 for coach
  const char*  getMyTeam()    {return comm->getMyTeam();}
  playing_side getMySide()    {return comm->getMySide();}
  play_mode    getPlayMode()  {return comm->getPlayMode();}
  
  // training coach actions
  void         changePlayerType(int player, int type);
  void         changeMode(play_mode pm);
  void         ear(bool earOn);         // enables/disables hearing
  void         moveObject (const char *object, double x, double y);
  void         moveObject (const char *object, double x, double y, 
			   double dir_rad);
  void         moveObject (const char *object, double x, double y, 
			   double dir_rad, 
			   double xvel, double yvel);
  void         recover(); // resets all players' stamina and effort
  
  // returns the timestep the server is in, our current time.
  // (There is only ONE current time. This is it.)
  long         getTime() {return comm->getServerTime();} 
  
  // sends an audiomessage to other clients immediately
  void         say(AudioMsg msg) {comm->say(msg);}

  // returns true, if a new audiomessage is available in msg, false otherwise
  bool         getNextMsg(AudioMsg &msg) {return comm->getNextMsg(msg);}

  Worldmodel *world;
  
private:

  Comm *comm;

  bool writeWorld;
  ofstream out;
};


#endif
