//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "unistd.h"
#include <cmath>
#include "mytrace.h"   // kick it!!!
#include "worldmodel.h"
#include "Ensure.h"
#include "Polygon.h"
#include "Config.h"

#define MOD_WM
#include "Trace.h"
#define FUNSSELMACROS
#include "drawMacros.h"

using namespace std;

// output generated from this command can be plotted by gplot <file> <time>.
// It shows the update of our position as determined by the flags seen.
//#define FLAGTRACE(output) cout << output

#ifndef FLAGTRACE
#define FLAGTRACE(output)
#else
#warning FLAGTRACE debug trace FLAGTRACE is activated!
#endif


///////////////////////////////////////////////////////////////////////////////////////////////
//
// utility functions concerning positions of objects
//
///////////////////////////////////////////////////////////////////////////////////////////////

bool inField( const Vector2d &pos, double security )
{
  if( fabs( pos.getX() ) > 0.5 * FIELD_LENGTH - security )
    return false;
  if( fabs( pos.getY() ) > 0.5 * FIELD_WIDTH - security )
    return false;
  return true;
}


bool inOurPenaltyArea( const Vector2d &pos, double security )
{
  return ( ( pos.getX() > - 0.5 * FIELD_LENGTH + security ) &&
	   ( pos.getX() < - 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - security ) &&
	   ( fabs( pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - security ) );
}


bool inOtherPenaltyArea( const Vector2d &pos, double security )
{
  return ( ( pos.getX() < 0.5 * FIELD_LENGTH - security ) &&
	   ( pos.getX() > 0.5 * FIELD_LENGTH - PENALTY_AREA_LENGTH + security ) &&
	   ( fabs( pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - security ) );
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// utility functions for quantization and de-quantization
//
///////////////////////////////////////////////////////////////////////////////////////////////

inline double
quantize( double val, double qstep )
  // simulates quantization of an exact variable
{
  return rint( val / qstep ) * qstep;
}


inline double
nodaDist( double dist, double qstep )
  // simulates quantization of an exact distance according to Noda
  // return value is equivalent to return value from server
{
  return quantize( exp( quantize(log( dist + ( 1.0e-10 ) ), qstep ) ), 0.1 );
}


inline Angle
nodaDir( Angle dir )
  // simulates quantization of an exact direction according to Noda
  // return value is equivalent to return value from server
{
  return DEG_TO_RAD( rint( RAD_TO_DEG( dir ) ) );
}


inline Angle
leftDirLimit( Angle nodaDir )
  // returns minimum Angle that would still be quantized to nodaDir
{
  // this "rint" has nothing to do with quantization but with the fact
  // that soccerserver only sends integers and so "nodaDir" must be
  // integer when transformed back to DEG
  return DEG_TO_RAD( rint( RAD_TO_DEG( nodaDir ) ) - 0.5 );
}


inline Angle
rightDirLimit( Angle nodaDir )
  // returns maximum Angle that would still be quantized to nodaDir
{
  // this "rint" has nothing to do with quantization but with the fact
  // that soccerserver only sends integers and so "nodaDir" must be
  // integer when transformed back to DEG
  return DEG_TO_RAD( rint( RAD_TO_DEG( nodaDir ) ) + 0.5 );
}


double
upperDistLimit( double nodaDist, double qstep )
  // returns maximum distance that would still be quantized to nodaDist
{  
  // From the outer quantization, adding 0.05 to nodaDist gives the absolute 
  // maximum for the possible values of the exp(...) expression. 
  
  double d = nodaDist + 0.05;
  
  // Invert exp()

  if( d > 0 )
    d = log( d );
  else
    {
      cerr << "Illegal quantization result from server !" << endl;
      return 1.0;   // senseless value (but d < 0 is senseless as well)
    }

  // As exp is monotonous, log(nodaDist+0.05) is the absolute upper limit
  // for the possible values of the inner quantization. But as these values
  // must also have precision qstep, we can restrict d further to the lower of 
  // the two quantization limits of the inner quantization.
  
  d = floor( d / qstep ) * qstep;

  // Now invert the inner quantization to upper limit.

  d = d + 0.5 * qstep;

  // Finally invert log.

  return exp( d );
}


double
lowerDistLimit( double nodaDist, double qstep )
  // returns minimum distance that would still be quantized to nodaDist
{
  // Like above, only choose lower limits when inverting quantizations,  
  // and pick the upper limit when restricting the inner quantizations value. 

  double d = nodaDist - 0.05;

  // distances are positive
  if( d <= EPS )
    return 0.0;

  // Invert exp()

  d = log(d);

  d = ceil( d / qstep ) * qstep;

  d = d - 0.5 * qstep;

  return exp( d );
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WObject
//
///////////////////////////////////////////////////////////////////////////////////////////////


double
WObject::distToLine( const Vector2d& source, const Vector2d& target ) const
{
  double sx = source.getX();
  double sy = source.getY();

  double tx = target.getX();
  double ty = target.getY();

  double ox = pos.getX();
  double oy = pos.getY();

  // project object onto the line source->target
  double proj = ( ( (sx-tx) * (sx-ox) + (sy-ty) * (sy-oy) ) /
		  ( (sx-tx) * (sx-tx) + (sy-ty) * (sy-ty) ) );
  
  return ( proj < 0 ? distTo( source ) :
	   proj > 1 ? distTo( target ) :
	   sqrt( sqr( ox - sx - proj * (tx-sx) ) + sqr( oy - sy - proj * (ty-sy) ) ) );
}


bool
WObject::inCorridor( Vector2d source, Vector2d target, double d, Angle alpha ) const
{
  double sx = source.getX();
  double sy = source.getY();

  double tx = target.getX();
  double ty = target.getY();

  double ox = pos.getX();
  double oy = pos.getY();

  // check if object is in circle of diameter d around target
  if( distTo( target ) <= 0.5 * d )
    return true;

  // check if object is in very small circle around source
  // so that inCorridor is consistent if source is (e.g.) the calling player's position
  if( distTo( source ) <= 1e-4 )
    return true;

  // check if source is close to target
  if ( norm( target - source ) <= 0.5 * d )
    return false;

  // project object onto the line source->target
  double proj = ( ( (sx-tx) * (sx-ox) + (sy-ty) * (sy-oy) ) /
		  ( (sx-tx) * (sx-tx) + (sy-ty) * (sy-ty) ) );
  
  if ( proj < 0 || proj > 1 )
    // projection point is outside the line
    return false;

  // calculate distance between object and the line source->target
  double dist = sqr( ox - sx - proj * (tx-sx) ) + sqr( oy - sy - proj * (ty-sy) );

  if ( dist > 0.25 * d * d )
    // object is too far away from line
    return false;

  if ( fabs( angleBetween( target - source, pos - source ) ) > alpha )
    // object is not in cone with angle 2 * alpha
    return false;

  return true;
}


bool
WObject::inField( double security ) const
{
  if( fabs( pos.getX() ) > 0.5 * FIELD_LENGTH   - security )
    return false;
  if( fabs( pos.getY() ) > 0.5 * FIELD_WIDTH - security )
    return false;
  return true;
}


bool
WObject::inOurPenaltyArea( double security ) const
{
  return ( ( pos.getX() > - 0.5 * FIELD_LENGTH + security ) &&
	   ( pos.getX() < - 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - security ) &&
	   ( fabs( pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - security ) );
}


bool
WObject::inOtherPenaltyArea( double security ) const
{
  return ( ( pos.getX() < 0.5 * FIELD_LENGTH - security ) &&
	   ( pos.getX() > 0.5 * FIELD_LENGTH - PENALTY_AREA_LENGTH + security ) &&
	   ( fabs( pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - security ) );
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WMovedObject
//
///////////////////////////////////////////////////////////////////////////////////////////////


void
WMovedObject::calcVelocity( const WMovedObject& relObject, double distChange, double dirChange )
{
  Vector2d relpos = pos - relObject.pos;

  v = ( distChange / norm( relpos ) ) * relpos;
  relpos.perp();
  acc( dirChange * relpos + relObject.v );
}


void
WMovedObject::setPos( const SensorObject& object, const WPlayer& my )
{
  // minimize distance error
  pos = my.pos + radAng2Vector2d( 0.5 * ( lowerDistLimit( object.dist, QUANTIZE_STEP ) +
					  upperDistLimit( object.dist, QUANTIZE_STEP ) ),
				  object.dir + my.faceDir );
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WBall
//
///////////////////////////////////////////////////////////////////////////////////////////////


bool
WBall::willCrossOurGoalLine( double distance )
{
  // position where the ball will stop
  Vector2d ballEndPos = pos + ( 1.0 / ( 1.0 - decay ) ) * v;

  // check if ball will cross the line
  if( pos.getX() > -0.5 * FIELD_LENGTH && ballEndPos.getX() < -0.5 * FIELD_LENGTH )
    {
      // check if the crossing will be between the goal posts
      double factor = ( 0.5 * FIELD_LENGTH + pos.getX() ) / ( pos.getX() - ballEndPos.getX() );
      if( fabs( pos.getY() + factor * ( ballEndPos.getY() - pos.getY() ) ) <= distance )
	return true;
    }
  return false;
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WPlayer
//
///////////////////////////////////////////////////////////////////////////////////////////////


bool
WPlayer::ballCatchable( const WBall& ball, const Angle &catchAngle ) const
{
  // check if ball is in penalty area (for >= 2003 servers, player was checked before)
  // REMARK: server checks if ball is PARTIALLY in enalty area, we don't
  // we also do not check if ball is out of field (we might think so even if it isn't!)
  if( ball.pos.getX() > - 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH || fabs( ball.pos.getY() ) > 0.5 * PENALTY_AREA_WIDTH )
    return false;

  // turn ball relpos until catchAngle is zero
  Vector2d relPos = radAng2Vector2d( distTo( ball ), bodyAngleTo( ball ) - catchAngle );

  // check if ball is in catchable area
  if( fabs( relPos.getY() ) <= 0.5 * CATCHABLE_AREA_W )
    if( relPos.getX() >= 0.0 && relPos.getX() <= CATCHABLE_AREA_L )
      return true;

  return false;
}


double
WPlayer::kickPowerRate( const WBall& ball ) const
{
  // Here I use M_PI instead of PI because it is no Vector2d calculation
  return KICK_POWER_RATE * ( 1 - 0.25 * fabs( bodyAngleTo( ball ) ) / M_PI -
			     0.25 * ( distTo( ball ) - BALL_SIZE - size ) / 
			     kickableMargin );
}


double
WPlayer::tackleProb( const WBall& ball ) const
{
  Vector2d relBallPos( ball.pos - pos );
  relBallPos.turn( -bodyDir );

  const double d( relBallPos.getX() > 0 ? TACKLE_DIST : TACKLE_BACK_DIST );
  const double w( TACKLE_WIDTH );
  const double dwp( power( d*w, TACKLE_EXP ) );

  const double P( 1.0 - ( power( relBallPos.getX() * w, TACKLE_EXP ) +
			  power( fabs( relBallPos.getY()) * d, TACKLE_EXP ) ) / dwp );
    
  return max( 0.0, min( 1.0, P ) );
}


void
WPlayer::setBySensorData( SensorPlayer player, const WPlayer& my, WObjectState newState, play_mode playMode )
{
  // approximate player's distance if not valid (should not happen)
  if( !IsDistValid( player.dist ) )
    player.dist = dist( my.pos, pos );

  Vector2d oldPos = pos;
  setPos( player, my );

  // try to simulate stamina loss from position difference
  if( state == VISUAL && newState == VISUAL && playMode == PLAY_MODE_PLAY_ON )
    {
      // player movement since last stamina simulation
      double d = dist( oldPos, pos ) * 0.9;   // error of about 10 % because of noise and sight blur

      double tmp = ( 1.0 - power( decay, age ) ) / ( 1.0 - decay );

      // maximum effect achieved by "sliding"
      d = max( 0.0, d - vmax * decay * tmp );

      // subtract minimum stamina needed for that position change
      stamina += min( 0.0, age * staminaIncMax - d / tmp / dashPowerRate );
      stamina = min( STAMINA_MAX, max( EFFORT_DEC_THR * STAMINA_MAX, stamina ) );
    }

  age = 0;
  coneTime = 0;
  state = newState;
  if( player.goalie != GOALIE_UNKNOWN )
    isGoalie = player.goalie;
  if( IsDirValid( player.body_dir ) )
    {
      bodyAge = 0;
      bodyDir = player.body_dir + my.faceDir;
      if( IsDirValid( player.face_dir ) )
	{
	  faceDir = player.face_dir;
	  headAngle = faceDir - bodyDir;
	}
    }
  if( IsDistValid( player.dist_change ) && IsDirValid( player.dir_change ) )
    calcVelocity( my, player.dist_change, player.dir_change );

  // check if we have seen the player tackling
  if( player.tackle )
    {
      // this is very simple implementation
      // we should check when we have seen this player for the last time
      // and then maybe set cyclesSinceTackle to a lower value
      // but I think it would not work very often
      if( cyclesSinceTackle > TACKLE_CYCLES )
	cyclesSinceTackle = TACKLE_CYCLES;
    }
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// class WPlayerType
//
///////////////////////////////////////////////////////////////////////////////////////////////


ostream & operator << ( ostream &os, const WPlayerType &pt )
{
  return os << "(PlayerType "          << pt.state
	    << " ((speed_max "         << pt.playerSpeedMax 
	    << ") (player_decay "      << pt.playerDecay
	    << ") (inertia_moment "    << pt.inertiaFactor
	    << ") (dash_power_rate "   << pt.dashPowerRate
	    << ")) (player_size "      << pt.playerSize
	    << ") ((kickable_margin "  << pt.kickableMargin
	    << ") (kick_rand "         << pt.kickRand
	    << ")) ((stamina_inc_max " << pt.staminaIncMax
	    << ") (extra_stamina "     << pt.extraStamina
	    << ") (effort_max "        << pt.effortMax
	    << ") (effort_min "        << pt.effortMin
	    << ")))";    
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WAction
//
///////////////////////////////////////////////////////////////////////////////////////////////


void
WAction::store( const CycleAction &newAction )
{
  command = newAction.command;

  moment = 0.0;    // cancel old data
  power  = 0.0;
  pos.setXY( 0.0, 0.0 );

  switch( command )
    {
    case NO_COMMAND:    // param1 == param2 == 0
      break;
      
    case TURN_COMMAND:  // param1 == moment_rad, param2 == 0
      moment = newAction.param1;
      break;
      
    case DASH_COMMAND:  // param1 == power, param2 == 0
      power = newAction.param1;
      break;
      
    case KICK_COMMAND:  // param1 == power, param2 == dir_rad
      power  = newAction.param1;
      moment = newAction.param2;
      break;
      
    case CATCH_COMMAND: // for goalie only; param1 == dir_rad, param2 == 0
      moment = newAction.param1;
      break;
      
    case MOVE_COMMAND:  // only for coach or in certain playmodes; param1 == x, param2 == y
      pos.setXY( newAction.param1, newAction.param2 );
      break;

    case TACKLE_COMMAND:  // param1 == power, param2 == 0
      power  = newAction.param1;
      break;

    default:
      break;
    }
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class Worldmodel
//
///////////////////////////////////////////////////////////////////////////////////////////////

Worldmodel::Worldmodel( Comm* _comm )
{
  comm = _comm;

  time      = 0;   // -1?
  visualAge = ETERNAL;

  // set goal positions (remember: coordinate system is always the same!)
  ourGoal.pos   = GOALPOS( GOAL_LEFT );
  otherGoal.pos = GOALPOS( GOAL_RIGHT );

  cyclesSinceCatch = ETERNAL;
  cyclesSincePointTo = ETERNAL;

  playMode      = PLAY_MODE_UNKNOWN;
  ourSide       = SIDE_UNKNOWN;
  penaltySide   = SIDE_UNKNOWN;
  my.number     = UNKNOWN;
  my.type       = DEFAULT_TYPE;
  my.bodyAge    = 0;

  ourTeam[0].isGoalie = GOALIE_FALSE;   // coach
  ourTeam[0].tacticalNo = 0;

  for( int j = 1; j <= 11; j++ )
    {
      ourTeam[j].tacticalNo = j;   // init tacticalNo with player number
      ourTeam[j].type = DEFAULT_TYPE;
      ourTeam[j].number = j;
      otherTeam[j].number = j;

      // set positions [C]
      ourTeam[j].pos.setXY( -3.0 * j, -FIELD_WIDTH - 3.0 );
      otherTeam[j].pos.setXY( 3.0 * j, -FIELD_WIDTH - 3.0 );
    }

  ourGoalieNo   = UNKNOWN;
  otherGoalieNo = UNKNOWN;

  maxPlayersOurTeam   = max( 0, min( 11, Config::getValueAsInt( "world_maxPlayersOurTeam" ) ) );
  maxPlayersOtherTeam = max( 0, min( 11, Config::getValueAsInt( "world_maxPlayersOtherTeam" ) ) );

  ourScore   = 0;
  otherScore = 0;

  if( comm )
    {
      playMode      = comm->getPlayMode();
      ourSide       = comm->getMySide();
      my.number     = comm->getMyNumber();
      my.pos        = ourTeam[my.number].pos;
      my.tacticalNo = my.number;
      my.isGoalie   = comm->amIGoalie() ? GOALIE_TRUE : GOALIE_FALSE;

      if( comm->amIGoalie() )
	adjustTacticalNumbers( my.number );

      ourScore   = comm->getOurScore();
      otherScore = comm->getOtherScore();
    }
      
  oldMy = my;

  // player types (including default type (ZERO))
  playerType = new WPlayerType[playerParams.player_types];
  playerType[0].state = DEFAULT_TYPE;

  // space for allowed substitutions plus subsitutions before kick off
  ourSubs = new Substitution[playerParams.subs_max + 11];
  otherSubs = new Substitution[playerParams.subs_max + 11];
  numOurSubs = 0;
  numOtherSubs = 0;

  collState = NONE;

  kickCount        = 0;
  dashCount        = 0;
  turnCount        = 0;
  sayCount         = 0;
  turnNeckCount    = 0;
  catchCount       = 0;
  moveCount        = 0;
  changeViewCount  = 0;
  tackleCount      = 0;
  pointToCount     = 0;
  attentionToCount = 0;

  viewWidth        = VIEW_WIDTH_NORMAL;
  viewQuality      = VIEW_QUALITY_HIGH;

  for( int i = 0; i < VISUALSTEPS; i++ )
    {
      vDir[i]   = 0.0;
      vAngle[i] = 0.0;
      vAge[i]   = ETERNAL;

      aOppDir[i] = 0.0;
      aOppAge[i] = ETERNAL;
      aOurDir[i] = 0.0;
      aOurAge[i] = ETERNAL;
    }
  vCounter = VISUALSTEPS - 1;
  aOppCounter = VISUALSTEPS - 1;
  aOurCounter = VISUALSTEPS - 1;

  waitForSenseBody = true;
}


Worldmodel::Worldmodel( const Worldmodel& wm )
{
  comm = wm.comm;
  oldBallRelpos = wm.oldBallRelpos;
  oldMy = wm.oldMy;

  my = wm.my;

  for( int i = 0; i <= 11; i++ )
    {
      ourTeam[i]   = wm.ourTeam[i];
      otherTeam[i] = wm.otherTeam[i];
    }

  ourGoalieNo   = wm.ourGoalieNo;
  otherGoalieNo = wm.otherGoalieNo;

  maxPlayersOurTeam   = wm.maxPlayersOurTeam;
  maxPlayersOtherTeam = wm.maxPlayersOtherTeam;

  ball = wm.ball;
  
  ourGoal   = wm.ourGoal;
  otherGoal = wm.otherGoal;

  collState = wm.collState;

  time = wm.time;
  visualAge = wm.visualAge;

  cyclesSinceCatch   = wm.cyclesSinceCatch;
  cyclesSincePointTo = wm.cyclesSincePointTo;

  ourSide  = wm.ourSide;
  playMode = wm.playMode;

  ourScore   = wm.ourScore;
  otherScore = wm.otherScore;

  action = wm.action;

  kickCount        = wm.kickCount;
  dashCount        = wm.dashCount;
  turnCount        = wm.turnCount;
  sayCount         = wm.sayCount;
  turnNeckCount    = wm.turnNeckCount;
  catchCount       = wm.catchCount;
  moveCount        = wm.moveCount;
  changeViewCount  = wm.changeViewCount;
  tackleCount      = wm.tackleCount;
  pointToCount     = wm.pointToCount;
  attentionToCount = wm.attentionToCount;

  viewWidth   = wm.viewWidth;
  viewQuality = wm.viewQuality;

  playerType = new WPlayerType[playerParams.player_types];
  for( int j = 0; j < playerParams.player_types; j++ )
    playerType[j] = wm.playerType[j];

  ourSubs   = new Substitution[playerParams.subs_max + 11];
  otherSubs = new Substitution[playerParams.subs_max + 11];
  numOurSubs   = wm.numOurSubs;
  numOtherSubs = wm.numOtherSubs;
  for( int j = 0; j < numOurSubs; j++ )
    ourSubs[j % ( playerParams.subs_max + 11 )] = wm.ourSubs[j % ( playerParams.subs_max + 11 )];
  for( int j = 0; j < numOtherSubs; j++ )
    otherSubs[j % ( playerParams.subs_max + 11 )] = wm.otherSubs[j % ( playerParams.subs_max + 11 )];

  for( int i = 0; i < VISUALSTEPS; i++ )
    {
      vDir[i]   = wm.vDir[i];
      vAngle[i] = wm.vAngle[i];
      vAge[i]   = wm.vAge[i];

      aOppDir[i] = wm.aOppDir[i];
      aOppAge[i] = wm.aOppAge[i];
      aOurDir[i] = wm.aOurDir[i];
      aOurAge[i] = wm.aOurAge[i];
    }
  vCounter = wm.vCounter;
  aOppCounter = wm.aOppCounter;
  aOurCounter = wm.aOurCounter;

  waitForSenseBody = wm.waitForSenseBody;
}


Worldmodel& 
Worldmodel::operator = ( const Worldmodel& wm )
{
  comm = wm.comm;
  oldBallRelpos = wm.oldBallRelpos;
  oldMy = wm.oldMy;

  my = wm.my;

  for( int i = 0; i <= 11; i++ )
    {
      ourTeam[i]   = wm.ourTeam[i];
      otherTeam[i] = wm.otherTeam[i];
    }

  ourGoalieNo   = wm.ourGoalieNo;
  otherGoalieNo = wm.otherGoalieNo;

  maxPlayersOurTeam   = wm.maxPlayersOurTeam;
  maxPlayersOtherTeam = wm.maxPlayersOtherTeam;

  ball = wm.ball;
  
  ourGoal   = wm.ourGoal;
  otherGoal = wm.otherGoal;

  collState = wm.collState;

  time = wm.time;
  visualAge = wm.visualAge;

  cyclesSinceCatch   = wm.cyclesSinceCatch;
  cyclesSincePointTo = wm.cyclesSincePointTo;

  ourSide  = wm.ourSide;
  playMode = wm.playMode;

  ourScore   = wm.ourScore;
  otherScore = wm.otherScore;

  action = wm.action;

  kickCount        = wm.kickCount;
  dashCount        = wm.dashCount;
  turnCount        = wm.turnCount;
  sayCount         = wm.sayCount;
  turnNeckCount    = wm.turnNeckCount;
  catchCount       = wm.catchCount;
  moveCount        = wm.moveCount;
  changeViewCount  = wm.changeViewCount;
  tackleCount      = wm.tackleCount;
  pointToCount     = wm.pointToCount;
  attentionToCount = wm.attentionToCount;

  viewWidth   = wm.viewWidth;
  viewQuality = wm.viewQuality;

  delete [] playerType;
  delete [] ourSubs;
  delete [] otherSubs;

  playerType = new WPlayerType[playerParams.player_types];
  for( int j = 0; j < playerParams.player_types; j++ )
    playerType[j] = wm.playerType[j];

  ourSubs   = new Substitution[playerParams.subs_max + 11];
  otherSubs = new Substitution[playerParams.subs_max + 11];
  numOurSubs   = wm.numOurSubs;
  numOtherSubs = wm.numOtherSubs;
  for( int j = 0; j < numOurSubs; j++ )
    ourSubs[j % ( playerParams.subs_max + 11 )] = wm.ourSubs[j % ( playerParams.subs_max + 11 )];
  for( int j = 0; j < numOtherSubs; j++ )
    otherSubs[j % ( playerParams.subs_max + 11 )] = wm.otherSubs[j % ( playerParams.subs_max + 11 )];

  for( int i = 0; i < VISUALSTEPS; i++ )
    {
      vDir[i]   = wm.vDir[i];
      vAngle[i] = wm.vAngle[i];
      vAge[i]   = wm.vAge[i];

      aOppDir[i] = wm.aOppDir[i];
      aOppAge[i] = wm.aOppAge[i];
      aOurDir[i] = wm.aOurDir[i];
      aOurAge[i] = wm.aOurAge[i];
    }
  vCounter = wm.vCounter;
  aOppCounter = wm.aOppCounter;
  aOurCounter = wm.aOurCounter;

  waitForSenseBody = wm.waitForSenseBody;
  return *this;
}


Worldmodel::~Worldmodel() 
{
  delete [] playerType;
  delete [] ourSubs;
  delete [] otherSubs;
}


bool
Worldmodel::checkViewWidth( SensorData &data )
{
  if( data.type != SENSOR_TYPE_SEE )
    return false;

  view_width thisvWidth = VIEW_WIDTH_NARROW;

  double angleNarrow = getViewAngle( VIEW_WIDTH_NARROW );
  double angleNormal = getViewAngle( VIEW_WIDTH_NORMAL );

  for( int i = 0; i < NUM_FLAGS; i++ )
    if( data.flag[i].valid && data.flag[i].which != FLAG_UNKNOWN && IsDistValid( data.flag[i].dist ) )
      if( lowerDistLimit( data.flag[i].dist, QUANTIZE_STEP_L ) > VISIBLE_DISTANCE )
	if( fabs( data.flag[i].dir ) > angleNarrow )
	  {
	    thisvWidth = VIEW_WIDTH_NORMAL;
	    if( fabs( data.flag[i].dir ) > angleNormal )
	      {
		thisvWidth = VIEW_WIDTH_WIDE;
		break;
	      }
	  }

  if( thisvWidth != VIEW_WIDTH_WIDE )
    for( int i = 0; i < 21; i++ )
      if( data.player[i].valid && IsDistValid( data.player[i].dist ) )
	if( lowerDistLimit( data.player[i].dist, QUANTIZE_STEP ) > VISIBLE_DISTANCE )
	  if( fabs( data.player[i].dir ) > angleNarrow )
	    {
	      thisvWidth = VIEW_WIDTH_NORMAL;
	      if( fabs( data.player[i].dir ) > angleNormal )
		{
		  thisvWidth = VIEW_WIDTH_WIDE;
		  break;
		}
	    }

  if( thisvWidth != VIEW_WIDTH_WIDE )
    if( data.ball.valid && IsDistValid( data.ball.dist ) )
      if( lowerDistLimit( data.ball.dist, QUANTIZE_STEP ) > VISIBLE_DISTANCE )
	if( fabs( data.ball.dir ) > angleNarrow )
	  {
	    thisvWidth = VIEW_WIDTH_NORMAL;
	    if( fabs( data.ball.dir ) > angleNormal )
	      thisvWidth = VIEW_WIDTH_WIDE;
	  }

  data.see_view_width = thisvWidth;
  viewWidth = thisvWidth;
  return true;
}


bool 
Worldmodel::calcMyPos( const SensorData &data )
{
  int   numLinesFound = 0;   // number of lines player has seen
  int   numFlagsFound = 0;   // number of flags player has seen

  // step 1 : try to calculate face direction with one line (only when player not out of field)

  int li = 0;

  for( int i = 0; i < 3; i++ )
    {
      if( data.line[i].valid && data.line[i].which != LINE_UNKNOWN )
	{
	  numLinesFound++;
	  li = i;
	}
    }

  if( numLinesFound == 1 )   // this means player is inside the field
    {
      CONDITION("Found only one line");
      Angle alpha = data.line[li].dir;

      switch( data.line[li].which )
	{
	case LINE_RIGHT:
	  alpha = alpha > 0 ? PI_2 - alpha : -PI_2 - alpha;
	  VALUE("right line");
	  break;
	      
	case LINE_BOTTOM:
	  VALUE("bottom line");
	  alpha = alpha > 0 ? PI - alpha : -alpha;
	  break;
	      
	case LINE_LEFT:
	  VALUE("left line");
	  alpha = alpha > 0 ? -PI_2 - alpha : PI_2 - alpha;
	  break;
	      
	case LINE_TOP:
	  VALUE("top line");
	  alpha = alpha > 0 ? -alpha : -PI - alpha;
	  break;
	      
	default:
	  ensure( false );
	}

      Angle l( leftDirLimit( alpha ) );
      Angle r( rightDirLimit( alpha ) );
      alpha = bisect( l, r );
      my.faceDir = alpha;
    }

  // step 1 : count valid flags and save the two closest ones

  int    closest     = -1;
  double closestDist = 10000;
  int    second      = -1;
  double secondDist  = 10000;

  for( int i = 0; i < NUM_FLAGS; i++ )
    {
      if( data.flag[i].valid && data.flag[i].which != FLAG_UNKNOWN && IsDistValid( data.flag[i].dist ) )
	{
	  numFlagsFound++;
	  if( data.flag[i].dist < secondDist )
	    {
	      second = i;
	      secondDist = data.flag[i].dist;
	    }
	  if( data.flag[i].dist < closestDist )
	    {
	      second = closest;
	      secondDist = closestDist;
	      closest = i;
	      closestDist = data.flag[i].dist;
	    }
	}
    }

  if( numFlagsFound == 0 )
    {
      CONDITION("Cannot determine any flag!");
      FLAGTRACE( "Cannot determine any flag!" << endl );

      // if faceDir could be calculated and we are playing from right to left
      // turn coordinate system by 180 deg
      if( numLinesFound == 1 && ourSide == SIDE_RIGHT) {
	my.faceDir += PI;
      }

      return false;
    }

  // step 3 : calculate face direction, if it could not be calculated with one line

  if( numLinesFound != 1 )
    {
      CONDITION("Found more than one line");
      if( numFlagsFound == 1 )
	{
	  FLAGTRACE( "Only determined one flag!" << endl );
	  return false;
	}

      Vector2d diffAbsPos = FLAGPOS( data.flag[closest].which ) - FLAGPOS( data.flag[second].which );
      
      Vector2d diffRelPos = radAng2Vector2d( data.flag[closest].dist, data.flag[closest].dir )
	- radAng2Vector2d( data.flag[second].dist, data.flag[second].dir );
      
      my.faceDir = diffAbsPos.getAngle() - diffRelPos.getAngle();
    }

  // step 4 : special case: if one flag is very close use flag position as my position

  if( closest != -1 && closestDist < upperDistLimit( 0.0, QUANTIZE_STEP_L ) )
    {
      my.pos = FLAGPOS( data.flag[closest].which );
      my.age = 0;
      my.state = VISUAL;

      // if we are playing from right to left turn coordinate system by 180 deg
      if( ourSide == SIDE_RIGHT )
	{
	  my.pos *= -1;
	  my.faceDir += PI;
	}
      return true;
    }

  // step 5 : calculate absolute position with polygon intersection
  //          if intersection fails use old method (weighted use of flag positions)

  Polygon posPoly;                      // intersection polygon of all flag polygons
  bool firstFlag = true;                // is the current flag the first valid flag?

  Vector2d myWeightedPos( 0.0, 0.0 );   // my position calculated with old method (if intersecting fails)
  double weightsum = 0.0;               // weightsum for old calculation method

  for( int i = 0; i < NUM_FLAGS; i++ )
    {
      if( data.flag[i].valid && data.flag[i].which != FLAG_UNKNOWN
	  && IsDistValid( data.flag[i].dist ) )
	{
	  Polygon flagPoly;   // position polygon for current flag

	  Vector2d flagpos = FLAGPOS( data.flag[i].which );   // get flag position
	  
	  // de-quantized absolute area of flag position
	  double top = upperDistLimit( data.flag[i].dist, QUANTIZE_STEP_L );
	  double bot = lowerDistLimit( data.flag[i].dist, QUANTIZE_STEP_L );

	  // assume maximal error of 0.6 deg
	  Angle l( leftDirLimit( data.flag[i].dir ) + my.faceDir - DEG_TO_RAD( 0.6 ) );
	  Angle r( rightDirLimit( data.flag[i].dir ) + my.faceDir + DEG_TO_RAD( 0.6 ) );
	  
	  // limitations of our own position, as determined by flag
	  Vector2d lowleft = flagpos - radAng2Vector2d( top, r );
	  Vector2d lowmiddle = flagpos - radAng2Vector2d( top, bisect( l, r ) );
	  Vector2d lowright = flagpos - radAng2Vector2d( top, l );
	  Vector2d hileft = flagpos - radAng2Vector2d( bot, r );
	  Vector2d hiright = flagpos - radAng2Vector2d( bot, l );
	  
	  // flag polygon coordinates trace
 	  FLAGTRACE( data.time << " "  << lowright.getX() << " " << lowright.getY() << endl );
 	  FLAGTRACE( data.time << " " << hiright.getX() << " " << hiright.getY() << endl );
 	  FLAGTRACE( data.time << " " << hileft.getX() << " " << hileft.getY() << endl );
 	  FLAGTRACE( data.time << " " << lowleft.getX() << " " << lowleft.getY() << endl ); 
 	  FLAGTRACE( data.time << " " << lowmiddle.getX() << " " << lowmiddle.getY() << endl );
 	  FLAGTRACE( data.time << " " << lowright.getX() << " " << lowright.getY() << endl << endl );

	  if( firstFlag )
	    {
	      // just fill posPoly (nothing to intersect yet)
	      posPoly.addPoint( lowleft );
	      posPoly.addPoint( hileft );
	      posPoly.addPoint( hiright );
	      posPoly.addPoint( lowright );
	      posPoly.addPoint( lowmiddle );

	      firstFlag = false;
	    }
	  else
	    {
	      // intersect posPoly with new flag polygon
	      flagPoly.addPoint( lowleft );
	      flagPoly.addPoint( hileft );
	      flagPoly.addPoint( hiright );
	      flagPoly.addPoint( lowright );
	      flagPoly.addPoint( lowmiddle );

	      posPoly ^= flagPoly;
	    }

	  // old position calculation method
	  double weight = 1.0 / ( 1.0 + data.flag[i].dist );
	  weightsum += weight;
	  myWeightedPos += ( flagpos - radAng2Vector2d( 0.5 * ( top + bot ), bisect( l, r ) ) ) * weight;
	}
    }
  
  // intersection polygon coordinates trace
  for( int k = 0; k < posPoly.nop(); k++ )
    FLAGTRACE( data.time << " " << posPoly.points()[k].getX()
	       << " " << posPoly.points()[k].getY() << endl );
  FLAGTRACE( endl );
  
  if( posPoly.nop() > 0 )
    {
      // new method (polygon intersection) worked
      my.pos = posPoly.getInside();

      // new method position trace
      FLAGTRACE( data.time << " " << my.pos.getX()-0.2 << " " << my.pos.getY()-0.2 << endl );
      FLAGTRACE( data.time << " " << my.pos.getX()+0.2 << " " << my.pos.getY()+0.2 << endl << endl );
      FLAGTRACE( data.time << " " << my.pos.getX()-0.2 << " " << my.pos.getY()+0.2 << endl );
      FLAGTRACE( data.time << " " << my.pos.getX()+0.2 << " " << my.pos.getY()-0.2 << endl << endl );
    }
  else
    {
      // new method (polygon intersection) failed, use weighted position
      my.pos = myWeightedPos * ( 1.0 / weightsum );

      // old method position trace
      FLAGTRACE( data.time << " " << my.pos.getX()-0.2 << " " << my.pos.getY() << endl );
      FLAGTRACE( data.time << " " << my.pos.getX()+0.2 << " " << my.pos.getY() << endl << endl );
      FLAGTRACE( data.time << " " << my.pos.getX() << " " << my.pos.getY()-0.2 << endl );
      FLAGTRACE( data.time << " " << my.pos.getX() << " " << my.pos.getY()+0.2 << endl << endl );
    }

  // use my.v (from sense body) for determination of position 
  Vector2d guessPos = 1.0 / my.decay * my.v + oldMy.pos;

  // if we are playing from right to left turn coordinate system by 180 deg
  if( ourSide == SIDE_RIGHT )
    guessPos *= -1;

  // but only if no collision and inside of the polygon
  if( collState == NONE && posPoly.nop() > 0 && posPoly.isInside( guessPos ) )
    my.pos = guessPos;

  // if we are playing from right to left turn coordinate system by 180 deg
  if( ourSide == SIDE_RIGHT )
    {
      my.pos *= -1;
      my.faceDir += PI;
    }

  my.age = 0;
  my.state = VISUAL;

  return true;
}


bool
Worldmodel::adjustPlayers( const SensorData &data ) 
{
  WPlayer dummy;        // for calculating distances in search algorithm
  bool    marked[21];   // for marking sensorPlayers that have been processed

  // ----------------------------------------------------------------------------------------
  // step 1 : update all players with visible team and number (simply overwrite)
  // ----------------------------------------------------------------------------------------

  for( int i = 0; i < 21; i++ )
    {
      marked[i] = false;
      const SensorPlayer &player = data.player[i];
      if( player.valid && player.number > 0 && IsDistValid( player.dist ) )
	{
	  if( player.team == TEAM_US )
	    {
	      // set goalie flags and tactical numbers if goalie is seen for the first time
	      if( player.goalie == GOALIE_TRUE && ourGoalieNo == UNKNOWN )
		adjustTacticalNumbers( player.number );

	      ourTeam[player.number].setBySensorData( player, my, VISUAL, playMode );
	      marked[i] = true;
	    }
	  else if( player.team == TEAM_THEM )
	    {
	      // set goalie flags if goalie is seen for the first time
	      if( player.goalie == GOALIE_TRUE && otherGoalieNo == UNKNOWN )
		{
		  for( int j = 1; j <= 11; j++ )
		    otherTeam[j].isGoalie = GOALIE_FALSE;
		  otherTeam[player.number].isGoalie = GOALIE_TRUE;
		  otherGoalieNo = player.number;
		}
	      otherTeam[player.number].setBySensorData( player, my, VISUAL, playMode );
	      marked[i] = true;
	    }
	}
    }

  // ----------------------------------------------------------------------------------------
  // step 2 : try to update "felt" players (inside VISIBLE_DISTANCE area):
  //          search in both teams for players with small distance to new
  //          sensor information and replace the youngest among them
  // ----------------------------------------------------------------------------------------

  for( int i = 0; i < 21; i++ )
    {
      const SensorPlayer &player = data.player[i];
      if( !marked[i] && player.valid && player.number <= 0 && player.team == TEAM_UNKNOWN )
	if( IsDistValid( player.dist ) && lowerDistLimit( player.dist, QUANTIZE_STEP ) <= VISIBLE_DISTANCE )
	  {
	    long   minAgeOurTeam = ETERNAL;
	    int    minOurIndex   = 0;
	    double minOurDist    = 10000.0;

	    long   minAgeOtherTeam = ETERNAL;
	    int    minOtherIndex   = 0;
	    double minOtherDist    = 10000.0;

	    bool putInOtherTeam = true;
	    dummy.setPos( player, my );
	    double dummyDist = 0.0;

	    // search in both teams
	    for( int j = 1; j <= 11; j++ )
	      {
		// search in our team
		// do not overwrite player himself and players that have already been updated
		// do not overwrite goalie with players that are too far away from our goal
		if( ( j <= maxPlayersOurTeam ) && ( j != my.number ) && ( ourTeam[j].age > 0 ) &&
		    ( j != ourGoalieNo || dummy.pos.getX() < -FIELD_LENGTH / 3.0 ) )
		  {
		    dummyDist = ourTeam[j].distTo( dummy );
		    // look for youngest player among close players
		    if( dummyDist < 2.0 + ourTeam[j].vmax * min( ourTeam[j].age, 3 ) )
		      {
			if( ourTeam[j].age < minAgeOurTeam ||
			    ( ourTeam[j].age == minAgeOurTeam && dummyDist < minOurDist ) )
			  {
			    minAgeOurTeam = ourTeam[j].age;
			    minOurIndex = j;
			    minOurDist = dummyDist;
			  }
		      }
		  }

		// search in opponent team
		// do not overwrite players that have already been updated
		// do not overwrite goalie with players that are too far away from opponent goal
		if( ( j <= maxPlayersOtherTeam ) && ( otherTeam[j].age > 0 ) &&
		    ( j != otherGoalieNo || dummy.pos.getX() > FIELD_LENGTH / 3.0 ) )
		  {
		    dummyDist = otherTeam[j].distTo( dummy );
		    // look for youngest player among close players
		    if( dummyDist < 2.0 + otherTeam[j].vmax * min( otherTeam[j].age, 3 ) )
		      {
			if( otherTeam[j].age < minAgeOtherTeam ||
			    ( otherTeam[j].age == minAgeOtherTeam && dummyDist < minOtherDist ) )
			  {
			    minAgeOtherTeam = otherTeam[j].age;
			    minOtherIndex = j;
			    minOtherDist = dummyDist;
			  }
		      }
		  } 
	      }

	    // if there are close players in both teams decide which to take
	    if( minAgeOtherTeam < ETERNAL && minAgeOurTeam < ETERNAL )
	      {
		marked[i] = true;
		// only put player in our team if age and distance fit very well
		// [MW] Spieler in Feelrange sind immer gleich alt, deshalb <= statt <
		if( minAgeOurTeam <= minAgeOtherTeam && minOurDist < minOtherDist )
		  putInOtherTeam = false;
	      }
	    // is there any close player in the opponent team?
	    else if( minAgeOtherTeam < ETERNAL )
	      {
		marked[i] = true;
	      }
	    // if not - maybe in our team?
	    else if( minAgeOurTeam < ETERNAL )
	      {
		marked[i] = true;
		putInOtherTeam = false;
	      }
	    // replace player
	    if( marked[i] )
	      {
		if( putInOtherTeam )
		  otherTeam[minOtherIndex].setBySensorData( player, my,
							    minAgeOtherTeam < ETERNAL &&
							    otherTeam[minOtherIndex].state != UNSAFE ? FUZZY : UNSAFE );
		else
		  ourTeam[minOurIndex].setBySensorData( player, my,
							minAgeOurTeam < ETERNAL &&
							ourTeam[minOurIndex].state != UNSAFE ? FUZZY : UNSAFE );
	      }
	  }
    }

  // ----------------------------------------------------------------------------------------
  // step 3 : try to update players with visible color but no visible number
  //          by comparing them with the simulated players:
  //          search for players with small distance to new sensor information
  //          and replace the youngest among them
  // ----------------------------------------------------------------------------------------

  for( int i = 0; i < 21; i++ )
    {
      const SensorPlayer &player = data.player[i];
      if( !marked[i] && player.valid && player.number <= 0 && IsDistValid( player.dist ) )
	{
	  // maximal error of distance caused by Noda's quantization
	  double distError = ( upperDistLimit( player.dist, QUANTIZE_STEP ) -
			       lowerDistLimit( player.dist, QUANTIZE_STEP ) );

	  if( player.team == TEAM_US )
	    {
	      long   minAge = ETERNAL;
	      double minDist = 10000;
	      int    minAgeIndex = 0;
	      dummy.setPos( player, my );
	      
	      for( int j = 1; j <= maxPlayersOurTeam; j++ )
		{
		  // do not overwrite player himself and players that have already been updated
		  // do not overwrite goalie with players that are too far away from our goal
		  if( ( j != my.number ) && ( ourTeam[j].age > 0 ) &&
		      ( j != ourGoalieNo || dummy.pos.getX() < -FIELD_LENGTH / 3.0 ) )
		    {
		      // look for youngest player among close players
		      if( ourTeam[j].distTo( dummy ) <
			  2.0 + ourTeam[j].vmax * min( ourTeam[j].age, 3 ) + distError )
			{
			  if( ourTeam[j].age < minAge - 1 ||
			      ( ourTeam[j].age <= minAge + 1 && ourTeam[j].distTo( dummy ) < minDist ) )
			    {
			      minAge = ourTeam[j].age;
			      minAgeIndex = j;
			      minDist = ourTeam[j].distTo( dummy );
			    }
			}
		    } 
		}
	      // we have found one close player
	      if( minAge < ETERNAL )
		{
		  ourTeam[minAgeIndex].setBySensorData( player, my,
							ourTeam[minAgeIndex].state != UNSAFE ? FUZZY : UNSAFE );
		  marked[i] = true;
		}
	    }
	  else if( player.team == TEAM_THEM )
	    {
	      long   minAge = ETERNAL;
	      double minDist = 10000;
	      int    minAgeIndex = 0;
	      dummy.setPos( player, my );

	      for( int j = 1; j <= maxPlayersOtherTeam; j++ )
		{
		  // do not overwrite players that have already been updated
		  // do not overwrite goalie with players that are too far away from opponent goal
		  if( ( otherTeam[j].age > 0 ) &&
		      ( j != otherGoalieNo || dummy.pos.getX() > FIELD_LENGTH / 3.0 ) )
		    {
		      // look for youngest player among close players
		      if( otherTeam[j].distTo( dummy ) <
			  2.0 + otherTeam[j].vmax * min( otherTeam[j].age, 3 ) + distError )
			{
			  if( otherTeam[j].age < minAge - 1 ||
			      ( otherTeam[j].age <= minAge + 1 && otherTeam[j].distTo( dummy ) < minDist ) )
			    {
			      minAge = otherTeam[j].age;
			      minAgeIndex = j;
			      minDist = otherTeam[j].distTo( dummy );
			    }
			}
		    } 
		}
	      // we have found one close player
	      if( minAge < ETERNAL )
		{
		  otherTeam[minAgeIndex].setBySensorData( player, my,
							  otherTeam[minAgeIndex].state != UNSAFE ? FUZZY : UNSAFE );
		  marked[i] = true;
		}
	    }
	}
    }

  // ----------------------------------------------------------------------------------------
  // step 4 : try to update the remaining felt players by comparing them with players in
  //          the opponent team:
  //          search for player with smallest distance to new sensor information and replace
  //          him if he is old enough or has coneTime > 0
  // ----------------------------------------------------------------------------------------

  for( int i = 0; i < 21; i++ )
    {
      const SensorPlayer &player = data.player[i];
      if( !marked[i] && player.valid && player.number <= 0 && player.team == TEAM_UNKNOWN )
	if( IsDistValid( player.dist ) && lowerDistLimit( player.dist, QUANTIZE_STEP ) <= VISIBLE_DISTANCE )
	  {
	    int minDistIndex = -1;
	    double minDist = 10000;
	    dummy.setPos( player, my );

	    for( int j = 1; j <= maxPlayersOtherTeam; j++ )
	      {
		// do not overwrite players that have already been updated
		// do not overwrite goalie with players that are too far away from opponent goal
		if( ( otherTeam[j].age > 0 ) &&
		    ( j != otherGoalieNo || dummy.pos.getX() > FIELD_LENGTH / 3.0 ) )
		  {
		    // look for closest player
		    if( minDistIndex == -1 || otherTeam[j].distTo( dummy ) < minDist )
		      if( otherTeam[j].age > 3 || otherTeam[j].coneTime > 0 )
			{
			  minDist = otherTeam[j].distTo( dummy );
			  minDistIndex = j;
			}
		  }
	      }
	    // we have found one close player
	    if( minDistIndex != -1 )
	      {
		otherTeam[minDistIndex].setBySensorData( player, my, UNSAFE );
		marked[i] = true;
	      }
	  }
    }

  // ----------------------------------------------------------------------------------------
  // step 5 : try to update the remaining players with visible color but no visible number
  //          by comparing them with the simulated players:
  //          search for player with smallest distance to new sensor information and replace
  //          him if he is old enough
  // ----------------------------------------------------------------------------------------

  for( int i = 0; i < 21; i++ )
    {
      const SensorPlayer &player = data.player[i];
      if( !marked[i] && player.valid && player.number <= 0 && IsDistValid( player.dist ) )
	{
	  if( player.team == TEAM_US )
	    {
	      int minDistIndex = -1;
	      double minDist = 10000;
	      dummy.setPos( player, my );

	      for( int j = 1; j <= maxPlayersOurTeam; j++ )
		{
		  // do not overwrite player himself and players that have already been updated
		  // do not overwrite goalie with players that are too far away from our goal
		  if( ( j != my.number ) && ( ourTeam[j].age > 0 ) &&
		      ( j != ourGoalieNo || dummy.pos.getX() < -FIELD_LENGTH / 3.0 ) )
		    {
		      // look for closest player
		      if( minDistIndex == -1 || ourTeam[j].distTo( dummy ) < minDist )
			if( ourTeam[j].age > 3 )
			  {
			    minDist = ourTeam[j].distTo( dummy );
			    minDistIndex = j;
			  }
		    }
		}
	      // we have found one close player
	      if( minDistIndex != -1 )
		{
		  ourTeam[minDistIndex].setBySensorData( player, my, UNSAFE );
		  marked[i] = true;
		}
	    }
	  else if( player.team == TEAM_THEM )
	    {
	      int minDistIndex = -1;
	      double minDist = 10000;
	      dummy.setPos( player, my );

	      for( int j = 1; j <= maxPlayersOtherTeam; j++ )
		{
		  // do not overwrite players that have already been updated
		  // do not overwrite goalie with players that are too far away from opponent goal
		  if( ( otherTeam[j].age > 0 ) &&
		      ( j != otherGoalieNo || dummy.pos.getX() > FIELD_LENGTH / 3.0 ) )
		    {
		      // look for closest player
		      if( minDistIndex == -1 || otherTeam[j].distTo( dummy ) < minDist )
			if( otherTeam[j].age > 3 )
			  {
			    minDist = otherTeam[j].distTo( dummy );
			    minDistIndex = j;
			  }
		    }
		}
	      // we have found one close player
	      if( minDistIndex != -1 )
		{
		  otherTeam[minDistIndex].setBySensorData( player, my, UNSAFE );
		  marked[i] = true;
		}
	    }
	}
    }

  // ----------------------------------------------------------------------------------------
  // step 6 : - update coneTime for all players
  //          - cleanup : set for players which are in view cone (and not too far away)
  //            or in "feel radius" but too old age to ETERNAL and state to UNSAFE
  // ----------------------------------------------------------------------------------------

  for( int i = 1; i <= 11; i++ )
    {
      if( i != my.number )
	if( ( fabs( my.faceAngleTo( ourTeam[i] ) ) < fabs( getViewAngle( data.see_view_width ) ) &&
	      my.distTo( ourTeam[i] ) < TEAM_FAR_LENGTH ) ||
	    my.distTo( ourTeam[i] ) <= VISIBLE_DISTANCE )
	  if( ourTeam[i].age > 0 )
	    ourTeam[i].coneTime++;

      if( ( fabs( my.faceAngleTo( otherTeam[i] ) ) < fabs( getViewAngle( data.see_view_width ) ) &&
	    my.distTo( otherTeam[i] ) < TEAM_FAR_LENGTH ) ||
	  my.distTo( otherTeam[i] ) <= VISIBLE_DISTANCE )
	if( otherTeam[i].age > 0 )
	  otherTeam[i].coneTime++;
    }

  for( int i = 1; i <= 11; i++ )
    {
      // [MW] don't set opponents' age to eternal who just slipped out of the feelrange
      if( i != my.number )
	if( ( my.distTo( ourTeam[i] ) < 0.8 * VISIBLE_DISTANCE && ourTeam[i].age > 0 ) ||
	    ourTeam[i].coneTime > 3 )
	  {
	    ourTeam[i].age = ETERNAL;
	    ourTeam[i].state = UNSAFE;
	  }
      // do not set opponent goalie's age to ETERNAL  
      if( i != otherGoalieNo )
	if( ( my.distTo( otherTeam[i] ) < 0.8 * VISIBLE_DISTANCE && otherTeam[i].age > 0 ) ||
	    otherTeam[i].coneTime > 3 )
	  {
	    otherTeam[i].age = ETERNAL;
	    otherTeam[i].state = UNSAFE;
	  }
    }

  return true;
}


bool 
Worldmodel::adjustBall( SensorData data )
{
  bool positionAdjusted = false;
  bool velocityAdjusted = false;

  DECISION("adjustball my.v: " << my.v);
  // set velocity to zero in standard situations
  switch( playMode )
    {
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
    case PLAY_MODE_PAUSE :
    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_HUMAN_JUDGE :
    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_GOALIE_CATCH_US :
    case PLAY_MODE_GOALIE_CATCH_THEM :
    case PLAY_MODE_OFFSIDE_US :
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_BACKPASS_US :
    case PLAY_MODE_BACKPASS_THEM :
    case PLAY_MODE_FREEKICK_FAULT_US :
    case PLAY_MODE_FREEKICK_FAULT_THEM :
      ball.v.setXY( 0.0, 0.0 );
      velocityAdjusted = true;
      break;

    default :
      break;
    }

  // set ball position in standard situations
  switch( playMode )
    {
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_GOAL_THEM :
      ball.pos.setXY( 0.0, 0.0 );
      positionAdjusted = true;
      ball.state = VISUAL;   // this makes sense even if we have not seen the ball!
      break;

    case PLAY_MODE_CORNER_KICK_US :
      // of course, this only works when we have seen the ball just before
      ball.pos.setXY( 0.5 * FIELD_LENGTH - CKICK_MARGIN,
		      ball.pos.getY() < 0.0 ? -0.5 * FIELD_WIDTH + CKICK_MARGIN
		      : 0.5 * FIELD_WIDTH - CKICK_MARGIN );
      // it is better not to use the following line!
      // positionAdjusted = true;
      break;

    case PLAY_MODE_CORNER_KICK_THEM :
      // of course, this only works when we have seen the ball just before
      ball.pos.setXY( -0.5 * FIELD_LENGTH + CKICK_MARGIN,
		      ball.pos.getY() < 0.0 ? -0.5 * FIELD_WIDTH + CKICK_MARGIN
		      : 0.5 * FIELD_WIDTH - CKICK_MARGIN );
      // it is better not to use the following line!
      // positionAdjusted = true;
      break;

    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_KICK_IN_THEM :
      // of course, this only works when we have seen the ball just before
      if( fabs( ball.pos.getY() ) > 0.5 * FIELD_WIDTH - 3.0 )
	ball.pos.setY( ball.pos.getY() < 0.0 ? -0.5 * FIELD_WIDTH : 0.5 * FIELD_WIDTH );
      // it is better not to use the following line!
      // positionAdjusted = true;
      break;

    default :
      break;
    }

  // if we have got visual information AND seen the ball, calculate its position
  // and velocity if this has not been done above
  if( data.type == SENSOR_TYPE_SEE )
    if( data.ball.valid )
      {
	DECISION("seen ball");
	// keep simulated ball (need that for collision check and position difference comparison)
	WBall simBall = ball;

	// we have seen the ball, so set its age and coneTime to zero
	ball.age = 0;
	ball.coneTime = 0;

	// approximate dist if not valid; in any case ball is a little away, so it needs a closer look!
	// this is only needed for bad quality sensor data - but we never use that feature!
	if( !IsDistValid( data.ball.dist ) )
	  data.ball.dist = my.distTo( ball ) < 20.0 ? 20.0 : my.distTo( ball );

	// update ball position (update of my position should be done before!)
	if( !positionAdjusted )
	  {
	    // minimize dist error by taking the arithmetic mean of lower and upper bound
	    double newDist = 0.5 * ( lowerDistLimit( data.ball.dist, QUANTIZE_STEP ) +
				     upperDistLimit( data.ball.dist, QUANTIZE_STEP ) );
	    ball.pos = my.pos + radAng2Vector2d( newDist, data.ball.dir + my.faceDir );
	    positionAdjusted = true;
	  }

	ball.state = VISUAL;

	// update ball velocity
	if( !velocityAdjusted )
	  {
	    // check if dist_change and dir_change are valid, then update ball velocity
	    if( IsDistValid( data.ball.dist_change ) && IsDirValid( data.ball.dir_change ) )
	      {
		VALUE("with velocity relpos: "<< ball.pos-my.pos << " distChange: " << 
		      data.ball.dist_change << " dirChange: " << data.ball.dir_change );
		ball.calcVelocity( my, data.ball.dist_change, data.ball.dir_change );
		velocityAdjusted = true;
	      }
	  }

	// check if there might have been a collision and set velocity to zero
	// this does not work in any case! (because of distance quantization)
	// (e.g. the possible values 0.385 ... 0.472 (for a player with standard
	// size 0.3) are all quantized to 0.4)
	if( !velocityAdjusted )
	  {
	    if( lowerDistLimit( data.ball.dist, QUANTIZE_STEP ) <= BALL_SIZE + my.size &&
		upperDistLimit( data.ball.dist, QUANTIZE_STEP ) >= BALL_SIZE + my.size )
	      {
		// if there was a collision player velocity (from sense body) must be small
		// [C] soccerserver also uses 0.01 constant for velocity quantization
		if( norm( my.v ) - 0.005 <= 0.1 * my.vmax * my.decay )
		  {
		    VALUE("possible player collision");
		    // of course ball velocity is not exactly zero, but very small
		    ball.v.setXY( 0.0, 0.0 );
		    velocityAdjusted = true;
		    ball.state = FUZZY;   // because information might be incorrect
		  }
		else
		  {
		    VALUE("no player collision");
		    if( collState != NONE )
		      {
			VALUE("multiplying ball.v with -10.0");
			// simulation of ball collision must have been wrong
			ball.v *= -10.0;
		      }
		  }
	      }
	    else if( collState != NONE )
	      {
			VALUE("multiplying ball.v with -10.0");
		// simulation of ball collision must have been wrong
		ball.v *= -10.0;
	      }
	  }

	// or calculate velocity from position difference, but only if ball is young
	// enough and the position difference does not fit our simulated velocity
	if( !velocityAdjusted && data.ball.dist <= 2 * VISIBLE_DISTANCE )
	  {
	    switch( simBall.age )
	      {
	      case 1 :   // visual/audial info in the last cycle
		{
		  // problem: there might have been a collision with another player
		  // in this case we can not use my.pos and my.v for the calculation
		  bool maybePlayerCollision = false;

		  // in case of a collision my velocity must be low
		  // [C] soccerserver also uses 0.01 constant for velocity quantization
		  if( norm( my.v ) - 0.005 <= 0.1 * my.vmax * my.decay )
		    {
		      // check maximum player size
		      double maxPlayerSize = PLAYER_SIZE;
		      for( int i = 0; i < playerParams.player_types; i++ )
			if( playerType[i].state != UNKNOWN )
			  if( playerType[i].playerSize > maxPlayerSize )
			    maxPlayerSize = playerType[i].playerSize;

		      for( int i = 0; i < 21; i++ )
			{
			  SensorPlayer &player = data.player[i];
			  if( player.valid && IsDistValid( player.dist ) )
			    if( lowerDistLimit( player.dist, QUANTIZE_STEP ) <= my.size + maxPlayerSize )
			      {
				maybePlayerCollision = true;
				break;
			      }
			}
		    }

		  VALUE( "COLL: Posdiff = " << simBall.distTo( ball ) << ", ball.age was 1" << endl );

		  // in case of collision with other player (guessed) use old algorithm
		  if( maybePlayerCollision )
		    {
		      if( !velocityAdjusted )
			{
			  VALUE("using old algorithm");
			  simBall.pos -= simBall.v * ( 1.0 / BALL_DECAY );
			  ball.v = ( ball.pos - simBall.pos ) * BALL_DECAY;
			  velocityAdjusted = true;
			  ball.state = FUZZY;   // because information might be incorrect
			}
		    }
		  else
		    {
		      VALUE("using position difference algorithm");
		      // calculate ball velocity from position difference
		      ball.v = ( ball.pos - my.pos + my.v * ( 1.0 / my.decay ) - oldBallRelpos ) * BALL_DECAY;
		      velocityAdjusted = true;
		    }
		}
		break;

	      case 2 :   // latest visual/audial info two cycles ago

		if( simBall.distTo( ball ) > 0.6 && !velocityAdjusted )   // [C] (but ok)
		  {
		    VALUE("using 2 step algorithm");
		    // The simulation in this case is not very reliable.
		    // The main assumption is that the view is NORMAL (because we receive
		    // visual data every 2 cycles). In this case the absolute position
		    // of the player is not too bad.

		    // ball position two timesteps ago
		    simBall.pos -= simBall.v * ( ( 1.0 + BALL_DECAY ) / sqr( BALL_DECAY ) );

		    // calculate ball velocity from position difference
		    ball.v = ( ball.pos - simBall.pos ) * ( sqr( BALL_DECAY ) / ( 1.0 + BALL_DECAY ) );
		    velocityAdjusted = true;
		    ball.state = FUZZY;   // because information might be incorrect
		  }
		break;

	      default :  // simulated ball too old
		if( !velocityAdjusted )
		  ball.state = FUZZY;
		break;
	      }
	  }
      }
    else
      {
	// we have got visual information but not seen the ball
	// check if simulated ball position is in view cone or in "feel radius" and update coneTime
	if( fabs( my.faceAngleTo( ball ) ) < fabs( getViewAngle( data.see_view_width ) ) ||
	    my.distTo( ball ) <= VISIBLE_DISTANCE )
	  ball.coneTime++;
      }

  // values for ball.state now:
  // -----------------------------------------------------------------------------------
  // case                                                                        state
  // -----------------------------------------------------------------------------------
  // ball vel. seen                                                              VISUAL
  // ball vel. not seen, maybe player<->ball collision                           FUZZY
  // ball vel. not seen, simBall's age == 1, sim. fits posdiff                   VISUAL
  // ball vel. not seen, simBall's age == 1, sim. does not fit posdiff           VISUAL
  // ball vel. not seen, simBall's age == 1, sim. does not fit posdiff,
  //      collision with other player(s) possible                                FUZZY
  // ball vel. not seen, simBall's age == 2, sim. fits posdiff                   VISUAL
  // ball vel. not seen, simBall's age == 2, sim. does not fit posdiff           FUZZY
  // ball vel. not seen, simBall's age > 2                                       FUZZY

  return ( positionAdjusted || velocityAdjusted );
}


bool
Worldmodel::adjustGoals( const SensorData &data )
{
  // set goal positions (remember: coordinate system is always the same!)
  ourGoal.pos = GOALPOS( GOAL_LEFT );
  otherGoal.pos = GOALPOS( GOAL_RIGHT );

  // use visual information only to set "age" (so that player knows when he
  // has seen the goals for the last time
  for( int i = 0; i < 2; i++ )
    if( data.goal[i].valid )
      if( data.goal[i].team == TEAM_THEM )
	{
	  otherGoal.age = 0;
	  otherGoal.state = VISUAL;
	}
      else if( data.goal[i].team == TEAM_US )
	{
	  ourGoal.age = 0;
	  otherGoal.state = VISUAL;
	}
  
  return true;
}


void
Worldmodel::adjustTacticalNumbers( int newOurGoalieNo )
{
  ourGoalieNo = newOurGoalieNo;

  // set tactical numbers of our team
  for( int j = 1; j <= 11; j++ )
    {
      ourTeam[j].isGoalie = GOALIE_FALSE;
      if( j < ourGoalieNo )
	ourTeam[j].tacticalNo = j + 1;
      else
	ourTeam[j].tacticalNo = j;
    }
  ourTeam[ourGoalieNo].isGoalie = GOALIE_TRUE;
  ourTeam[ourGoalieNo].tacticalNo = 1;

  my.isGoalie = ourTeam[my.number].isGoalie;
  my.tacticalNo = ourTeam[my.number].tacticalNo;
}


bool   
Worldmodel::adjustExact( const SensorData &data )
{
  time = data.time;

  // adjust my (== coach's) position
  my.pos.setXY( 0.0, 0.0 );
  my.v.setXY( 0.0, 0.0 );
  my.age      = 0;
  my.decay    = 0.0;
  my.vmax     = 0.0;
  my.bodyDir  = 0.0;
  my.faceDir  = 0.0;
  my.number   = 0;
  my.isGoalie = GOALIE_FALSE;   // is coach !

  // adjust ball
  if( data.exactball.valid )
    {
      ball.age = 0;
      if( ourSide == SIDE_LEFT )
	{
	  ball.pos.setXY( data.exactball.x, data.exactball.y );
	  ball.v.setXY( data.exactball.x_vel, data.exactball.y_vel );
	}
      else
	{
	  ball.pos.setXY( - data.exactball.x, - data.exactball.y );
	  ball.v.setXY( - data.exactball.x_vel, - data.exactball.y_vel );
	}
    }

  // adjust players
  for( int i = 0; i < 22; i++ )
    {
      const ExactPlayer &player = data.exactplayer[i];
      if( player.valid && player.number > 0 )   // then all information is complete, esp. the numbers
	{
	  if( player.team == TEAM_US )
	    {
	      // set goalie flags and tactical numbers if goalie is seen for the first time
	      if( player.goalie == GOALIE_TRUE && ourGoalieNo == UNKNOWN )
		adjustTacticalNumbers( player.number );

	      ourTeam[player.number].age = 0;
	      ourTeam[player.number].state = VISUAL;
	      ourTeam[player.number].isGoalie = player.goalie;
	      ourTeam[player.number].bodyDir = player.body_dir;
	      ourTeam[player.number].faceDir = player.face_dir;
	      if( ourSide == SIDE_LEFT )
		{
		  ourTeam[player.number].pos.setXY( player.x, player.y );
		  ourTeam[player.number].v.setXY( player.x_vel, player.y_vel );
		}
	      else
		{
		  ourTeam[player.number].pos.setXY( - player.x, - player.y );
		  ourTeam[player.number].v.setXY( - player.x_vel, - player.y_vel );
		}
	    }
	  else if( player.team == TEAM_THEM )
	    {
	      // set goalie flags if goalie is seen for the first time
	      if( player.goalie == GOALIE_TRUE && otherGoalieNo == UNKNOWN )
		{
		  for( int j = 1; j <= 11; j++ )
		    otherTeam[j].isGoalie = GOALIE_FALSE;
		  otherTeam[player.number].isGoalie = GOALIE_TRUE;
		  otherGoalieNo = player.number;
		}
	      
	      otherTeam[player.number].age = 0;
	      otherTeam[player.number].state = VISUAL;
	      otherTeam[player.number].isGoalie = player.goalie;
	      otherTeam[player.number].bodyDir = player.body_dir;
	      otherTeam[player.number].faceDir = player.face_dir;
	      if( ourSide == SIDE_LEFT )
		{
		  otherTeam[player.number].pos.setXY( player.x, player.y );
		  otherTeam[player.number].v.setXY( player.x_vel, player.y_vel );
		}
	      else
		{
		  otherTeam[player.number].pos.setXY( - player.x, - player.y );
		  otherTeam[player.number].v.setXY( - player.x_vel, - player.y_vel );
		}
	    }
	}
    }

  // adjust goals
  for( int i = 0; i < 2; i++ )
    {
      if( data.exactgoal[i].valid )
	{
	  if( data.exactgoal[i].team == TEAM_US )
	    {
	      if( ourSide == SIDE_LEFT )
		ourGoal.pos.setXY( data.exactgoal[i].x, data.exactgoal[i].y );
	      else
		ourGoal.pos.setXY( -data.exactgoal[i].x, -data.exactgoal[i].y );
	      ourGoal.age = 0;
	    }
	  if( data.exactgoal[i].team == TEAM_THEM )
	    {
	      if( ourSide == SIDE_LEFT )
		otherGoal.pos.setXY( data.exactgoal[i].x, data.exactgoal[i].y );
	      else
		otherGoal.pos.setXY( -data.exactgoal[i].x, -data.exactgoal[i].y );
	      otherGoal.age = 0;
	    }
	}
    }
  
  return true;
}


Angle 
Worldmodel::getViewAngle( view_width thisVwidth ) const
{
  switch( thisVwidth )
    {
    case VIEW_WIDTH_NARROW : return DEG_TO_RAD( 22.5 );
    case VIEW_WIDTH_NORMAL : return DEG_TO_RAD( 45.0 );
    case VIEW_WIDTH_WIDE   : return DEG_TO_RAD( 90.0 );
    default : return( 0.0 );
    }

  return 0.0;
}


long
Worldmodel::dirAge( const Angle& alpha ) const
{
  long age = ETERNAL;

  for( int i = 0; i < VISUALSTEPS; i++ )
    if( fabs( vDir[i] - alpha ) <= fabs( vAngle[i] ) )
      if( vAge[i] < age )
	age = vAge[i];

  return age;
}


int
Worldmodel::lookAroundFinished( ) const
{
  //[hack] bser, bser Hack, irgendwann mal ganz neu schreiben [MW]
  const long maxAge = 7;
  int notSeen = 0;

  for( double i = -M_PI; i < M_PI; i+= M_PI / 8.0 )
      if ( dirAge( Angle(i + my.bodyDir) ) > maxAge ) notSeen++;

  return ( notSeen < 8 ) ;
}


int
Worldmodel::ourTeamClosestTo( const Vector2d& target, long maxAge, int index ) const
{
  bool          sorted[12];                // for marking players that have already been sorted
  bool          validEntryFound = false;   // Is there a player left who can be sorted?
  int           validEntryIndex = 0;       // index of next player to be sorted
  
  if( index < 1 ) return UNKNOWN;
  if( index > 11 ) return UNKNOWN;

  for( int i = 1; i <= 11; i++ )
    sorted[i] = false;
  
  // search until "index" players are marked in index array
  for( int i = 0; i < index; i++ )
    {
      validEntryFound = false;
      
      // search for next player for index array
      for( int j = 1; j <= 11; j++ )
	{
	  // Is player young enough and farer away than those we already sorted?
	  if ( !sorted[j] && ourTeam[j].age <= maxAge )
	    {
	      if( validEntryFound )
		{
		  // do not use dist() for distance check (avoid sqare root)
		  if( sqrnorm( ourTeam[j].pos - target ) < sqrnorm( ourTeam[validEntryIndex].pos - target ) )
		    validEntryIndex = j;
		}
	      else
		{
		  validEntryFound = true;
		  validEntryIndex = j;
		}
	    }
	}

      if( validEntryFound )
	sorted[validEntryIndex] = true;
      else continue;
    }

  if( validEntryFound )
    return validEntryIndex;   // return player number

  return UNKNOWN;   // not enough valid players found
}


int
Worldmodel::otherTeamClosestTo( const Vector2d& target, long maxAge, int index ) const
{
  bool          sorted[12];                // for marking players that have already been sorted
  bool          validEntryFound = false;   // Is there a player left who can be sorted?
  int           validEntryIndex = 0;       // index of next player to be sorted

  if( index < 1 ) return UNKNOWN;
  if( index > 11 ) return UNKNOWN;

  for( int i = 1; i <= 11; i++ )
    sorted[i] = false;

  // search until "index" players are marked in index array
  for( int i = 0; i < index; i++ )
    {
      validEntryFound = false;
      
      // search for next player for index array
      for( int j = 1; j <= 11; j++ )
	{
	  // Is player young enough and farer away than those we already sorted?
	  if ( !sorted[j] && otherTeam[j].age <= maxAge )
	    {
	      if( validEntryFound )
		{
		  // do not use dist() for distance check (avoid sqare root)
		  if( sqrnorm( otherTeam[j].pos - target ) < sqrnorm( otherTeam[validEntryIndex].pos - target ) )
		    validEntryIndex = j;
		}
	      else
		{
		  validEntryFound = true;
		  validEntryIndex = j;
		}
	    }
	}

      if( validEntryFound )
	sorted[validEntryIndex] = true;
      else continue;
    }

  if( validEntryFound )
    return validEntryIndex;   // return player number

  return UNKNOWN;   // not enough valid players found
}


int
Worldmodel::ourTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			       const Angle& alpha, long maxAge ) const
{
  int numPlayers = 0;

  for( int i = 1; i <= 11; i++ )
    {
      if( ourTeam[i].age > maxAge )
	continue;

      if( ourTeam[i].inCorridor( source, target, d, alpha ) )
	numPlayers++;
    }

  return numPlayers;
}


int
Worldmodel::otherTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
				 const Angle& alpha, long maxAge ) const
{
  int numPlayers = 0;

  for( int i = 1; i <= 11; i++ )
    {
      if( otherTeam[i].age > maxAge )
	continue;

      if( otherTeam[i].inCorridor( source, target, d, alpha ) )
	numPlayers++;
    }

  return numPlayers;
}


double
Worldmodel::ourOffsideLine( long maxAge ) const
{
  double   offX = 0.0;          // x-value for offside line
  double   maxX = 0.0;          // maximum x-value of opponent team (closest to opponent goal)

  if ( playMode == PLAY_MODE_KICK_OFF_THEM ||
       playMode == PLAY_MODE_KICK_OFF_US ||
       playMode == PLAY_MODE_GOAL_THEM ||
       playMode == PLAY_MODE_GOAL_US ||
       playMode == PLAY_MODE_BEFORE_KICK_OFF )
    return 0;

  // this is soccerserver code: it does not know any goalie...

  for( int i = 1; i <= 11; i++ )
    {
      if( otherTeam[i].age <= maxAge )
	{
	  double playerX = otherTeam[i].pos.getX();
	  // we are careful: players could have moved away from their goal
	  playerX -= ( min( otherTeam[i].age, min( 2, maxAge / 2 ) ) * otherTeam[i].vmax );

	  if( playerX > offX )      // If a player is beyond the offside-line...
	    {
	      offX = playerX;       // ...take his x-coordinate for the offside value
	      if( offX > maxX )     // But if he is even beyond the max-line...
		{
		  double tmpX = maxX;   // ...swap offX and maxX
		  maxX = offX;                  
		  offX = tmpX;
		}
	    }

	}
    }

  if( ball.pos.getX() > offX )
    offX = ball.pos.getX();

  return offX;
}


double
Worldmodel::otherOffsideLine( long maxAge ) const
{
  double   offX = 0.0;          // x-value for offside line
  double   minX = 0.0;          // minimum x-value of our team (closest to our goal)

  if( ( playMode == PLAY_MODE_KICK_IN_US ) ||
      ( playMode == PLAY_MODE_FREE_KICK_US ) )
    return min( 0., max( -FIELD_LENGTH/2, ball.pos.getX() - 10 ) );
  
  if( ( playMode == PLAY_MODE_CORNER_KICK_US ) ||
      ( playMode == PLAY_MODE_GOAL_KICK_THEM ) )
    return 0;

  if( playMode == PLAY_MODE_GOAL_KICK_US ) 
    return -FIELD_LENGTH/2 + PENALTY_AREA_LENGTH;

  if( ( playMode == PLAY_MODE_KICK_IN_THEM ) ||
      ( playMode == PLAY_MODE_FREE_KICK_THEM ) )
    return min( 0., max( -FIELD_LENGTH/2, ball.pos.getX() - 20 ) );
  
  if( playMode == PLAY_MODE_CORNER_KICK_THEM )
    return 0;

  // this is soccerserver code: it does not know any goalie...

  for( int i = 1; i <= 11; i++ )
    {
      if( ourTeam[i].age <= maxAge )
	{
	  double playerX = ourTeam[i].pos.getX();

	  if( playerX < offX )      // If a player is beyond the offside-line...
	    {
	      offX = playerX;       // ...take his x-coordinate for the offside value
	      if( offX < minX )     // But if he is even beyond the min-line...
		{
		  double tmpX = minX;   // ...swap offX and minX
		  minX = offX;                  
		  offX = tmpX;
		}
	    }

	}
    }
  
  if( ball.pos.getX() < offX )
    offX = ball.pos.getX();

  return offX;
}

  
void
Worldmodel::encodeAudio( MessageType msgType, char* msg, WBall comball )
{
  long c = long( msgType );   // encoded in Noda bytes 0,1,2
  long d = 0;         // encoded in Noda bytes 3,4,5

  // currently there is space for 10 values (for msgType) in the first 3 Noda bytes
  // 74^3 == 405224, 405224 / ( 180 * 28 * 8 ) == 10.0502

  if( msgType != MSGTYPE_SIM_ACTION )
    comball = ball;

  switch( msgType )
    {
    case MSGTYPE_POSITION :

      // there is some space left to say who is the intended receiver of this message

      // player's stamina(0..4000)
      c += 10 * max( 0, min( 4000, int( my.stamina / STAMINA_MAX * 4000 ) ) );

      // currently there is space for 3 values in the second 3 Noda bytes
      // 74^3 == 405224, 405224 / ( 421 * 273 ) == 3.5257

      // player's position (0..420, 0..272)
      d += max( 0, min( 420, int( 4.0 * ( my.pos.getX() + 0.5 * FIELD_LENGTH ) + 0.5 ) ) );
      d += 421 * max( 0, min( 272, int( 4.0 * ( my.pos.getY() + 0.5 * FIELD_WIDTH ) + 0.5 ) ) );
      break;

    case MSGTYPE_SIM_ACTION :
    case MSGTYPE_ACTION :
    case MSGTYPE_BALL :
    case MSGTYPE_TACKLE :
    default :

      // ball velocity (radius 0..179, length 0..27)
      c += 10 * max( 0, min( 179, int( ( RAD_TO_DEG( comball.v.getAngle() ) + 180 ) * 0.5 + 0.5 ) ) );
      c += 10 * 180 * max( 0, min( 27, int( comball.v.getLength() * 10.0 + 0.5 ) ) );

      // ball age (0..3, 3 stands for old data) and state
      long bc = min( 3, comball.age );
      // decodeAudio() only distinguishes between FUZZY and not FUZZY
      if( comball.state != VISUAL )
	bc += 4;

      c += 10 * 180 * 28 * bc;

      // currently there is space for 3 values in the second 3 Noda bytes
      // 74^3 == 405224, 405224 / ( 421 * 273 ) == 3.5257

      // ball position (0..420, 0..272)
      d += max( 0, min( 420, int( 4.0 * ( comball.pos.getX() + 0.5 * FIELD_LENGTH ) + 0.5 ) ) );
      d += 421 * max( 0, min( 272, int( 4.0 * ( comball.pos.getY() + 0.5 * FIELD_WIDTH ) + 0.5 ) ) );

      // player's distance to middle of field (for determination of position)
      // here information from passmodule is overwritten
      // [C] 64 is a little more than sqrt( sqr( 0.5 * FIELD_WIDTH ) + sqr( 0.5 * FIELD_LENGTH ) )
      // to allow players to leave the field for kick-ins
      msg[9] = msgEncodeChars[(int)char( max( 0, min( 73, int( my.pos.getLength() * 73.0 / 64.0 + 0.5 ) ) ) )];
      break;
    }

  // fill first three Noda bytes
  msg[0] = msgEncodeChars[(int)char( c % 74 )];
  msg[1] = msgEncodeChars[(int)char( c / 74 % 74 )];
  msg[2] = msgEncodeChars[(int)char( c / 74 / 74 % 74 )];

  // fill next three Noda bytes
  msg[3] = msgEncodeChars[(int)char( d % 74 )];
  msg[4] = msgEncodeChars[(int)char( d / 74 % 74 )];
  msg[5] = msgEncodeChars[(int)char( d / 74 / 74 % 74 )];
}


bool
Worldmodel::decodeAudio( char* msg, int sender, double senderDir )
{
  long c = 0;   // needed for decoding
  long d = 0;
  long e = 0;
  long f = 0;

  // MCMD( MDLine(my.pos,radAng2Vector2d(40.0,senderDir)) );

  bool position = false;

  WBall audioBall;   // received ball data

  // get Noda bytes 0,1,2
  c = msgDecodeChars[(int)msg[0]];
  c += 74 * msgDecodeChars[(int)msg[1]];
  c += 74 * 74 * msgDecodeChars[(int)msg[2]];

  // get Noda bytes 3,4,5
  d = msgDecodeChars[(int)msg[3]];
  d += 74 * msgDecodeChars[(int)msg[4]];
  d += 74 * 74 * msgDecodeChars[(int)msg[5]];
      
  // get Noda bytes 6,7,8 for stamina
  e = msgDecodeChars[(int)msg[6]];
  e += 74 * msgDecodeChars[(int)msg[7]];
  e += 74 * 74 * msgDecodeChars[(int)msg[8]];

  // get Noda byte 9 for player position information (distance to zero)
  f = msgDecodeChars[(int)msg[9]];

  MessageType msgType = MessageType( c % 10 );
  c /= 10;

  DECISION( "received audiomsg from "<< sender << " type : "
	    << ( msgType == MSGTYPE_POSITION ?  "Position" :
		 msgType == MSGTYPE_SIM_ACTION ? "sim_Action" :
		 msgType == MSGTYPE_TACKLE ? "tackle":
		 msgType == MSGTYPE_ACTION ? "action":
		 msgType == MSGTYPE_BALL ? "ball": "unknown" ) );

  switch( msgType )
    {
    case MSGTYPE_POSITION :
      position = true;
      ourTeam[sender].stamina = c * STAMINA_MAX / 4000.0;

      // use player position information if we have not seen him in this cycle
      if( ourTeam[sender].age > 0 )
	{
	  ourTeam[sender].pos.setX( d % 421 / 4.0 - 0.5 * FIELD_LENGTH );
	  d /= 421;
	  ourTeam[sender].pos.setY( d / 4.0 - 0.5 * FIELD_WIDTH );
	  ourTeam[sender].age = 1;
	  ourTeam[sender].coneTime = 0;
	}
      break;

    case MSGTYPE_TACKLE :
      ourTeam[sender].cyclesSinceTackle = 0;
      // no break!

    case MSGTYPE_SIM_ACTION :
    case MSGTYPE_ACTION :
    case MSGTYPE_BALL :
    default :

      audioBall.v.setRadAng( c / 180 % 28 / 10.0, DEG_TO_RAD( c % 180 * 2 - 180 ) );
      c /= 28;
      c /= 180;

      audioBall.age = c % 4;
      audioBall.state = ( c / 4 == 1 ? FUZZY : UNSAFE );

      audioBall.pos.setX( d % 421 / 4.0 - 0.5 * FIELD_LENGTH );

      d /= 421;
      audioBall.pos.setY( d / 4.0 - 0.5 * FIELD_WIDTH );

      // update ball if audial information is essentially younger than my visual information
      // if audioBall.age == 3 it might be much older!!! (<- sserver 8+ communication)
      VALUE( "audioball : age: " << audioBall.age );
      if( msgType == MSGTYPE_SIM_ACTION && ball.age > 0 )
	{
	  VALUE( "using simulated audioball" );
	  ball.pos      = audioBall.pos;
	  ball.v        = audioBall.v;
	  ball.age      = 1;
	  ball.state    = audioBall.state;
	  ball.coneTime = 0;
	}
      else if( ( ball.coneTime > 0 || audioBall.age < ball.age  ) && audioBall.age < 3 )
	{
	  VALUE( "using audioball" );
	  ball.pos      = audioBall.pos;
	  ball.v        = audioBall.v;
	  ball.age      = audioBall.age + 1;
	  ball.state    = audioBall.state;
	  ball.coneTime = 0;
	  ball.simulate();   // information is from one cycle ago
	}

      ourTeam[sender].stamina = e / 9 / 12 / 12 / 9 % 17 / 16.0 * STAMINA_MAX;

      // determine player's position from distance to zero and senderDir
      // but only if we have not seen him in this cycle
      if( ourTeam[sender].age > 0 )
	{
	  double l1, l2;

	  // [C] for the 64 see encodeAudio()
	  if( intersectLineCircle( radAng2Vector2d( 1.0, senderDir + my.bodyDir ),
				   -my.pos, f / 73.0 * 64.0, l1, l2 ) == Valid )
	    {
	      if( l2 > 0 )
		{
		  bool usePos = true;
		  Vector2d playerPos = my.pos + l2 * radAng2Vector2d( 1.0, senderDir + my.bodyDir );

		  if( l1 > 0 )   // there may be two possible solutions
		    {
		      VALUE("two solutions while calculating position");
		      usePos = false;
		      Vector2d playerPos2 = my.pos + l1 * radAng2Vector2d( 1.0, senderDir + my.bodyDir );

		      // only use one of the positions if it is much closer to the simulated position [C]
		      if( dist( playerPos, ourTeam[sender].pos ) < 0.5 * dist( playerPos2, ourTeam[sender].pos ) )
			usePos = true;
		      else if( dist( playerPos2, ourTeam[sender].pos ) < 0.5 * dist( playerPos, ourTeam[sender].pos ) )
			{
			  usePos = true;
			  playerPos = playerPos2;
			}
		    }

		  if( usePos )
		    {
		      VALUE("using distance position");
		      ourTeam[sender].pos = playerPos;
		      ourTeam[sender].age = 1;
		      ourTeam[sender].coneTime = 0;
		    }
		}
	    }
	}
      break;
    }

  return position;
}


bool
Worldmodel::writeToStream( ostream &out ) const
{
  // writes worldmodel to stream in non-binary format. 
  
  if( !out )
    {
      cerr << "Worldmodel: output stream not open when trying to write" 
	   << endl;
      return false;   // error: stream not open
    }

  const char *BLANK = " ";

  // write all data to the file

  // write time and player data 
  out << time               << BLANK
      << (int)ourSide       << BLANK
      << my.pos.getX()      << BLANK
      << my.pos.getY()      << BLANK
      << my.number          << BLANK
      << (double)my.bodyDir << BLANK
      << (double)my.faceDir << BLANK
      << my.age             << BLANK
      << (int)viewWidth     << BLANK
      << (double)my.v.getX() << BLANK
      << (double)my.v.getY() << BLANK;

  // write data of our team
  for( int i = 1; i <= 11; i++ )
    out << ourTeam[i].pos.getX() << BLANK
	<< ourTeam[i].pos.getY() << BLANK
	<< ourTeam[i].age << BLANK
	<< (double)(ourTeam[i].bodyDir) << BLANK
	<< (double) ourTeam[i].v.getX() << BLANK
	<< (double) ourTeam[i].v.getY() << BLANK;

  // write data of other team
  for( int i = 1; i <= 11; i++ )
    out << otherTeam[i].pos.getX() << BLANK
	<< otherTeam[i].pos.getY() << BLANK
	<< otherTeam[i].age << BLANK
	<< (double)(otherTeam[i].bodyDir) << BLANK
	<< (double) otherTeam[i].v.getX() << BLANK
	<< (double) otherTeam[i].v.getY() << BLANK;

  // write data of the ball and the goals
  out << ball.pos.getX() << BLANK
      << ball.pos.getY() << BLANK
      << ball.age        << BLANK
      << otherGoal.age   << BLANK
      << ourGoal.age     << BLANK
      << ball.v.getX()   << BLANK
      << ball.v.getY()   << BLANK;

  // write sensebody data
  out << my.stamina       << BLANK
      << my.effort        << BLANK
      << my.headAngle     << BLANK
      << kickCount        << BLANK
      << dashCount        << BLANK
      << turnCount        << BLANK
      << sayCount         << BLANK
      << turnNeckCount    << BLANK
      << catchCount       << BLANK
      << moveCount        << BLANK
      << changeViewCount  << BLANK
      << tackleCount      << BLANK
      << pointToCount     << BLANK
      << attentionToCount << BLANK;

  if( !out ) 
    return false;   // some error
  else 
    return true;    // writing was ok
}


bool
Worldmodel::readFromStream( istream &in )
{
  // reads non-binary coded Worldmodel from stream. 

  if( !in )
    {
      cerr << "Worldmodel: input stream not open or faulty when trying to read"
	   << endl;
      return false; 
    }
  
  // variables used for reading complex data (Vector2d, ...)
  double x, y, angle;
  int theSide, theViewWidth;
  
  // read time and some body-data 
  in >> time           
     >> theSide
     >> x >> y
     >> my.number;
  ourSide = (playing_side)theSide;
  my.pos.setXY( x, y ); 
       
  in >> angle;
  my.bodyDir = Angle(angle);    
  in >> angle
     >> my.age
     >> theViewWidth;
  viewWidth= (view_width)theViewWidth;
  my.faceDir = Angle(angle);
  
  // read data of our team
  for( int i = 1; i <= 11; i++ )
    {
      in >> x >> y
	 >> ourTeam[i].age
	 >> angle;
      ourTeam[i].pos.setXY( x, y );
      ourTeam[i].bodyDir = Angle(angle);
      ourTeam[i].number  = i;
    } 
  // read data of other team
  for( int i = 1; i <= 11; i++ )
    {
      in >> x >> y  
	 >> otherTeam[i].age
	 >> angle;
      otherTeam[i].pos.setXY( x, y );
      otherTeam[i].bodyDir = Angle(angle);
      otherTeam[i].number  = i;
    }

  // read data of the ball and the goals
  in >> x >> y
     >> ball.age
     >> otherGoal.age
     >> ourGoal.age;
  ball.pos.setXY( x, y );      

  in >> x >> y;
  ball.v.setXY( x, y );
  
  // read sensebody data
  in >> my.stamina       
     >> my.effort
     >> angle
     >> kickCount     
     >> dashCount     
     >> turnCount     
     >> sayCount      
     >> turnNeckCount
     >> catchCount
     >> moveCount
     >> changeViewCount
     >> tackleCount
     >> pointToCount
     >> attentionToCount;

  my.headAngle = (Angle)angle;     

  if( !in ) 
    return false;   // some weird error appeared while reading from stream
  else 
    return true;    // reading was ok
}


void
Worldmodel::simCycle()
{
  double   turnMoment;
  double   actualPower;   // needed for simulation of a dash command

  // step 1 : save old ball relpos (for later calculation of ball velocity) and player

  oldBallRelpos = ball.pos - my.pos;
  oldMy = my;

  // step 2 : if player has desired an action, simulate it now

  switch( action.command )
    {
    case DASH_COMMAND: 
      // no dash after tackle
      if( my.cyclesSinceTackle > TACKLE_CYCLES )
	{
	  actualPower = action.power;

	  if( actualPower < 0 )   // dashing backward costs double stamina
	    {
	      actualPower *= -2.0;
	      // "will-power" is new in soccerserver 7.0
	      if( my.stamina + my.extraStamina < actualPower )
		actualPower = my.stamina + my.extraStamina;
	      actualPower /= -2.0;
	    }
	  else
	    if( my.stamina + my.extraStamina < actualPower )
	      actualPower = my.stamina + my.extraStamina;

	  actualPower *= ( my.effort * my.dashPowerRate );

	  my.acc( radAng2Vector2d( actualPower, my.bodyDir ) );

	  // decrease stamina

	  if( action.power < 0 )
	    my.stamina += 2.0 * action.power;
	  else
	    my.stamina -= action.power;
	  if( my.stamina < 0.0 )
	    my.stamina = 0.0;

	  // move ball if goalie dashes after a catch command
	  // caution: this does not check if goalie really sent a catch command!
	  if( my.isGoalie == GOALIE_TRUE && playMode == PLAY_MODE_FREE_KICK_US )
	    if( my.ballKickable( ball ) )
	      ball.pos += my.v;
	}
      break;

    case TURN_COMMAND:
      // no turn after tackle
      if( my.cyclesSinceTackle > TACKLE_CYCLES )
	{
	  turnMoment = action.moment;
	  if( turnMoment > MAXMOMENT / 180.0 * M_PI ) turnMoment = MAXMOMENT / 180.0 * M_PI;
	  if( turnMoment < MINMOMENT / 180.0 * M_PI ) turnMoment = MINMOMENT / 180.0 * M_PI;
	  my.bodyDir += turnMoment / my.inertiaMoment();
	  my.faceDir += turnMoment / my.inertiaMoment();
	}
      break;
  
    case MOVE_COMMAND:
      // check if playMode allows a move command
      switch( playMode )
	{
	case PLAY_MODE_FREE_KICK_US:
	  if( my.isGoalie != GOALIE_TRUE )
	    break;
	  // caution: this does not check if goalie sent a catch command!
	  if( my.ballKickable( ball ) )
	    ball.pos = action.pos + ball.pos - my.pos;
	case PLAY_MODE_BEFORE_KICK_OFF:
	case PLAY_MODE_GOAL_US:
	case PLAY_MODE_GOAL_THEM:
	  my.pos = action.pos;
	  my.v.setXY( 0.0, 0.0 );
	  break;
	default:
	  break;
	}
      
    case KICK_COMMAND:
      // no kick after tackle
      if( my.cyclesSinceTackle > TACKLE_CYCLES )
	{
	  // is the ball within kick radius?
	  if( my.ballKickable( ball ) )
	    ball.acc( radAng2Vector2d( action.power * my.kickPowerRate( ball ),
				       my.bodyDir + Angle( action.moment ) ) );	       
	}
      break;
	  
    case CATCH_COMMAND:
      // no catch after tackle
      if( my.cyclesSinceTackle > TACKLE_CYCLES )
	{
	  // check if player is goalie, if player is allowed to catch again and if ball is in catchable area
	  if( my.isGoalie == GOALIE_TRUE ) 
	    if( cyclesSinceCatch > CATCH_BAN_CYCLES )
	      {
		cyclesSinceCatch = 0;
		if( my.ballCatchable( ball, action.moment ) )
		  {
		    ball.v.setXY( 0.0, 0.0 );
		    ball.pos = my.pos + my.v +
		      radAng2Vector2d( my.size + BALL_SIZE, ( ball.pos - my.pos ).getAngle() );
		  }
	      }
	}
      break;

    case TACKLE_COMMAND:
      // no second tackle after tackle
      if( my.cyclesSinceTackle > TACKLE_CYCLES )
	my.cyclesSinceTackle = 0;
      break;

      /*
	case POINTTO_COMMAND:
	if( cyclesSincePointTo > POINTTO_BAN_CYCLES )
	  cyclesSincePointTo = 0;
	break;
      */

    default:
      break;
    }

  // set action invalid

  action.command = NO_COMMAND;
  action.moment = 0.0;
  action.power  = 0.0;
  action.pos.setXY( 0.0, 0.0 );

  // step 3 : simulate movement of objects

  my.simulate();                   // simulate my movement  
  ball.simulate();                 // simulate ball movement
  for( int i = 1; i <= 11; i++ )   // simulate other players' movement
    {
      ourTeam[i].simulate();
      otherTeam[i].simulate();
    }

  // step 4 : check for collision with ball (we do NOT check player<->player collisions)

  collState = NONE;

  // soccerserver 7.09+ algorithm

  if( my.number > 0 )   // coach should not collide...
    {
      if( my.distTo( ball ) < my.size + BALL_SIZE )
	{
	  Vector2d middle = ( my.pos + ball.pos ) * 0.5;
	  Vector2d myOffset;
	  Vector2d ballOffset;

	  DECISION( "simCycle: checking for player<->ball collision" );

	  if( my.pos == ball.pos )
	    {
	      // Here soccerserver works with a random angle

	      // new player position
	      myOffset.setRadAng( ( my.size + BALL_SIZE ) * 0.5, 0.0 );

	      // new ball position
	      ballOffset.setRadAng( ( my.size + BALL_SIZE ) * 0.5, PI );

	      collState = CRITICAL;
	      VALUE( "simCycle: maybe player<->ball collision (CRITICAL)" );
	    }
	  else
	    {
	      // new player position
	      myOffset.setRadAng( ( my.size + BALL_SIZE ) * 0.5, ( my.pos - middle ).getAngle() );

	      // new ball position
	      ballOffset.setRadAng( ( my.size + BALL_SIZE ) * 0.5, ( ball.pos - middle ).getAngle() );

	      collState = PROCESSED;
	      VALUE( "simCycle: player<->ball collision (PROCESSED)" );
	    }

	  // This is soccerserver implementation. It does not make sense since the ball
	  // can also be kicked around in play mode GOAL_KICK_US/THEM!
	  if( playMode != PLAY_MODE_PLAY_ON )
	    {
	      my.pos = middle + myOffset - ballOffset;

	      // this is not implemented in soccerserver 7.10 but it should be...
	      while( my.distTo( ball ) < my.size + BALL_SIZE )
		{
		  myOffset *= 1.0001;
		  ballOffset *= 1.0001;
		  my.pos = middle + myOffset - ballOffset;
		  VALUE( "simCycle: adjusting distance after player<->ball collision" );
		}
	    }
	  else
	    {
	      my.pos = middle + myOffset;
	      ball.pos = middle + ballOffset;

	      // this IS implemented in soccerserver 7.10...
	      while( my.distTo( ball ) < my.size + BALL_SIZE )
		{
		  myOffset *= 1.0001;
		  my.pos = middle + myOffset;

		  ballOffset *= 1.0001;
		  ball.pos = middle + ballOffset;
		  VALUE( "simCycle: adjusting distance after player<->ball collision" );
		}
	    }

	  // set new player and ball velocity
	  my.v *= -0.1;
	  ball.v *= -0.1;
	}
    }

  // step 5 : increase stamina (considering effort and recovery parameters)

  if( my.number > 0 )   // coach has no stamina
    {
      if( my.stamina <= RECOVER_DEC_THR * STAMINA_MAX )
	if( my.recovery > RECOVER_MIN )
	  my.recovery -= RECOVER_DEC ;

      if( my.stamina <= EFFORT_DEC_THR * STAMINA_MAX )
	if( my.effort > my.effortMin )
	  {
	    my.effort -= EFFORT_DEC ;
	    if( my.effort < my.effortMin )
	      my.effort = my.effortMin;
	  }

      if( my.stamina >= EFFORT_INC_THR * STAMINA_MAX )
	if( my.effort < my.effortMax )
	  {
	    my.effort += EFFORT_INC ;
	    if( my.effort > my.effortMax )
	      my.effort = my.effortMax;
	  }

      my.stamina += my.recovery * my.staminaIncMax;
      if( my.stamina > STAMINA_MAX )
	my.stamina = STAMINA_MAX;
    }

  // simulate stamina for other players (always under worst case assumption)

  for( int i = 1; i <= 11; i++ )
    {
      ourTeam[i].stamina = max( EFFORT_DEC_THR * STAMINA_MAX,
				min( STAMINA_MAX, ourTeam[i].stamina + ourTeam[i].staminaIncMax - MAXPOWER ) );
				
      otherTeam[i].stamina = min( STAMINA_MAX, otherTeam[i].stamina + otherTeam[i].staminaIncMax );
    }

  // step 6 : Birthday party for all objects

  if( my.age < ETERNAL ) my.age++;
  if( my.cyclesSinceTackle < ETERNAL )
    my.cyclesSinceTackle++;

  if( ball.age < ETERNAL ) ball.age++;
  if( otherGoal.age < ETERNAL ) otherGoal.age++;
  if( ourGoal.age < ETERNAL ) ourGoal.age++;

  for( int i = 1; i <= 11; i++ )
    {
      if( ourTeam[i].age < ETERNAL ) ourTeam[i].age++;
      if( ourTeam[i].bodyAge < ETERNAL ) ourTeam[i].bodyAge++;
      if( otherTeam[i].age < ETERNAL ) otherTeam[i].age++;
      if( otherTeam[i].bodyAge < ETERNAL ) otherTeam[i].bodyAge++;
      if( ourTeam[i].cyclesSinceTackle < ETERNAL )
	ourTeam[i].cyclesSinceTackle++;
      if( otherTeam[i].cyclesSinceTackle < ETERNAL )
	otherTeam[i].cyclesSinceTackle++;
    }

  for( int i = 0; i < VISUALSTEPS; i++ )
    {
      if( vAge[i] < ETERNAL )
	vAge[i]++;
      if( aOppAge[i] < ETERNAL )
	aOppAge[i]++;
      if( aOurAge[i] < ETERNAL )
	aOurAge[i]++;
    }

  if( visualAge < ETERNAL )
    visualAge++;
  if( cyclesSinceCatch < ETERNAL )
    cyclesSinceCatch++;
  if( cyclesSincePointTo < ETERNAL )
    cyclesSincePointTo++;

  // finally copy player into team array
  if( my.number > 0 )
    ourTeam[my.number] = my;
}


SensorData
Worldmodel::adjust()
{
  SensorData    data;                    // new data received from soccerserver
  int           type;                    // submitted new player type
  WPlayerType   newType;                 // for a player change

  CycleAction   newAction;               // player's latest action
  CycleAction   lostAction;              // player's evt. lost action

  SensorData    audioData;               // received audiomessage from teammate

  audioData.type = SENSOR_TYPE_ERROR;

  // get agent's latest action (to insert it at the right time)

  comm->getLastAction( newAction );
  comm->getLostAction( lostAction );

  // as long as there is data in the buffer - get it

  while( comm->getNextData( data ) )
    {
      switch( data.type ) 
	{

	case SENSOR_TYPE_HEAR :   // should be a message from another player
 
	  // only keep messages from team members
	  if( data.sender == SENDER_PLAYER_US )
	    {
	      if( data.sender_num >= 1 && data.sender_num <= 11 )
		if( data.sender_dir != INVALID_DIR )
		  if( data.message[0] != 0 )
		    {
		      audioData = data;
		      if( !decodeAudio( data.message, data.sender_num, data.sender_dir ) )
			{ // [MW] take direction only in account if no position was sent
			  aOurCounter = ( aOurCounter + 1 ) % VISUALSTEPS;
			  aOurDir[aOurCounter] = data.sender_dir + my.bodyDir;
			  aOurAge[aOurCounter] = 0;
			}

		      // Comm::sendMCmd(LINE(my.pos,my.pos+radAng2Vector2d(20,data.sender_dir+my.bodyDir)));
		    }
	    }
	  else if( data.sender == SENDER_PLAYER_THEM && data.sender_dir != INVALID_DIR )
	    {
	      // [MW] remeber direction of the opponent
	      aOppCounter = ( aOppCounter + 1 ) % VISUALSTEPS;
	      aOppDir[aOppCounter] = data.sender_dir + my.bodyDir;
	      aOppAge[aOppCounter] = 0;
	      //Comm::sendMCmd(LINE(my.pos,my.pos+radAng2Vector2d(30,data.sender_dir+my.bodyDir)));
	    }
	  else if( data.sender == SENDER_REFEREE )
	  
            {
	      if (penaltySide == SIDE_UNKNOWN)
		  penaltySide = data.penaltySide;
            }
	  
	  break;

	case SENSOR_TYPE_PLAYER_TYPE :

	  type = data.playerType.player_type;
	  if( type >= 0 && type < playerParams.player_types )
	    {
	      playerType[type].state          = type;

	      playerType[type].playerSpeedMax = data.playerType.player_speed_max;
	      playerType[type].playerDecay    = data.playerType.player_decay;
	      playerType[type].playerSize     = data.playerType.player_size;
	      playerType[type].staminaIncMax  = data.playerType.stamina_inc_max;
	      playerType[type].inertiaFactor  = data.playerType.inertia_moment;
	      playerType[type].dashPowerRate  = data.playerType.dash_power_rate;
	      playerType[type].kickableMargin = data.playerType.kickable_margin;
	      playerType[type].kickRand       = data.playerType.kick_rand;
	      playerType[type].extraStamina   = data.playerType.extra_stamina;
	      playerType[type].effortMin      = data.playerType.effort_min;
	      playerType[type].effortMax      = data.playerType.effort_max;

	      if( playerType[type].dashPowerRate * MAXPOWER / ( 1.0 - playerType[type].playerDecay )
		  < playerType[type].playerSpeedMax )
		playerType[type].playerSpeedMax = playerType[type].dashPowerRate * MAXPOWER /
		  ( 1.0 - playerType[type].playerDecay );
	    }
	  break;

	case SENSOR_TYPE_PLAYER_CHANGED :

	  if( data.heteroType == UNKNOWN )
	    newType = playerType[DEFAULT_TYPE];
	  else
	    newType = playerType[data.heteroType];

	  if( data.ourTeamChanged )
	    {
	      int numSubs = numOurSubs;

	      if( newType.state == UNKNOWN )
		cerr << "Nr. " << my.number
		     << " : Player " << data.number << " was changed to type " 
		     << data.heteroType << ", but I don't know that type!" << endl;

	      ourTeam[data.number].type           = data.heteroType;
	      ourTeam[data.number].vmax           = newType.playerSpeedMax;
	      ourTeam[data.number].decay          = newType.playerDecay;
	      ourTeam[data.number].size           = newType.playerSize;
	      ourTeam[data.number].staminaIncMax  = newType.staminaIncMax;
	      ourTeam[data.number].inertiaFactor  = newType.inertiaFactor;
	      ourTeam[data.number].dashPowerRate  = newType.dashPowerRate;
	      ourTeam[data.number].kickableMargin = newType.kickableMargin;
	      ourTeam[data.number].kickRand       = newType.kickRand;
	      ourTeam[data.number].extraStamina   = newType.extraStamina;
	      ourTeam[data.number].effortMin      = newType.effortMin;
	      ourTeam[data.number].effortMax      = newType.effortMax;

	      if( ourTeam[data.number].dashPowerRate * MAXPOWER / ( 1.0 - ourTeam[data.number].decay ) <
		  ourTeam[data.number].vmax )
		ourTeam[data.number].vmax = ourTeam[data.number].dashPowerRate * MAXPOWER /
		  ( 1.0 - ourTeam[data.number].decay );

	      if( my.number == data.number )
		{
		  my.type           = data.heteroType;
		  my.vmax           = newType.playerSpeedMax;
		  my.decay          = newType.playerDecay;
		  my.size           = newType.playerSize;
		  my.staminaIncMax  = newType.staminaIncMax;
		  my.inertiaFactor  = newType.inertiaFactor;
		  my.dashPowerRate  = newType.dashPowerRate;
		  my.kickableMargin = newType.kickableMargin;
		  my.kickRand       = newType.kickRand;
		  my.extraStamina   = newType.extraStamina;
		  my.effortMin      = newType.effortMin;
		  my.effortMax      = newType.effortMax;

		  if( my.dashPowerRate * MAXPOWER / ( 1.0 - my.decay ) < my.vmax )
		    my.vmax = my.dashPowerRate * MAXPOWER / ( 1.0 - my.decay );
		}

	      numOurSubs++;
	      ourSubs[numSubs % ( playerParams.subs_max + 11 )].number = data.number;
	      ourSubs[numSubs % ( playerParams.subs_max + 11 )].time   = data.time;
	    }
	  else
	    {
	      int numSubs = numOtherSubs;

	      otherTeam[data.number].type           = data.heteroType;
	      otherTeam[data.number].vmax           = newType.playerSpeedMax;
	      otherTeam[data.number].decay          = newType.playerDecay;
	      otherTeam[data.number].size           = newType.playerSize;
	      otherTeam[data.number].staminaIncMax  = newType.staminaIncMax;
	      otherTeam[data.number].inertiaFactor  = newType.inertiaFactor;
	      otherTeam[data.number].dashPowerRate  = newType.dashPowerRate;
	      otherTeam[data.number].kickableMargin = newType.kickableMargin;
	      otherTeam[data.number].kickRand       = newType.kickRand;
	      otherTeam[data.number].extraStamina   = newType.extraStamina;
	      otherTeam[data.number].effortMin      = newType.effortMin;
	      otherTeam[data.number].effortMax      = newType.effortMax;

	      if( otherTeam[data.number].dashPowerRate * MAXPOWER / ( 1.0 - otherTeam[data.number].decay ) <
		  otherTeam[data.number].vmax )
		otherTeam[data.number].vmax = otherTeam[data.number].dashPowerRate * MAXPOWER /
		  ( 1.0 - otherTeam[data.number].decay );

	      numOtherSubs++;
	      otherSubs[numSubs % ( playerParams.subs_max + 11 )].number = data.number;
	      otherSubs[numSubs % ( playerParams.subs_max + 11 )].time   = data.time;
	    }
	  break;

	case SENSOR_TYPE_SEE:

	  // only care about data for current cycle
	  if( data.time == time && !waitForSenseBody )
	    {
	      checkViewWidth( data );
	      calcMyPos( data );
	      adjustBall( data );
	      adjustPlayers( data );
	      adjustGoals( data );

	      // this has to be done after adjustPlayers() (stamina simulation)
	      if( playMode == PLAY_MODE_HALF_TIME ||
		  playMode == PLAY_MODE_TIME_EXTENDED ||
		  playMode == PLAY_MODE_BEFORE_KICK_OFF )
		{
		  for( int j = 1; j <= 11; j++ )
		    {
		      ourTeam[j].stamina    = STAMINA_MAX;
		      ourTeam[j].effort     = 1.0;
		      ourTeam[j].recovery   = 1.0;
		      otherTeam[j].stamina  = STAMINA_MAX;
		      otherTeam[j].effort   = 1.0;
		      otherTeam[j].recovery = 1.0;
		    }
		}

	      if( comm->amIGoalie() && ourGoalieNo == UNKNOWN )
		adjustTacticalNumbers( my.number );

	      // try to keep up with headAngle changes
	      // this works because we always process senseBody messages first
	      my.bodyDir = my.faceDir - my.headAngle;

	      // copy player into team array
	      ourTeam[my.number] = my;

	      collState = NONE;   // just forget simulation - we have got visual data!

	      // save faceDir to vDir array
	      vCounter = ( vCounter + 1 ) % VISUALSTEPS;
	      vDir[vCounter] = my.faceDir;
	      vAngle[vCounter] = getViewAngle( data.see_view_width );
	      vAge[vCounter] = 0;

	      // do not accept further visual info before next senseBody message
	      waitForSenseBody = true;
	      visualAge = 0;
	    }
	  break;

	case SENSOR_TYPE_SENSE_BODY:

	  // only care about new data
	  if( data.time >= time )
	    {
	      waitForSenseBody = false;
	      // simulate world until data.time, then update with newer information
	      if( data.time == time )
		{
		  if( time == newAction.time )
		    action.store( newAction );
		  simCycle();
		}
	      else
		{
		  while( time < min( data.time, newAction.time ) )
		    {
		      simCycle();
		      time++;
		    }
		  if( time == newAction.time )
		    action.store( newAction );
		  while( time < data.time )
		    {
		      simCycle();
		      time++;
		    }
		}

	      playMode   = comm->getPlayMode();
	      ourSide    = comm->getMySide();
	      ourScore   = comm->getOurScore();
	      otherScore = comm->getOtherScore();

	      /* if we are doing a penalty shoot out*/
	      if (penaltySide != UNKNOWN) {

		/* if we are attacking */
		if (playMode == PLAY_MODE_PENALTY_SETUP_US  ||
		    playMode == PLAY_MODE_PENALTY_MISS_THEM  ||
		    playMode == PLAY_MODE_PENALTY_SCORE_THEM ||
		    playMode == PLAY_MODE_PENALTY_TAKEN_US)
		  {
		    ourSide = (penaltySide == SIDE_LEFT) ?
		      SIDE_RIGHT : SIDE_LEFT;
		  }

		/* if they are attacking */
		if (playMode == PLAY_MODE_PENALTY_SETUP_THEM  ||
		    playMode == PLAY_MODE_PENALTY_MISS_US     ||
		    playMode == PLAY_MODE_PENALTY_SCORE_US    ||
		    playMode == PLAY_MODE_PENALTY_TAKEN_THEM)
		  {
		    ourSide = penaltySide;
		  }
	      }

	      if( comm->amIGoalie() && ourGoalieNo == UNKNOWN )
		adjustTacticalNumbers( my.number );

	      // update ball (only does something in standard situations in this case!)
	      adjustBall( data );

	      my.stamina = data.stamina;
	      my.effort = data.effort;

	      kickCount        = data.kick_count;
	      dashCount        = data.dash_count;
	      turnCount        = data.turn_count;
	      sayCount         = data.say_count;
	      turnNeckCount    = data.turn_neck_count;

	      catchCount       = data.catch_count;      // new in soccerserver 7
	      moveCount        = data.move_count;
	      changeViewCount  = data.change_view_count;

	      tackleCount      = data.tackle_count;     // new in soccerserver 8
	      pointToCount     = data.point_to_count;
	      attentionToCount = data.attention_to_count;

	      viewWidth        = data.my_view_width;
	      viewQuality      = data.my_view_quality;

	      my.headAngle     = data.head_angle;
	      my.faceDir       = my.bodyDir + my.headAngle;

	      // update player velocity (caution: velocity is relative!)
	      // (maybe the velocity vector could also be helpful to detect and handle collisions!)
	      my.v = radAng2Vector2d( data.speed, Angle( data.speed_angle ) + my.faceDir );

	      // update simulated position with felt velocity
	      if( collState == NONE )
		my.pos = 1.0 / my.decay * my.v + oldMy.pos;

	      // copy player into team array
	      ourTeam[my.number] = my;
	    }
	  break;
	  
	case SENSOR_TYPE_EXACT:   // for the coach
    
	  // only care about newer data
	  if( data.time >= time )
	    {
	      // simulate world until data.time, then overwrite with exact information
	      if( data.time == time )
		{
		  simCycle();
		}
	      else while( time < data.time )
		{
		  simCycle();
		  time++;
		}

	      playMode   = comm->getPlayMode();
	      ourSide    = comm->getMySide();
	      ourScore   = comm->getOurScore();
	      otherScore = comm->getOtherScore();

	      // update ball (only does something in standard situations in this case!)
	      adjustBall( data );

	      adjustExact( data );

	      visualAge = 0;
	    }
	  break;

	default:
	  break;
	}
    }

  return audioData;
}


bool
Worldmodel::feedbackAudio( AudioMsg &audioMsg )
{
  // check for coach messages (encoded as before)
  if( !audioMsg.isCoachMsg() )
    return false;

  bool ok;
  audioMsg >> ok;

  if( !ok )
    return false;

  // don't use old messages
  // CAUTION: during standard situations time does not change
  // so the message might be old even if its time is the current time!
  // Maybe coach should also send the number of steps that had passed
  // since the beginning of the standard situation when he sent the
  // message? But then players would also have to count this...
  if( long( audioMsg.getServerTime() ) < time - 1 )
    return false;

  NodaChar nc;
  double d;

  audioMsg >> nc;
  int audioOurGoalieNo = int( nc ) - 1;
  // set goalie flags and tactical numbers if goalie is seen for the first time
  if( audioOurGoalieNo != UNKNOWN && ourGoalieNo == UNKNOWN )
    adjustTacticalNumbers( audioOurGoalieNo );

  audioMsg >> nc;
  int audioOtherGoalieNo = int( nc ) - 1;
  if( audioOtherGoalieNo != UNKNOWN && otherGoalieNo == UNKNOWN )
    {
      otherGoalieNo = audioOtherGoalieNo;
      for( int i = 1; i <= 11; i++ )
	otherTeam[i].isGoalie = GOALIE_FALSE;
      otherTeam[otherGoalieNo].isGoalie = GOALIE_TRUE;
    }

  audioMsg >> d;
  ball.pos.setX( d );
  audioMsg >> d;
  ball.pos.setY( d );
  ball.age = 0;
  ball.coneTime = 0;
  ball.state = VISUAL;
  ball.v.setXY( 0.0, 0.0 );

  for( int i = 1; i <= 11; i++ )
    {
      audioMsg >> d;
      ourTeam[i].pos.setX( d );
      audioMsg >> d;
      ourTeam[i].pos.setY( d );
      ourTeam[i].age = 0;
      ourTeam[i].coneTime = 0;
      ourTeam[i].state = VISUAL;

      audioMsg >> d;
      otherTeam[i].pos.setX( d );
      audioMsg >> d;
      otherTeam[i].pos.setY( d );
      otherTeam[i].age = 0;
      otherTeam[i].coneTime = 0;
      otherTeam[i].state = VISUAL;
    }

  ourTeam[my.number] = my;

  // from tactical module: opponent player types
  for( int i = 1; i <= 11; i++ )
    {
      audioMsg >> nc;
      otherTeam[i].type = int( nc ) - 1;
    }

  return true;
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Operators for giving worldmodel (from coach) to other players using an audio msg
//
///////////////////////////////////////////////////////////////////////////////////////////////


class AudioMsg& operator << ( class AudioMsg& aum, const Worldmodel& w )
{
  // only used for coach messages in server 8+
  if( aum.isCoachMsg() )
    {
      aum << NodaChar( w.ourGoalieNo + 1 );
      aum << NodaChar( w.otherGoalieNo + 1 );

      aum << w.ball.pos.getX() << w.ball.pos.getY();
      for( int i = 1; i <= 11; i++ )
	aum << w.ourTeam[i].pos.getX() << w.ourTeam[i].pos.getY()
	    << w.otherTeam[i].pos.getX() << w.otherTeam[i].pos.getY();
    }

  return aum;
}


ostream & operator << ( ostream &os, Worldmodel& w )
{
  int i;

  os << "time       : " << w.time << endl
     << "ourSide    : " << ( w.ourSide == SIDE_LEFT ? "LEFT" : w.ourSide == SIDE_RIGHT ? "RIGHT" : "UNKNOWN" ) << endl
     << endl
     << "my.pos       : " << w.my.pos << endl 
     << "my.bodyDir   : " << w.my.bodyDir << endl
     << "my.faceDir   : " << w.my.faceDir << endl
     << "my.headAngle : " << w.my.headAngle << endl
     << "my.stamina   : " << w.my.stamina << endl
     << "my.effort    : " << w.my.effort << endl
     << "my.age       : " << w.my.age << endl << endl;
  
  for( i = 1; i <= 11; i++ )
    {
      os << "ourTeam(" << i << ") : pos : " << w.ourTeam[i].pos
	 << ", age : " << w.ourTeam[i].age 
	 << endl << endl;
    }
  for( i = 1; i <= 11; i++ )
    {
      os << "otherTeam(" << i << ") : pos : " << w.otherTeam[i].pos
	 << ", age : " << w.otherTeam[i].age
	 << endl << endl;
    }

  os << "ball : pos : " << w.ball.pos
     << ", age : " << w.ball.age << endl
     << "otherGoal.age : " << w.otherGoal.age << endl
     << "ourGoal.age   : " << w.ourGoal.age << endl << endl;
  
  os << "kickCount        : " << w.kickCount << endl
     << "dashCount        : " << w.dashCount << endl
     << "turnCount        : " << w.turnCount << endl
     << "sayCount         : " << w.sayCount << endl
     << "turnNeckCount    : " << w.turnNeckCount << endl
     << "catchCount       : " << w.catchCount << endl
     << "moveCount        : " << w.moveCount << endl
     << "changeViewCount  : " << w.changeViewCount << endl
     << "tackleCount      : " << w.tackleCount << endl
     << "pointToCount     : " << w.pointToCount << endl
     << "attentionToCount : " << w.attentionToCount << endl;

  os << "***************************************" << endl;
  
  return os;
}
