//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

using namespace std;


/////////////////////////////////////////////////////////////////////////
//
// ACT
//
/////////////////////////////////////////////////////////////////////////

bool 
BallHandlingModule::act()
{
  // main action method, called by master control to execute action previously evaluated

  acted = true;
  lastAction = action;
  lastActCycles = 0;

  switch( action )
    {
    case WAIT:
      // if player is immobile after tackling, do not do anything
      if( my.cyclesSinceTackle <= TACKLE_CYCLES )
	skills->noAction();
      else
	{
	  // turn perpetual to ball to maximize reaction possibilities
	  // choose that direction pointing into opponent goal
	  // since we want to catch the ball later in that direction if possible
	  Angle dir = ( ball.pos - my.pos ).getAngle();
	  dir += ( dir > 0 ? -PI_2 : PI_2 );
	  AdvSkills::TurnCommand turnParams = skills->turn( dir );
	  if( turnParams.state == AdvSkills::ROK || turnParams.state == AdvSkills::RANGLE )
	    skills->exec( turnParams );
	  else
	    skills->noAction();
	}
      break;

    case FIND_BALL:
      findBall();
      break;

    case GOALIE_INTERCEPT:
      updateTA();
      goalieIntercept();
      break;        

    case GOALIE_CATCH:
      updateTA();
      skills->exec( catchParams );
      break;

    case GOALIE_PLACE:
      updateTA();
      skills->exec( placeParams );
      break;

    case PLAYER_INTERCEPT:
      updateTA();
      playerIntercept();
      break;

    case PLAYER_FAST_DRIBBLE:
      updateTA();
      kickParams = skills->kick( fd_ballV, 1 );
      skills->exec( kickParams );
      break;

    case PLAYER_DRIBBLE:
      updateTA();
      skills->exec( dribbleParams );
      break;

    case PLAYER_PLACE:
      updateTA();
      // if the place command would not do much, turn around instead
      // to get a better view of the world and avoid deadlocks
      if( norm( ball.v ) < 0.1 * my.vmax && norm( placeParams.v ) < 0.1 * my.vmax &&
	  dist( ball.pos + ball.v, my.pos + my.v ) < my.kickRange() )
	{
	  AdvSkills::TurnCommand tcmd;

	  if( my.pos.getY() < -FIELD_WIDTH / 3.0 )
	    tcmd = skills->turn( M_PI_2 );
	  else if( my.pos.getY() < -FIELD_WIDTH / 6.0 )
	    tcmd = skills->turn( 0.25 * M_PI );
	  else if( my.pos.getY() < FIELD_WIDTH / 6.0 )
	    tcmd = skills->turn( 0.0 );
	  else if( my.pos.getY() < FIELD_WIDTH / 3.0 )
	    tcmd = skills->turn( -0.25 * M_PI );
	  else
	    tcmd = skills->turn( -M_PI_2 );

	  if( tcmd.state != AdvSkills::RFINISHED )
	    skills->exec( tcmd );
	  else
	    skills->exec( placeParams );
	}
      else
	{
	  skills->exec( placeParams );
	}
      break;

    case PLAYER_GET_PASS:
      getPass();
      break;

    default:
      skills->noAction();
      return false;
    }
  return true;
}


void
BallHandlingModule::updateTA()
{
  // if any player of our team except for me is near the
  // ball, be quiet. I do not use the reachTime but real
  // distance since reachTime is instable and I don't want
  // to communicate nonsense
  int clst = world->ourTeamClosestTo( ball.pos );
  if( clst != UNKNOWN && clst != my.number &&
      ( world->ourTeam[clst].distTo( ball ) < 2 * world->ourTeam[clst].vmax ) )
    return;

  if( grade.is_better_than( Grade::AVERAGE ) )
    tact->reportOurTacticalAction( TA_CONTROL, my.number );
}


void
BallHandlingModule::findBall()
{
  // turn continuously to find ball on field (quite simple algorithm)
  if( !lookingForBall )
    {   // we just started the search this timestep
      lookingForBall = true;
    
      // turn left or right to find ball ? (take last ball dir into account)
      // NOTE: These values should work for view width NARROW and NORMAL
      turnAngle = ( my.bodyAngleTo( ball ) > 0 ) ? DEG_TO_RAD( +40.0 ) : DEG_TO_RAD( -40.0 );
    }

  if( world->visualAge <= 1 )
    skills->basicTurn( turnAngle );
  else
    skills->noAction();
}


inline
class std::ostream & operator << ( class std::ostream &os, const AdvSkills::TurtleCommand tc )
{
  return os << "state: " << ( ( tc.state == AdvSkills::TFINISHED ) ? "TFINISHED" :
			      ( tc.state == AdvSkills::TONWAY ) ? "TONWAY" :
			      ( tc.state == AdvSkills::TGOINGBACKWARDS ) ? "TGOINGBA":
			      "other? "  )
	    << " command: " << ( ( tc.command == AdvSkills::TurtleCommand::DASH ) ? "DASH" :
				 ( tc.command == AdvSkills::TurtleCommand::TURN ) ? "TURN" : 
				 "other? " )
	    << " time: " << tc.time << " v: " << tc.v << " bDir: " << tc.bodyDir;
}


void
BallHandlingModule::goalieIntercept()
{
  WBall meet;   // first point where ball can be intercepted (with controlBall)
  AdvSkills::TurtleCommand turtleParams;

  turtleParams = interceptWithFallback( my.stamina,
					1.0 * CATCHABLE_AREA_L,
					0.5 * CATCHABLE_AREA_L,
					0.5 * CATCHABLE_AREA_L,
					CATCHABLE_AREA_L,
					5,
					meet );

  // just in case the goalie leaves the penalty area (strange...)
  if( meet.inField() && !meet.inOurPenaltyArea() )
    turtleParams = interceptWithFallback( availableStamina,
					  my.size + BALL_SIZE + 1.0 * my.kickableMargin,
					  my.size + BALL_SIZE + 0.55 * my.kickableMargin,
					  my.size + BALL_SIZE + 0.3 * my.kickableMargin,
					  CATCHABLE_AREA_L,
					  5,
					  meet );

  // if opponent controls ball intercept him
  if( tact->firstDangerousOpponent != UNKNOWN )
    if( tact->otherTeamReachBallTime[tact->firstDangerousOpponent] < turtleParams.time &&
	tact->otherTeamReachBallTime[tact->firstDangerousOpponent] <= 2 )
      turtleParams = skills->goToPosition( tact->otherTeamReachBallPos[tact->firstDangerousOpponent],
					   0.5 * CATCHABLE_AREA_L,
					   2 * MAXPOWER, true,
					   ( tact->otherTeamReachBallPos[tact->firstDangerousOpponent]
					     - my.pos ).getAngle() );

  // check if ball crosses goal line before controlBall reaches it
  if( ball.pos.getX() > -0.5 * FIELD_LENGTH && meet.pos.getX() < -0.5 * FIELD_LENGTH )
    {
      // check if ball crosses the line between the goal posts
      double factor = - ( 0.5 * FIELD_LENGTH + ball.pos.getX() ) / ( meet.pos.getX() - ball.pos.getX() );
      if( fabs( ball.pos.getY() + factor * ( meet.pos.getY() - ball.pos.getY() ) ) <= 0.5 * GOAL_AREA_WIDTH )
	{
	  Vector2d icptPos( -FIELD_LENGTH / 2.0 + CATCHABLE_AREA_W, 0.0 );
	  icptPos.setY( ball.pos.getY() -
			( 0.5 * FIELD_LENGTH - CATCHABLE_AREA_W + ball.pos.getX() ) /
			( meet.pos.getX() - ball.pos.getX() ) * ( meet.pos.getY() - ball.pos.getY() ) );

	  // try to get ball just before it enters the goal
	  turtleParams = skills->goToPosition( icptPos, 0.5 * CATCHABLE_AREA_L, 2 * MAXPOWER,
					       true, ( icptPos - my.pos ).getAngle() );
	}
    }

  if( turtleParams.state != AdvSkills::TFINISHED )
    {
      lastGoalieInterceptTime = world->time;
      skills->exec( turtleParams );
      return;
    }

  // finished - turn to ball (might happen because fallback position is reached)
  // maybe this is not the optimum (might be difficult to react when the
  // intercept position has to be changed!)
  AdvSkills::TurnCommand turnParams = skills->turnTo( ball );
  skills->exec( turnParams );
}


AdvSkills::TurtleCommand
BallHandlingModule::interceptWithFallback( double stamina,
					   double maxPrecision,
					   double stdPrecision,
					   double minPrecision,
					   double secBorder,
					   long   maxFallbackTime,
					   WBall& meet )
{
  // calculate earliest position where rolling ball can be reached, and run
  // to that position (takes slight changes of this position into account)

  AdvSkills::TurtleCommand turtleParams;

  // see what controlBall will report
  turtleParams = skills->controlBall( stamina, stdPrecision, meet );

  if( turtleParams.time > 1 )
    {
      WBall meet2;
      AdvSkills::TurtleCommand t2params = skills->controlBall( stamina, maxPrecision, meet2 );

      // if player can reach outer circle within one step let him go
      if( t2params.time <= 1 )
	{
	  turtleParams = t2params;
	  meet = meet2;
	}
      else if( turtleParams.time > 4 ) // necessary. Otherwise we probably turn
	// away from the ball to let it roll and perhaps miss the information
	// that the ball is somewhere else...
	{
	  // We will need some time (more than 2 timesteps) until we get the ball.
	  // check if the minimum intercept is stable
	  // if it is, try to intercept with a minimum number of dashes
	  // (this is not done and will not be done in the skills because
	  // it heavily depends on other players)
	  long minTime     = turtleParams.time;
	  long simTime     = turtleParams.time;
	  long maxTime     = turtleParams.time + maxFallbackTime;

	  long oppTimeToReach = ( tact->firstDangerousOpponent == UNKNOWN ? 20 :
				  tact->otherTeamReachBallTime[tact->firstDangerousOpponent] );

	  // now let the ball roll
	  WBall simBall = meet;

	  while( oppTimeToReach > simTime + 2 && simTime <= maxTime )
	    {
	      simBall.simulate();
	      simTime++;

	      // stop simulation if ball is out of field 
	      if( !inField( simBall.pos, secBorder ) )
		break;

	      // stop simulation if ball gets too slow
	      if( norm( simBall.v ) < my.vmax )
		break;

	      AdvSkills::TurtleCommand tmp = skills->goToPositionTimed( simBall.pos, stamina, minPrecision );

	      // stop simulation when it takes longer again to reach the desired point
	      if( tmp.time > minTime )
		break;

	      minTime = tmp.time;
	      turtleParams = tmp;
	      meet = simBall;
	    }
	}
    }

  return turtleParams;
}


void
BallHandlingModule::playerIntercept()
{
  AdvSkills::TurtleCommand turtleParams;
  WBall meet;

  turtleParams = interceptWithFallback( availableStamina,
					my.size + BALL_SIZE + 1.0 * my.kickableMargin,
					my.size + BALL_SIZE + 0.55 * my.kickableMargin,
					my.size + BALL_SIZE + 0.3 * my.kickableMargin,
					2.0,
					5,
					meet );

  if( turtleParams.state != AdvSkills::TFINISHED )
    {
      skills->exec( turtleParams );
      return;
    }

  // finished - turn to ball (might happen because fallback position is reached)
  AdvSkills::TurnCommand turnParams = skills->turnTo( ball );
  skills->exec( turnParams );
}


void
BallHandlingModule::getPass()
{
  // getPass() should not be called in these cases
  if( tact->ourAction.state != TA_PASS || tact->ourAction.receiver != my.number ||
      tact->ourAction.performer < 1 || tact->ourAction.performer > 11 )
    {
      skills->noAction();
      cerr << "Ballhandling::getPass() is called in wrong situation!" << endl;
    }

  double partnerDir = ( world->ourTeam[tact->ourAction.performer].pos - my.pos ).getAngle();

  AdvSkills::TurnCommand tcmd;

  if( fabs( partnerDir ) < M_PI_2 )
    tcmd = skills->turn( 0.0 );
  else if( partnerDir >= M_PI_2 && partnerDir <= 0.75 * M_PI )
    tcmd = skills->turn( 0.25 * M_PI );
  else if( partnerDir > 0.75 * M_PI )
    tcmd = skills->turn( M_PI_2 );
  else if( partnerDir <= -M_PI_2 && partnerDir >= -0.75 * M_PI )
    tcmd = skills->turn( -0.25 * M_PI );
  else
    tcmd = skills->turn( -M_PI_2 );

  if( tcmd.state == AdvSkills::RFINISHED )
    skills->noAction();
  else
    skills->exec( tcmd );
}
