//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

using namespace std;


/////////////////////////////////////////////////////////////////////////
//
// EVALUATE
//
/////////////////////////////////////////////////////////////////////////

Grade 
BallHandlingModule::evaluate()
{
  // main evaluate method, called by master control to evaluate
  // current situation by module and return a grade of responsibility
  // and desired activity. Naturally plans some of the future actions
  // already in its computations.
  
  acted = false;
  waitForBall = false;

  // check whether to turn the head to the ball
  // done here since we ALWAYS need the turnHeadGrade
  // and it has to be done before turnHeadEvaluate to allow
  // body turns to be done.
  Grade lookG = evaluateFindBall();

  // if player is immobile after tackling, let positioning module do the job
  if( my.cyclesSinceTackle <= TACKLE_CYCLES )
    {
      action = WAIT;
      grade.set_BAD();
      return grade;
    }

  // check for standard situation -> perform FIND_BALL if needed
  // penalty modes are handled like PLAY_ON
  if( world->playMode != PLAY_MODE_PLAY_ON &&
      world->playMode != PLAY_MODE_PENALTY_TAKEN_THEM &&
      world->playMode != PLAY_MODE_PENALTY_TAKEN_US &&
      world->playMode != PLAY_MODE_PENALTY_SCORE_THEM &&
      world->playMode != PLAY_MODE_PENALTY_SCORE_US &&
      world->playMode != PLAY_MODE_PENALTY_MISS_THEM &&
      world->playMode != PLAY_MODE_PENALTY_MISS_US &&
      world->playMode != PLAY_MODE_PENALTY_READY_THEM &&
      world->playMode != PLAY_MODE_PENALTY_READY_US &&
      world->playMode != PLAY_MODE_PENALTY_SETUP_THEM &&
      world->playMode != PLAY_MODE_PENALTY_SETUP_US )
    {
      // special case: GOAL_KICK_US : Goalie should be able to catch
      // in case anything unexpected happens
      if( world->playMode != PLAY_MODE_GOAL_KICK_US || my.isGoalie != GOALIE_TRUE )
	{
	  action = FIND_BALL;
	  grade.set_BAD();
	  DECISION( "eval -> only FIND_BALL possible, planning to let standard sit have control" );
	  return grade;
	}
    }

  if( lookG.is_PERFECT() )
    {
      // this is only possible if ball data is corrupt in which case we cannot do anything else
      action = FIND_BALL;
      return lookG;
    }

  grade = ( my.isGoalie == GOALIE_TRUE &&
	    world->playMode != PLAY_MODE_PENALTY_TAKEN_US &&
	    world->playMode != PLAY_MODE_PENALTY_SCORE_THEM &&
	    world->playMode != PLAY_MODE_PENALTY_SCORE_US &&
	    world->playMode != PLAY_MODE_PENALTY_MISS_THEM &&
	    world->playMode != PLAY_MODE_PENALTY_MISS_US &&
	    world->playMode != PLAY_MODE_PENALTY_READY_US &&
	    world->playMode != PLAY_MODE_PENALTY_SETUP_US )
    ? evaluateGoalie() : evaluatePlayer();

  if( grade.is_worse_than( lookG ) )
    {
      action = FIND_BALL;
      DECISION( "eval -> ball too old, planning to look for ball" );
      return lookG;
    }

  // looking after ball is not really needed, so perform normal action
  return grade;
}


TurnHeadWish
BallHandlingModule::turnHeadEvaluate()
{
  TurnHeadWish thw;
  thw.lookNowhere( Grade::IMPOSSIBLE, VIEW_WIDTH_NARROW );

  // [MW] if it is likely that an opponent player will kick the ball, look at the ball  
  if( tact->firstDangerousOpponent != UNKNOWN && world->otherTeam[tact->firstDangerousOpponent].age < 2 &&
      world->otherTeam[tact->firstDangerousOpponent].distTo( ball ) <
      world->otherTeam[tact->firstDangerousOpponent].kickRange() * 1.2 &&
      my.distTo( ball ) < BALL_SPEED_MAX * 8.0 )
    {
      thw.lookAtBall( Grade::PERFECT,
		      ( my.distTo( ball ) < BALL_SPEED_MAX * 4.0 ? VIEW_WIDTH_NARROW : VIEW_WIDTH_NORMAL ) );
      DECISION("ball in opps kickrange, looking at ball");
      return thw;
    } 

  // step 1: handle FIND_BALL
  if( action == FIND_BALL && acted )
    {
      thw.lookAtBodyDir( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
      return thw;
    }

  if( turnHeadGrade.is_better_than( Grade::GOOD ) )
    { // [MW] coneTime is too big, better look at the ball
      thw.lookAtBall( turnHeadGrade, VIEW_WIDTH_NARROW );
      return thw;
    }

  // step 2: if player "feels" the ball try looking around
  if( my.distTo( ball ) < VISIBLE_DISTANCE )
    {
      // [MW] playerAttack needs visual information
      // -> let positionModule turn head in this case
      thw.lookAround( acted ? Grade::PERFECT : Grade::GOOD, VIEW_WIDTH_NARROW );
      return thw;
    }

  // step 3: if player performed an intercept (and the ball is not close) look at the ball frequently
  if( acted && ( action == GOALIE_INTERCEPT || action == PLAYER_INTERCEPT || action == PLAYER_GET_PASS ) )
    {
      if( ball.age > 0 )
	{
	  thw.lookAtBall( Grade::PERFECT, VIEW_WIDTH_NARROW );
	  return thw;
	}
      if( waitForBall && tact->firstDangerousOpponent != UNKNOWN && 
	  world->otherTeam[tact->firstDangerousOpponent].age > 1
	  && tact->otherTeamReachBallTime[tact->firstDangerousOpponent] < 
	  tact->ourTeamReachBallTime[world->my.number] + 3)
	{
	  thw.lookAtOppPlayer( Grade::PERFECT, tact->firstDangerousOpponent, VIEW_WIDTH_NARROW );
	  return thw;
	}

      thw.lookAround( Grade::GOOD, VIEW_WIDTH_NARROW );
      return thw;
    }

  if( turnHeadGrade.is_better_than( Grade::IMPOSSIBLE ) )
    {
      // looking at ball is the optimum
      thw.lookAtBall( turnHeadGrade, VIEW_WIDTH_NARROW );
    }

  return thw;
}


Grade
BallHandlingModule::evaluateGoalie()
{
  // evaluate goalie's possible actions
  Grade g;

  // check if the situation is dangerous
  bool danger =
    ( ball.pos.getX() < -FIELD_LENGTH / 4.0 ) &&   // ball near our goal
    ( tact->ourAction.state != TA_PASS ||          // do not run after kickoffs!
      tact->ourAction.performer != my.number );

  // check point where the ball could be stopped by an opponent
  WBall simBall = ball;
  if( tact->firstDangerousOpponent != UNKNOWN )
    for( int i = 0; i <= tact->otherTeamReachBallTime[tact->firstDangerousOpponent]; i++ )
      simBall.simulate();
  else
    simBall.pos += simBall.v * ( 1.0 / ( 1.0 - BALL_DECAY ) );

  // if this point is far from the goal (e.g. after goalkick) there is no direct danger
  if( simBall.pos.getX() > -FIELD_LENGTH / 4.0 )
    danger = false;

  if( !danger )
    {
      // no danger - positioning should do the job
      action = WAIT;
      g.set_BAD();
      DECISION( "eval -> no danger, positioning should do the job!" );
      return g;
    }

  // only get active in dangerous situations

  // in this case catching into ball direction is safe (at least if ball age is zero)
  if( my.distTo( ball ) < lowerDistLimit( CATCHABLE_AREA_L, QUANTIZE_STEP ) )
    {
      // try to catch ball, if possible -> PERFECT
      catchParams = skills->goalieCatch( ( ball.pos - my.pos ).getAngle() );
      if( catchParams.state == AdvSkills::COK )
	{
	  action = GOALIE_CATCH;
	  g.set_PERFECT();
	  DECISION( "eval -> planning to catch ball (into ball direction)" );
	  return g;
	}
    }
  // in this case catching into ball direction is unsafe
  else if( my.distTo( ball ) < upperDistLimit( CATCHABLE_AREA_L, QUANTIZE_STEP ) )
    {
      // better catch into different direction (this is unsafe if ball age is not zero)
      catchParams = skills->goalieCatch( ( ball.pos - my.pos ).getAngle() +
					 ( Vector2d( CATCHABLE_AREA_L, 0.5 * CATCHABLE_AREA_W ) ).getAngle() );
      if( catchParams.state == AdvSkills::COK )
	{
	  WBall wb;
	  // ok, catch is unsafe, check if we can wait
	  AdvSkills::TurtleCommand turtleParams =
	    skills->controlBall( 2 * MAXPOWER, 0.5 * CATCHABLE_AREA_L, wb );
	  // one dash helps and no opponent is too near
	  if( turtleParams.time == 1 &&
	      ( tact->firstDangerousOpponent == UNKNOWN ||
		tact->otherTeamReachBallTime[tact->firstDangerousOpponent] > 1 ) )
	    {
	      // check if ball will be safely in our penalty area in the next step
	      if( inOurPenaltyArea( ball.pos + ball.v, BALL_SIZE ) )
		{
		  // another dash makes catching safer, so intercept
		  action = GOALIE_INTERCEPT;
		  g.set_PERFECT();
		  DECISION( "eval -> catch unsafe, going near ball" );
		  return g;
		}
	    }
	  // here for some reason a security dash was impossible, so better try unsafe catch
	  action = GOALIE_CATCH;
	  g.set_PERFECT();
	  DECISION( "eval -> planning to catch ball (somewhere else)" );
	  return g;
	}
    }

  // catching is impossible, try to place ball (passmodule should do domething!)
  if( my.ballKickable( ball ) )
    {
      g.set_BAD();

      // place ball away from opponent
      action = GOALIE_PLACE;
      placeParams = skills->placeBall( radAng2Vector2d( skills->safeCircle, 0.0 ) );

      // placing is perfect if we can be sure that the opponent can not reach us before we can catch again
      if( ball.inOurPenaltyArea() &&
	  ( tact->firstDangerousOpponent == UNKNOWN ||
	    tact->otherTeamReachBallTime[tact->firstDangerousOpponent] >
	    max( 0, CATCH_BAN_CYCLES - world->cyclesSinceCatch ) + 3 ) )
	g.set_PERFECT();
      else if( tact->firstDangerousOpponent != UNKNOWN )
	{
	  Angle placeDir = ( my.pos - world->otherTeam[tact->firstDangerousOpponent].pos ).getAngle();
	  placeParams = skills->placeBall( radAng2Vector2d( skills->safeCircle, placeDir ) );
	}

      DECISION( "eval -> cannot catch ball, planning to place" );
      if( g.is_BAD() )
	DECISION( "placing is BAD -> passmodule should do something!" );

      return g;
    }

  // if we reach this point goalie could only intercept
  action = GOALIE_INTERCEPT;

  // goalie is most responsible but the ball is not in kick/catch range
  if( tact->firstResponsiblePlayer == my.number || tact->secondResponsiblePlayer == my.number )
    {
      // there is no dangerous opponent around
      if( tact->firstDangerousOpponent == UNKNOWN )
	{
	  g.set_PERFECT();
	  DECISION( "eval -> no opp. around, closest to ball, planning to intercept it" );
	  return g;
	}

      // goalie should not leave our penalty area
      if( inOurPenaltyArea( tact->ourTeamReachBallPos[my.number] ) )
	{
	  // goalie reaches the ball before opponents
	  if( tact->ourTeamReachBallTime[my.number] < tact->otherTeamReachBallTime[tact->firstDangerousOpponent] )
	    {
	      g.set_PERFECT();
	      DECISION( "eval -> closest to ball, planning to intercept it" );
	      return g;
	    }
	  // goalie does not reach the ball before opponents but already intercepted
	  if( world->time == lastGoalieInterceptTime + 1 &&
	      tact->ourTeamReachBallTime[my.number] <= tact->otherTeamReachBallTime[tact->firstDangerousOpponent] + 1 )
	    {
	      g.set_PERFECT();
	      DECISION( "eval -> not closest but started intercept, planning to intercept again" );
	      return g;
	    }
	}
    }

  // goalie reaches the ball within a few steps (in our penalty area)
  if( tact->ourTeamReachBallTime[my.number] <= 2 )
    if( inOurPenaltyArea( tact->ourTeamReachBallPos[my.number] ) )
      {
	g.set_PERFECT();
	DECISION( "eval -> closest to ball, but opp. even closer, planning to intercept it" );
	return g;
      }

  // if there is a goalkick, intercept
  // (use GOAL_AREA_WIDTH instead of GOAL_WIDTH for safety)
  if( ball.willCrossOurGoalLine( 0.5 * GOAL_AREA_WIDTH ) )
    {
      DECISION( "eval-> ball will cross goal line, planning to intercept it" );
      g.set_PERFECT();
      return g;
    }

  // in all other cases positioning module should do the job!
  g.set_BAD();
  DECISION( "eval -> planning to intercept, but positioning should do the job!" );
  return g;
}


Grade
BallHandlingModule::evaluatePlayer()
{
  Grade g( Grade::BAD );

  if( my.ballKickable( ball ) )
    return evaluateDribble();   // try to fastdribble, to dribble or to place

  action = PLAYER_INTERCEPT;

  // if I am pass receiver, find ball first, then intercept
  if( tact->ourAction.state == TA_PASS && tact->ourAction.receiver == my.number )
    {
      g.set_PERFECT();
      if( ball.age < tact->ourAction.age )
	haveSeenBallSincePass = true;

      if( !haveSeenBallSincePass )
	action = PLAYER_GET_PASS;
    }
  else
    {
      haveSeenBallSincePass = false;
      // get ball if first responsible
      if( tact->firstResponsiblePlayer == my.number )
	{
	  g.set_PERFECT();
	  DECISION( "eval -> 1. resp, icpt" );
	}
      // in this case a teammate seems to be closer
      else if( tact->ourAction.state == TA_CONTROL && tact->ourAction.performer == my.number )
	{
	  g.set_AVERAGE();
	  DECISION( "eval -> not 1. resp, icpt" );
	}
    }

  // downgrade if ball control is not SAFE
  if( tact->ourBallControl != BC_SAFE && 
      ( world->ball.pos.getX() < 0.0  || tact->ourTeamDefender( world->my.number ) ) )
    {
      g.downgrade();
      DECISION("eval -> should intercept, but ball control is LOST so let PLAYER_ATTACK do the job");
    }
  else if( opponentControlsBall )
    {
      g.downgrade();
      DECISION( "eval -> opp. controls ball, let PLAYER_ATTACK do the job" );
    }

  return g;
}


const BallHandlingModule::DribbleAreaData
BallHandlingModule::areas[] =
{
  // 1. right opponent corner-triangle
  {
    5, {
      { LEFT,           Grade::GOOD,    false },   // true means: no fastdribble
      { LEFT_LEFT_BACK, Grade::GOOD,    false },
      { LEFT_BACK,      Grade::AVERAGE, false },
      { BACK_LEFT_BACK, Grade::POOR,    false },
      { BACK,           Grade::POOR,    false }
    },
    Grade::AVERAGE
  },
  // 2. left opponent corner-triangle
  {
    5, {
      { RIGHT,            Grade::GOOD,    false },
      { RIGHT_RIGHT_BACK, Grade::GOOD,    false },
      { RIGHT_BACK,       Grade::AVERAGE, false },
      { BACK_RIGHT_BACK,  Grade::POOR,    false },
      {	BACK,             Grade::POOR,    false }
    },
    Grade::AVERAGE
  },
  // 3. right border
  {
    3, {
      { LEFT_FORWARD,         Grade::GOOD,    false },
      { FORWARD_LEFT_FORWARD, Grade::GOOD,    false },
      { FORWARD,              Grade::AVERAGE, true  }
    },
    Grade::AVERAGE
  },
  // 4. left border
  {
    3, {
      { RIGHT_FORWARD,         Grade::GOOD,    false },
      { FORWARD_RIGHT_FORWARD, Grade::GOOD,    false },
      { FORWARD,               Grade::AVERAGE, true  }
    },
    Grade::AVERAGE
  },
  // 5. our right backfield
  {
    5, {
      { RIGHT_FORWARD,         Grade::AVERAGE, false },
      { RIGHT_RIGHT_FORWARD,   Grade::AVERAGE, false },
      { RIGHT,                 Grade::AVERAGE, false },
      { FORWARD_RIGHT_FORWARD, Grade::AVERAGE, false },
      { FORWARD,               Grade::AVERAGE, false }
    },
    Grade::AVERAGE
  },
  // 6. our left backfield
  {
    5, {
      { LEFT_FORWARD,         Grade::AVERAGE, false },
      { LEFT_LEFT_FORWARD,    Grade::AVERAGE, false },
      { LEFT,                 Grade::AVERAGE, false },
      { FORWARD_LEFT_FORWARD, Grade::AVERAGE, false },
      { FORWARD,              Grade::AVERAGE, false }
    },
    Grade::AVERAGE
  },
  // 7. right front flank ("Axel's Leitplanke")
  {
    6, {
      { FORWARD,              Grade::PERFECT, false },
      { FORWARD_LEFT_FORWARD, Grade::PERFECT, false },
      { LEFT_FORWARD,         Grade::PERFECT, false },
      { LEFT_LEFT_FORWARD,    Grade::GOOD,    false },
      { LEFT,                 Grade::AVERAGE, false },
      { BACK,                 Grade::POOR,    false }
    },
    Grade::AVERAGE
  },
  // 8. left front flank
  {
    6, {
      { FORWARD,                Grade::PERFECT, false },
      { FORWARD_RIGHT_FORWARD,  Grade::PERFECT, false },
      { RIGHT_FORWARD,          Grade::PERFECT, false },
      { RIGHT_RIGHT_FORWARD,    Grade::GOOD,    false },
      { RIGHT,                  Grade::AVERAGE, false },
      { BACK,                   Grade::POOR,    false }
    },
    Grade::AVERAGE
  },
  // 9. somewhere else in the right body of the field
  {
    9, {
      { FORWARD,               Grade::PERFECT, false },
      { FORWARD_RIGHT_FORWARD, Grade::PERFECT, false },
      { FORWARD_LEFT_FORWARD,  Grade::GOOD,    false },
      { RIGHT_FORWARD,         Grade::AVERAGE, false },
      { RIGHT_RIGHT_FORWARD,   Grade::AVERAGE, false },
      { LEFT_FORWARD,          Grade::AVERAGE, false },
      { RIGHT,                 Grade::AVERAGE, true  },
      { LEFT,                  Grade::POOR,    false },
      { LEFT_LEFT_FORWARD,     Grade::POOR,    false }
    },
    Grade::AVERAGE
  },
  // 10. somewhere else in the left body of the field
  {
    9, {
      { FORWARD,               Grade::PERFECT, false },
      { FORWARD_LEFT_FORWARD,  Grade::PERFECT, false },
      { FORWARD_RIGHT_FORWARD, Grade::GOOD,    false },
      { LEFT_FORWARD,          Grade::AVERAGE, false },
      { LEFT_LEFT_FORWARD,     Grade::AVERAGE, false },
      { RIGHT_FORWARD,         Grade::AVERAGE, false },
      { LEFT,                  Grade::AVERAGE, true  },
      { RIGHT_RIGHT_FORWARD,   Grade::POOR,    false },
      { RIGHT,                 Grade::POOR,    false }
    },
    Grade::AVERAGE
  }
};


int BallHandlingModule::dribbleArea()
{
  const double sideBorder( 5.0 );
  const double backBorder( 1.5 * PENALTY_AREA_LENGTH );
  const double cornerDist( 0.5 * ( FIELD_WIDTH - GOAL_WIDTH ) );
  const double targetingDist( cornerDist + 0.5 * GOAL_WIDTH );

  // the dribble directions depend on the player's position on the field:

  // 1. right opponent corner-triangle
  if( 2 * ( 0.5 * FIELD_LENGTH - my.pos.getX() ) +
      ( 0.5 * FIELD_WIDTH - my.pos.getY() ) < cornerDist )
    return 0;

  // 2. left opponent corner-triangle
  if( 2 * ( 0.5 * FIELD_LENGTH - my.pos.getX() ) +
      ( 0.5 * FIELD_WIDTH + my.pos.getY() ) < cornerDist )
    return 1;

  // 3. right border
  if( my.pos.getY() > 0.5 * FIELD_WIDTH - sideBorder )
    return 2;

  // 4. left border
  if( my.pos.getY() < -0.5 * FIELD_WIDTH + sideBorder )
    return 3;

  // 5./6. our right and left backfield
  if( my.pos.getX() < -0.5 * FIELD_LENGTH + backBorder )
    {
      if( my.pos.getY() > 0 )
	return 4;
      return 5;
    }

  // 7. right front flank ("Axel's Leitplanke")
  if( 0.5 * FIELD_LENGTH - my.pos.getX() +
      0.5 * FIELD_WIDTH - my.pos.getY() < targetingDist )
    return 6;

  // 8. left front flank
  if( 0.5 * FIELD_LENGTH - my.pos.getX() +
      0.5 * FIELD_WIDTH + my.pos.getY() < targetingDist)
    return 7;

  // 9./10. somewhere else in the right or left part of the field
  if( my.pos.getY() > 0 )
    return 8;
  return 9;
}


Grade 
BallHandlingModule::evaluateDribble()
{
  Grade g( Grade::BAD );   // is returned

  // ------------------------------< step 1 >------------------------------
  // if we do not know if a collision has happened, we do not dribble
  // but only if we are more or less alone (else we risk it)

  if( ( world->collState == CRITICAL ) &&
      ( ( tact->firstDangerousOpponent == UNKNOWN ) ||
	( tact->otherTeamReachBallTime[tact->firstDangerousOpponent] >
	  world->otherTeam[tact->firstDangerousOpponent].age ) ) )
    {
      action = PLAYER_PLACE;
      placeParams = skills->placeBall( radAng2Vector2d( skills->safeCircle, ( ball.pos - my.pos ).getAngle() ) );
      g.set_PERFECT();
      return g;
    }

  // ------------------------------< step 2 >------------------------------
  // evaluate fastdribble

  // minimum dist to field border watched while dribbling
  const double sideBorder( 5.0 );

  const DribbleAreaData *adt = &areas[dribbleArea()];

  // keep back some stamina in case we lose the ball while dribbling,
  // so we can intercept it again within 2-3 steps.
  dribbleStamina = max( my.staminaIncMax, availableStamina - 2 * MAXPOWER );

  // dribble direction; length gives the point where we want to control the ball again
  Vector2d propReachVec;

  // Grade for currently evaluated direction and distance
  Grade tmp_g;

  // ball velocity of currently evaluated kick
  Vector2d tmp_ballV;

  AdvSkills::KickCommand tmp_kickParams;

  int oppNo = world->otherTeamClosestTo( my.pos );

  // check all possible directions
  for( int cdir = 0; cdir < adt->numDirs; cdir++ )
    {
      // check if fastdribble is allowed for that direction
      if( !adt->dirs[cdir].dribbleonly )
	{
	  // check several fastdribble distances 
	  // (iterating number of dashes needed for intercept)
	  Angle currentDir = adt->dirs[cdir].dir.getAngle();

	  for( int n = 28; n >= 4; n -= 4 )
	    {
	      // distance player can manage with n dashes with dribbleStamina
	      double rundist =  my.dashPowerRate * min( MAXPOWER, my.staminaIncMax + dribbleStamina / n )
		/ ( 1.0 - my.decay ) * ( n - my.decay * ( 1.0 - power( my.decay, n ) ) / ( 1.0 - my.decay ) );

	      // distance standard player can manage in n steps
	      double maxdist = n * PLAYER_SPEED_MAX;

	      propReachVec = rundist * adt->dirs[cdir].dir;

	      double corrAngle = fabs( ( Vector2d( rundist, maxdist ) ).getAngle() );

	      // no fastdribble at the field border ([C] age==30)
	      if( world->otherTeamInCorridor( my.pos, my.pos + propReachVec, 2 * maxdist, corrAngle, 30 ) == 0 &&
		  inField( ball.pos + propReachVec, sideBorder ) )
		{
		  tmp_g = adt->dirs[cdir].grade;

		  // do not fastdribble if no opponent is close and too many turns are needed
		  if( oppNo == UNKNOWN ||
		      my.distTo( world->otherTeam[oppNo] ) >
		      world->otherTeam[oppNo].kickRange() + world->otherTeam[oppNo].vmax )
		    {
		      if( world->dirAge( currentDir ) > 3 )
			{
			  tmp_g.downgrade();
			  if( world->dirAge( currentDir ) > 5 )
			    tmp_g.downgrade();
			}
		    }

		  if( tmp_g.is_better_than( g ) )
		    {
		      // check if ball will leave kickrange with this kick
		      tmp_ballV  = ( 1.0 - BALL_DECAY ) / ( 1.0 - power( BALL_DECAY, n + 3 ) ) * propReachVec;
		      tmp_kickParams = skills->kick( tmp_ballV, 1 );
		      if( dist( my.pos + my.v, ball.pos + tmp_kickParams.v ) > my.kickRange() )
			{
			  g          = tmp_g;
			  dribbleDir = currentDir;
			  fd_ballV   = tmp_ballV;
			}
		    }
		}
	    }
	}
    }

  Grade fastDribbleGrade = g;

  // ------------------------------< step 3 >------------------------------
  // check simple dribble

  g.set_BAD();
  AdvSkills::DribbleCommand testParams;

  Angle    holdDir;      // direction where to hold the ball when dribbling or placing
  double   oppReachRadius;

  // check all possible directions
  for( int cdir = 0; cdir < adt->numDirs; cdir++ )
    {
      dribbleDir = adt->dirs[cdir].dir.getAngle();
      tmp_g = adt->dirs[cdir].grade;
      tmp_g.upgrade();
      if( !tmp_g.is_better_than( g ) )
	break;

      for( int sign = 1; sign >= -1; sign -= 2 )
	{
	  // choose direction where to place ball while dribbling
	  holdDir = dribbleDir + ( ( ( ball.pos - my.pos ).getAngle() - dribbleDir > 0 ) ?
				   Angle( sign / 3.0 * M_PI ) :
				   Angle( -sign / 3.0 * M_PI ) );

	  testParams = skills->dribble( dribbleDir, holdDir, dribbleStamina );
	  // testParams = skills->dribble( dribbleDir, holdDir, my.staminaIncMax );

	  WBall simBall = ball;
	  simBall.v = testParams.v_b;
	  simBall.simulate();
	  WPlayer simMy = my;
	  simMy.v = testParams.v_p;
	  simMy.simulate();
	  WMovedObject meet;
	  AdvSkills::TurtleCommand icptParams;
	  icptParams = skills->interceptTimed( simMy, simBall, dribbleStamina, my.kickRange(), meet );

	  // check if opponent players could reach the ball
	  for( long t = 0; t <= icptParams.time; t ++ )
	    {
	      for( int i = 1; i <= 11; i++ )
		{
		  // do not consider opponents that are not close
		  if( my.distTo( world->otherTeam[i].pos ) > 1.5 * VISIBLE_DISTANCE )
		    continue;

		  oppReachRadius = ( t + 1 ) * world->otherTeam[i].vmax + world->otherTeam[i].kickRange();
		  if( world->otherTeam[i].age == 1 )
		    oppReachRadius += world->otherTeam[i].vmax;

		  if( dist( world->otherTeam[i].pos, simBall.pos ) < oppReachRadius )
		    {
		      // tmp_g.downgrade();
		      // if( !tmp_g.is_better_than( g ) )
		      // break;

		      Vector2d corrPos = ball.pos - world->otherTeam[i].pos;
		      corrPos.stretch( oppReachRadius / ( corrPos.getLength() + 0.0001 ) );

		      if( simBall.inCorridor( world->otherTeam[i].pos - corrPos,
					      world->otherTeam[i].pos + corrPos,
					      2 * world->otherTeam[i].kickRange() ) )
			{
			  tmp_g.downgrade();
			  if( !tmp_g.is_better_than( g ) )
			    break;
			}

		      corrPos = ball.pos + ball.v - world->otherTeam[i].pos;
		      corrPos.stretch( oppReachRadius / ( corrPos.getLength() + 0.0001 ) );

		      if( simBall.inCorridor( world->otherTeam[i].pos - corrPos,
					      world->otherTeam[i].pos + corrPos,
					      2 * world->otherTeam[i].kickRange() ) )
			{
			  tmp_g.downgrade();
			  if( !tmp_g.is_better_than( g ) )
			    break;
			}
		    }
		}
	    }

	  if( tmp_g.is_better_than( g ) )
	    {
	      g             = tmp_g;
	      dribbleParams = testParams;
	    }
	  if( g.is_PERFECT() )
	    break;
	}
      if( g.is_PERFECT() )
	break;
    }

  // now decide if to dribble or to fastdribble
  action = PLAYER_FAST_DRIBBLE;
  if( g.is_better_than( fastDribbleGrade ) )
    action = PLAYER_DRIBBLE;
  else
    g = fastDribbleGrade;

  if( g.is_better_than( adt->placeGrade ) )
    return g;

  // ------------------------------< step 4 >------------------------------
  // we have decided not to dribble but to place the ball

  action = PLAYER_PLACE;
  g = adt->placeGrade;

  // if we have already been placing for many cycles, do not upgrade
  // (to avoid deadlocks!)
  if( placeCycles < 50 )   // [C]
    {
      g.upgrade();
      if( placeCycles < 20 )
	g.upgrade();
    }

  // if all opponents are far away keep ball where it is
  if( ( tact->firstDangerousOpponent == UNKNOWN ) ||
      ( my.distTo( world->otherTeam[tact->firstDangerousOpponent] ) > VISIBLE_DISTANCE ) )
    holdDir = ( ball.pos - my.pos ).getAngle();
  else
    {
      // if two opponents are close try to place ball away from both (not easy...)
      if( ( tact->secondDangerousOpponent != UNKNOWN ) &&
	  ( my.distTo( world->otherTeam[tact->secondDangerousOpponent] ) < VISIBLE_DISTANCE ) )
	{
	  Angle opDir1 = ( world->otherTeam[tact->firstDangerousOpponent].pos - my.pos ).getAngle();
	  Angle opDir2 = ( world->otherTeam[tact->secondDangerousOpponent].pos - my.pos ).getAngle();
	  holdDir = PI + ( opDir1 + 1./3 * ( opDir2 - opDir1 ) );
	  g.downgrade();
	}
      // if there is only one close opponent place ball away from him
      else
	{
	  holdDir = PI + ( world->otherTeam[tact->firstDangerousOpponent].pos - my.pos ).getAngle();
	}
    }

  placeParams = skills->placeBall( radAng2Vector2d( skills->safeCircle, holdDir ) );
  return g;
}


Grade
BallHandlingModule::evaluateFindBall()
{
  Grade physTurnGrade;   // turn body?

  // turnHeadGrade is changed in this method; the result is used in turnHeadEvaluate()
  turnHeadGrade.set_BAD();
  physTurnGrade.set_BAD();   // init with BAD

  // IDEA: Both turnHeadGrade and physTurnGrade are upgraded as information about the
  // ball gets worse. If physTurnGrade is high enough findBall() should be called by the
  // mastercontrol. In this case the head is turned into player's body direction.
  // But if findBall() is not called the head is turned towards the ball if the
  // ballhandling is allowed to perform the turn head command.

  // ------------------------------< step 1 >------------------------------
  // check how much time ball has spent in view cone

  int maxSeeGap = 2;

  // should have seen ball (because position is in my view cone) but did not see it
  if( ball.coneTime >= maxSeeGap )
    {
      turnHeadGrade.set_PERFECT();   // -> turn head first
      if( ball.coneTime >= maxSeeGap + 2 )
	physTurnGrade.set_PERFECT();   // -> turn body
    }

  // ------------------------------< step 2 >------------------------------
  // check ball age; if it is too old, upgrade body and head turn

  long acceptableAge = 10;
  if( my.number == tact->firstResponsiblePlayer )
    acceptableAge = min( acceptableAge, maxSeeGap );
  else if( my.number == tact->secondResponsiblePlayer )
    acceptableAge = min( acceptableAge, ( 3 * maxSeeGap ) / 2 );

  long diff = ( ball.age - acceptableAge ) / 2;
  // for each delay make turn more attractive
  while( diff > 0 )
    {
      physTurnGrade.upgrade();
      turnHeadGrade.upgrade();
      diff--;
    }

  DECISION("FindBall: ball.age: " << ball.age << " conetime: " << ball.coneTime 
	   << " maxSeeGap: " << maxSeeGap << " AcceptableAge: " << acceptableAge 
	   << " physTurnGrade: " << physTurnGrade << " turnHeadGrade: " << turnHeadGrade);
  return physTurnGrade;
}

