//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef __BALL_HANDLING_MODULE_H__
#define __BALL_HANDLING_MODULE_H__


#include "grade.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"

/*
  short description:

  When called for feedback() some auxiliary variables (my, ball, etc.) are set.

  In evaluate() we first test for some simple special cases, and then call the
  more complex evaluatePlayer() and evaluateGoalie() routines. They do the rest
  of the evaluation, and make use of the further singled out 
  dribble evaluation. At the end of the eval() method, the variables action and
  g (for grade) have to be set to the desired values to be returnd to 
  master control. Also take care to set any local temporary variables (e.g. 
  for values already computed for the later execution of the command). All
  this is implemented in ballhandlingEvaluate.C.

  In act() we split up the different cases in a switch{}. Auxiliary routines
  then do the rest of the action execution in most cases. All this is 
  implemented in ballhandlingAct.C. Some simulation routines and specific 
  turnHead routines (taking care of the players current action) are 
  included in this file as well.
*/


// different actions this module can perform 
enum BHAction
{
  WAIT = 0,             // turn perpetual to ball direction
  FIND_BALL,            // turn to get ball in sight

  GOALIE_INTERCEPT,     // go for a rolling ball (uses also player intercept)
  GOALIE_CATCH,         // catch ball
  GOALIE_PLACE,         // keep ball away from opponent

  PLAYER_INTERCEPT,     // go for a rolling ball
  PLAYER_FAST_DRIBBLE,  // kick ball ahead a few steps
  PLAYER_DRIBBLE,       // keep ball close, but move
  PLAYER_PLACE,         // just keep ball away from opponent
  PLAYER_GET_PASS       // try to find ball before intercepting a pass
};


// dribble directions

const Vector2d LEFT                  = radAng2Vector2d( 1.0, -M_PI/2.0 );
const Vector2d LEFT_LEFT_FORWARD     = radAng2Vector2d( 1.0, -3.0*M_PI/8.0 );
const Vector2d LEFT_FORWARD          = radAng2Vector2d( 1.0, -M_PI/4.0 );
const Vector2d FORWARD_LEFT_FORWARD  = radAng2Vector2d( 1.0, -M_PI/8.0 );
const Vector2d FORWARD               = radAng2Vector2d( 1.0, 0.0 );
const Vector2d FORWARD_RIGHT_FORWARD = radAng2Vector2d( 1.0, M_PI/8.0 );
const Vector2d RIGHT_FORWARD         = radAng2Vector2d( 1.0, M_PI/4.0 );
const Vector2d RIGHT_RIGHT_FORWARD   = radAng2Vector2d( 1.0, 3.0*M_PI/8.0 );
const Vector2d RIGHT                 = radAng2Vector2d( 1.0, M_PI/2.0 );
const Vector2d RIGHT_RIGHT_BACK      = radAng2Vector2d( 1.0, 5.0*M_PI/8.0 );
const Vector2d RIGHT_BACK            = radAng2Vector2d( 1.0, 3.0*M_PI/4.0 );
const Vector2d BACK_RIGHT_BACK       = radAng2Vector2d( 1.0, 7.0*M_PI/8.0 );
const Vector2d BACK                  = radAng2Vector2d( 1.0, M_PI );
const Vector2d BACK_LEFT_BACK        = radAng2Vector2d( 1.0, -7.0*M_PI/8.0 );
const Vector2d LEFT_BACK             = radAng2Vector2d( 1.0, -3.0*M_PI/4.0 );
const Vector2d LEFT_LEFT_BACK        = radAng2Vector2d( 1.0, -5.0*M_PI/8.0 );


/////////////////////////////////////////////////////////////////////
//
// Class Ballhandling Module
//
/////////////////////////////////////////////////////////////////////


class BallHandlingModule
{
public:
  
  BallHandlingModule( Worldmodel * _world, TacticalModule * _tact, AdvSkills * _skills );
  
  ~BallHandlingModule() {}

  // feedback of communicated information
  void feedbackWorld();
  
  // evaluation of given situation into a Grade
  Grade evaluate();

  // Calculate a turn head wish
  TurnHeadWish turnHeadEvaluate();
    
  // perform action according to last evaluation
  bool act();
  
private:
  
  // ----------< basic data >----------

  Worldmodel     * world;
  TacticalModule * tact;
  AdvSkills      * skills;

  WPlayer my;
  WBall   ball;
  double  availableStamina;   // total stamina minus recovery threshold

  bool    opponentControlsBall;

  AdvSkills::CatchCommand   catchParams;
  AdvSkills::ExtKickCommand kickParams;
  AdvSkills::DribbleCommand dribbleParams;
  AdvSkills::KickCommand    placeParams;

  // ----------< evaluation and decision >----------

  Grade evaluateGoalie();

  Grade evaluatePlayer();

  Grade evaluateDribble(); 

  void updateTA();

  BHAction action;         // action calculated in evaluate(), to be executed in act()
  BHAction lastAction;     // action performed when module was called last time...
  long     lastActCycles;  // ...and the number of cycles that have passed since then

  // grade of this action
  Grade grade;
  // true if action was actually executed
  bool  acted;

  // ----------< intercept >----------

  // intercepts the freely rolling ball at the computed intercept point
  // (different for player and goalie)
  void  goalieIntercept();
  void  playerIntercept();

  // intercept routine called by goalieIntercept() and playerIntercept()
  // provides intercept with fallback
  AdvSkills::TurtleCommand
  interceptWithFallback( double stamina,             // maximum stamina to be used
			 double maxPrecision,        // maximum error allowed
			 double stdPrecision,        // standard error
			 double minPrecision,        // error allowed for fallback
			 double secBorder,           // distance kept to field border
			 long   maxFallbackTime,     // maximum number of fallback timesteps
			 WBall& meet );              // point where ball will be intercepted

  // if player knows that he gets a pass but has not seen the ball yet
  // he turns into a direction that might be good for later intercept
  void  getPass();

  bool  haveSeenBallSincePass;
  bool  waitForBall;  
  // [MW] Am I waiting for the ball? Look for dangerous opps

  long  lastGoalieInterceptTime;

  // ----------< findBall/turnHead >----------

  bool  lookingForBall;
  Angle turnAngle; // angle by which to turn for findBall
  Grade turnHeadGrade;

  Grade evaluateFindBall(); // checks if ball is too old

  void  findBall();         // turns to look for the ball

  // ----------< dribble, fastdribble and place >----------

  Vector2d fd_ballV;     // desired ball speed after fastdribble kick

  Angle    dribbleDir;   // for fastdribble and dribble
  Angle    kickDir;      // for kickaway
  double   dribbleStamina;

  long     placeCycles;  // no. of cycles we have been placing the ball without interrupt

  // stores parameters for evaluation of dribble and place in different areas of the field
  struct DribbleAreaData
  {
    int      numDirs;       // number of directions for this area
    struct
    {
      Vector2d dir;         // angle of this dribble direction
      Grade    grade;       // Grade for this direction in this area
      bool     dribbleonly; // true == fastdribble not allowed for this direction in this area
    } dirs[9];

    Grade    placeGrade;    // Grade for placing in this area
  };

  static const DribbleAreaData areas[];

  // test current ball position for the area code
  int  dribbleArea();
};

#endif
