//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "passeval.h"
#include <math.h>

using namespace std;


//////////////////////////////////////////////////////////////////////////
//
// class PassEval
//
//////////////////////////////////////////////////////////////////////////

PassEval::PassEval( const GoalShotEval * _goalshoteval )
{
  goalshoteval = _goalshoteval;
  SAMELEVEL_LENGTH    = Config::getValueAsDouble("pass_sameLevelLength");
  MAXKICK_DURATION    = Config::getValueAsInt("pass_maxKickDuration");
  MAXPASS_LENGTH      = Config::getValueAsDouble("pass_maxKickLength");
  RPASS_LATTICELENGTH = Config::getValueAsDouble("pass_rpassLatticeLength");
  SHORTPASS_LENGTH    = Config::getValueAsDouble("pass_shortPassLength");
  SHORTPASS_LENGTH_GOALIE = Config::getValueAsDouble("pass_shortPassLengthGoalie");
  SECURITY_DIST       = Config::getValueAsDouble("pass_securityDist");
  BORDER_DIST         = Config::getValueAsDouble("pass_borderDist");
  SCORE_POOR          = Config::getValueAsInt("pass_score_POOR");
  SCORE_AVERAGE       = Config::getValueAsInt("pass_score_AVERAGE");
  SCORE_GOOD          = Config::getValueAsInt("pass_score_GOOD");
  SCORE_PERFECT       = Config::getValueAsInt("pass_score_PERFECT");
}


////////////////////////////////////////////////////////////////////////
//
// PropOf PassEval::ourEvalProp
//
////////////////////////////////////////////////////////////////////////

PropOf PassEval::ourEvalProp( const Worldmodel *world, const TacticalModule * tact,
			      int plnumber, int panumber ) const
{
  ensure( plnumber >= 1 && plnumber <= 11 && panumber >= 1 && panumber <= 11 );

  // examine rangepasses for this grid:
  //
  //  + + + -    |       p = partner 
  //  * * + -    |       * = possible points for a rangepass
  //  p * + +    |       - = unwanted points
  //  * * + -   \ /      + = only in offensive situation
  //  + + + -    j       x- and y-distance beetween points: RPASS_LATTICELENGTH
  //
  //  ---------> i

  PropOf bestProp;

  bestProp.score = SCORE_POOR + 1;
  bestProp.agescore = 0;
  bestProp.grade.set_BAD();
  bestProp.number = panumber;
  bestProp.offside = false;
  bestProp.pos = world->ourTeam[panumber].pos;
  bestProp.dist = world->ourTeam[plnumber].distTo( bestProp.pos );
  bestProp.vel = 0.0;
  bestProp.xIndex = 0;
  bestProp.yIndex = 0;

  if( panumber == plnumber )
    {
      bestProp.score = SCORE_POOR + 21;
      return bestProp;
    }
  if( world->ourTeam[panumber].age > MAXAGE )
    {
      bestProp.score = SCORE_POOR + 22;
      return bestProp;
    }
  if( panumber == world->ourGoalieNo )
    {
      bestProp.score = SCORE_POOR + 23;
      return bestProp;
    }
  if( world->ourGoalieNo == UNKNOWN && world->ourTeam[panumber].pos.getX() < -FIELD_LENGTH / 3.0 )
    {
      bestProp.score = SCORE_POOR + 24;
      return bestProp;
    }
  if( world->ourTeam[plnumber].distTo( world->ourTeam[panumber] ) > MAXPASS_LENGTH )
    {
      bestProp.score = SCORE_POOR + 25;
      return bestProp;
    }

  double offsideLine = world->ourOffsideLine( MAXAGE );
  long maxTimeForPass = timeForPass( world, plnumber );

  PropOf helpProp;

  bool defensive = world->ourTeam[panumber].pos.getX() < 0.0 && !tact->ourTeamOffender(panumber);
  
  for( int i = 0; i <= ( defensive ? 1 : 10 ); i++ )   // check grid point [i][j]
    for( int j = ( defensive ? 1 : 0 ); j <= ( defensive ? 3 : 4 ); j++ )
      {
	// check all points but the ones marked with "+"
	if( i < 3 || j == 2 )
	  {
	    // calculate the position of this point
	    helpProp.pos = world->ourTeam[panumber].pos + RPASS_LATTICELENGTH * Vector2d( i, j - 2 );

	    evalPass( true, world, tact, plnumber, panumber, offsideLine, maxTimeForPass, BALL_SPEED_MAX, helpProp );

	    if( helpProp.score - helpProp.agescore < bestProp.score - bestProp.agescore )
	      {
		bestProp = helpProp;

		bestProp.xIndex = i;
		bestProp.yIndex = j - 2;
	      }
	  }
      }

  return bestProp;
}


////////////////////////////////////////////////////////////////////////
//
// Grade PassEval::otherEval
//
////////////////////////////////////////////////////////////////////////

Grade PassEval::otherEval( const Worldmodel* world, const TacticalModule * tact, 
			   int plnumber, int panumber ) const
{
  ensure( plnumber >= 1 && plnumber <= 11 & panumber >= 1 && panumber <= 11 );
  PropOf prop;
  prop.pos = world->otherTeam[panumber].pos;
  double offsideLine = world->otherOffsideLine( MAXAGE );
  evalPass( false, world, tact, plnumber, panumber, offsideLine, MAXKICK_DURATION, BALL_SPEED_MAX, prop );
  returnGRADE( prop.grade );
}


////////////////////////////////////////////////////////////////////////
//
// PassEval::timeForPass
//
////////////////////////////////////////////////////////////////////////

long PassEval::timeForPass( const Worldmodel *world, int player ) const
{
  // in these playmodes we have got enough time
  // just check if we must not place because ball is at field border
  if( ( world->playMode == PLAY_MODE_GOAL_KICK_US ) ||
      ( world->playMode == PLAY_MODE_FREE_KICK_US ) ||
      ( world->playMode == PLAY_MODE_KICK_IN_US ) ||
      ( world->playMode == PLAY_MODE_CORNER_KICK_US ) )
    return ( inField( world->ourTeam[player].pos, SECURITY_DIST ) ? MAXKICK_DURATION : 1 );

  // closest opponent to player
  int opp = world->otherTeamClosestTo( world->ourTeam[player].pos );

  if( opp == UNKNOWN )
    // player does not see any opponent so everything seems to be alright
    return ( inField( world->ourTeam[player].pos, SECURITY_DIST ) ? MAXKICK_DURATION : 1 );

  // distance between player and closest opponent
  double pod = max( 0.0, dist( world->ourTeam[player].pos, world->otherTeam[opp].pos )
		    - world->otherTeam[opp].kickRange() ); 

  // approximated time opponent will need to reach the ball
  return max( 1, min( MAXKICK_DURATION, long( pod / world->otherTeam[opp].vmax ) ) );
}


////////////////////////////////////////////////////////////////////////
//
// bool PassEval::evalPass
//
////////////////////////////////////////////////////////////////////////

#ifndef USEPASSETRACE
bool PassEval::evalPass( bool we, const Worldmodel *world, 
			 const TacticalModule * tact, 
			 int plnumber, int panumber,
			 double offsideLine, long maxTimeForPass,
			 double maxBallSpeed, PropOf &prop ) const
#else
bool PassEval::evalPass( bool we, const Worldmodel *world, 
			 const TacticalModule * tact, 
			 int plnumber, int panumber,
			 double offsideLine, long maxTimeForPass, 
			 double maxBallSpeed, PropOf &prop,
			 bool debug = false ) const
#endif
{
  const WPlayer &player  = ( we ? world->ourTeam[plnumber] : world->otherTeam[plnumber] );
  const WPlayer &partner = ( we ? world->ourTeam[panumber] : world->otherTeam[panumber] );
  
  prop.grade.set_BAD();
  prop.number    = panumber;
  prop.dist      = player.distTo( prop.pos );
  prop.xIndex    = 0;
  prop.yIndex    = 0;
  prop.score     = 0;
  prop.tactscore = 0;
  prop.corrscore = 0;
  prop.agescore  = 0;
  prop.numSteps  = 0;
  prop.duration  = 0;
  prop.vel       = 0.0;

  // target too far away
  if( prop.dist > MAXPASS_LENGTH )
    {
      prop.score = SCORE_POOR + 11;
      return false;
    }

  // keep away from the field border
  if( !inField( prop.pos, BORDER_DIST ) )
    {
      prop.score = SCORE_POOR + 12;
      return false;
    }

  prop.offside = ( we ? ( offsideLine <= partner.pos.getX() ) : ( offsideLine >= partner.pos.getX() ) );

  // test teammates: if one of them is offside better do not pass
  if( !prop.offside )
    for( int i = 1; i <= 11; i++ )
      if( i != plnumber && i != panumber )
	{
	  const WPlayer &dummy = ( we ? world->ourTeam[i] : world->otherTeam[i] );
	  if( ( we ? dummy.pos.getX() >= offsideLine : dummy.pos.getX() <= offsideLine ) &&
	      ( dummy.distToLine( player.pos, prop.pos ) < OFFSIDE_ACTIVE_AREA ) )
	  {
	      prop.offside = true;
	      break;
	  }
	}

  // partner offside or offside players in the way
  if( prop.offside )
    {
      prop.score = SCORE_POOR + 13;
      return false;
    }

  // stamina player can use; calculated for 10 steps, [MW] todo: check partner.stamina
  double dashStamina = min( MAXPOWER, partner.staminaIncMax +
			    max( 0.0, partner.stamina - EFFORT_DEC_THR * STAMINA_MAX ) / 10.0 );

  PASSECMD( if( debug ) PASSETRACE( "PASSEVAL-> partner.stamina : " << partner.stamina ); );

  // estimated number of dashes player will need to reach the target position 
  prop.numSteps = long( max( 0.0, partner.distTo( prop.pos ) - partner.kickRange() )
			* 1.2 * ( 1.0 - partner.decay ) / dashStamina / partner.dashPowerRate );
  // [MW] added experimental factor 1.2, to slow partner down

  // partner is immobile after tackling
  if( partner.cyclesSinceTackle <= TACKLE_CYCLES )
    prop.numSteps += TACKLE_CYCLES - partner.cyclesSinceTackle;

  // calculate maximal velocity the ball may have at the beginning so that
  // partner can reach it (calculated with additional three steps for turning etc.)
  prop.vel = min( maxBallSpeed, prop.dist * ( 1.0 - BALL_DECAY ) / 
		  ( 1.0 - power( BALL_DECAY, prop.numSteps + 4 ) ) );

  // if partner is too slow the ball might not reach the target
  if( prop.vel / ( 1.0 - BALL_DECAY ) < prop.dist )
    {
      prop.score = SCORE_POOR + 14;
      prop.numSteps = 0;
      prop.vel = 0.0;
      return false;
    }

  // [MW] slow passes are not good
  if( prop.vel < PLAYER_SPEED_MAX )
    {
      prop.score = SCORE_POOR + 15;
      prop.numSteps = 0;
      prop.vel = 0.0;
      return false;
    }

  // [MW] calculate how long it will take until the ball is at the
  // intercept position of our partner
  prop.duration = (int) ceil( log( 1 - prop.dist / prop.vel * ( 1 - BALL_DECAY ) ) / log( BALL_DECAY ) );

  WObject ourGoal   = ( we ? world->ourGoal : world->otherGoal );
  WObject otherGoal = ( we ? world->otherGoal : world->ourGoal );

  // better set ball pos to player.pos
  WBall simBall( player.pos, radAng2Vector2d( prop.vel, ( prop.pos - player.pos ).getAngle() ) );
  simBall.simulate();   // this is the ball after the kick command

  PASSECMD( ostringstream out; );

  Angle  oppAngle;
  double oppScore;
  const double maxV = 1.2;
  const double scoreV = 0.8;

  long n = 1;
  // simulate until we reach the ball (and a few steps more)
  // [MW] use prop.duration not prop.numSteps
  // also take into account that partner might intercept the pass earlier
  while( n <= min( 30, prop.duration + 1 ) &&
	 player.distTo( simBall ) < prop.dist - partner.kickRange() &&
	 scoreV * n * partner.vmax - partner.distTo( simBall ) + partner.kickRange() < 0 )
    {
      simBall.simulate();

      PASSECMD( if (debug) out << n << "( "; );

      for( int i = 1; i <= 11; i++ )
	{
	  const WPlayer &opp = ( we ? world->otherTeam[i] : world->ourTeam[i] );

	  // player is immobile after tackling
	  if( opp.age < MAXAGE && opp.cyclesSinceTackle + n > TACKLE_CYCLES )
	    {
	      // [MW] start counting the opponent if he could reach the ball
	      // with maxV times his maxspeed. Add full penalty if he can reach
	      // the ball with only scoreV times his maxspeed.

	      double playerAdvantage = maxV * n * opp.vmax - opp.distTo( simBall ) + opp.kickRange();

	      if( playerAdvantage > 0 )
		{
		  oppScore = 50;

		  PASSECMD( if( debug ) out << i << ":( " );

		  // [MW] take bodyDir into account if not too old
		  if( opp.bodyAge < 3 )
		    {
		      oppAngle = ( simBall.pos - opp.pos ).getAngle();
		      if( fabs( M_PI + oppAngle - opp.bodyDir ) > 3.0 * M_PI / 4.0 )
			{
			  PASSECMD( if( debug ) out << "bb "; );

			  oppScore -= ( opp.bodyAge == 0 ? 5 :
					opp.bodyAge == 1 ? 3 :
					opp.bodyAge == 2 ? 1 : 0 );
			}
		      if( norm( opp.v ) > opp.vmax / 4.0 &&
			  fabs( M_PI + oppAngle - opp.v.getAngle() ) > 3.0 * M_PI / 4.0 )
			{
			  PASSECMD( if( debug ) out << "bv "; );

			  oppScore -= ( opp.bodyAge == 0 ? 5 :
					opp.bodyAge == 1 ? 3 :
					opp.bodyAge == 2 ? 1 : 0 );
			}
		    }

		  prop.corrscore += (int) 
		    ceil( oppScore / ( maxV - scoreV ) * playerAdvantage / ( n * opp.vmax ) );

		  PASSECMD( if( debug )
			    out << oppScore << ": " 
			    << (int) ceil( oppScore / ( maxV - scoreV ) * playerAdvantage / ( n * opp.vmax ) )
			    << ") "; );
		}
	    }
	}
      PASSECMD( if( debug ) out << ") "; );
      n++;
    }

  PASSECMD( if( debug )
  {
    out << char(0);
    string  str = out.str();
    PASSETRACE( "PASSEVAL-> corridor : " << str );
  } );

  // ----------< determine tactscore >----------

  if( dist( prop.pos, otherGoal.pos ) < dist( prop.pos, ourGoal.pos ) )
    {
      // (linear) penalty for passes leading away from opponent goal
      double goaldist = norm( player.pos - otherGoal.pos ) - norm( prop.pos - otherGoal.pos );
      if( goaldist < 0 )
	{
	  prop.tactscore -= (int) ceil( 50 / SAMELEVEL_LENGTH * goaldist );
	}
    }
  else
    {
      // (linear) penalty for passes leading closer to our goal
      double goaldist = norm( prop.pos - ourGoal.pos ) - norm( player.pos - ourGoal.pos );
      if( goaldist < 0 )
	{
	  prop.tactscore -= (int) ceil( 50 / SAMELEVEL_LENGTH * goaldist );
	}
    }

  // upgrade passes in front of players
  if( we )
    if( world->ourTeam[panumber].pos.getX() < prop.pos.getX() )
      prop.tactscore -= 1 * (int) ceil( ( prop.pos.getX() - world->ourTeam[panumber].pos.getX() ) );

  // upgrade passes that cross the offside line
  /*
  if( ( we && offsideLine < prop.pos.getX() ) || ( !we && offsideLine > prop.pos.getX() ) )
    prop.tactscore -= 20 * (int) ceil( fabs( prop.pos.getX() - offsideLine ) );
  */

  if( ( we && offsideLine < simBall.pos.getX() ) || ( !we && offsideLine > simBall.pos.getX() ) )
    prop.tactscore -= 20 * (int) ceil( fabs( simBall.pos.getX() - offsideLine ) );

  // passes to defenders are not good
  if( we )
    {
      if( tact->ourTeamDefender( panumber ) )
	prop.tactscore += 100;
      if( tact->ourTeamDefMidfield( panumber ) )
	prop.tactscore += 20;
    }

  // check if partner is in a good goalshot position
  Grade goalshotGrade = ( we ? goalshoteval->ourEval( world, prop.pos ) :
			  goalshoteval->otherEval( world, prop.pos ) );
  if( goalshotGrade.is_better_than( Grade::AVERAGE ) )
    {
      prop.tactscore -= 50;
      if( goalshotGrade.is_better_than( Grade::GOOD ) )
	prop.tactscore -= 50;
    }

  // ----------< determine agescore >----------

  // consider partner's age
  prop.agescore += min( MAXAGE, partner.age ) * 25;

  // if it is my pass: consider age of pass direction
  if( we && plnumber == world->my.number )
    {
      long passDirAge = world->dirAge( ( prop.pos - player.pos ).getAngle() );
      prop.agescore += max( 0, min( MAXAGE, passDirAge ) * 25 );
    }

  // ----------< calculate final score >----------

  // partner close
  if( we )
    if( prop.dist < SHORTPASS_LENGTH )
      {
	if( player.isGoalie && prop.dist < SHORTPASS_LENGTH_GOALIE )
	  prop.score += 100;
	if( maxTimeForPass > 1 )
	  prop.score += 100;
      }

  // sum all different scores
  prop.score += prop.tactscore + prop.agescore + prop.corrscore;

  // calculate grade from score
  prop.grade = scoreToGrade( prop.score );

  // set velocity to target velocity
  // old: prop.vel *= power( BALL_DECAY, prop.numSteps + 4 );
  prop.vel *= power( BALL_DECAY, prop.duration );

  return true;
}
