//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "standardsituationmodule.h"

using namespace std;


// include the outsourced methods
#include "stdGoalie.C"
#include "stdKicksUs.C"
#include "stdPenalty.C"

// ------------------------------------------------------------------

StandardSituationModule::StandardSituationModule( Worldmodel * _world,
						  TacticalModule * _tact,
						  AdvSkills * _skills,
						  bool isGoalie, 
						  const PassEval *_passEval, 
						  const GoalShotEval *_goalShotEval, 
						  const PositionEval *_positionEval ) :
  world( _world ),
  tact( _tact ),
  skills( _skills ),
  kickUsPlayer( UNKNOWN ),
  goToPhase( 0 ),
  penaltyPlayer( 1 ),
  actPlayer( UNKNOWN ),
  goalShotEval( _goalShotEval ),
  passEval( _passEval ),
  positionEval( _positionEval )
{
  // read agent.conf
  activeStaminaFactor        = Config::getValueAsDouble( "std_activeStaminaFactor" );
  standardWaitingTime        = Config::getValueAsInt( "std_maxWaitingTime" );
  waitingTimeAfterGoalieMove = Config::getValueAsInt( "std_waitingTimeAfterGoalieMove" );
  penaltyKickGoaliePos       = Config::getValueAsVector( "std_PenaltyKickGoaliePos" );
  penaltyKickFieldplayerPos  = Config::getValueAsVector( "std_PenaltyKickFieldplayerPos" );
  maxObstacleDist            = Config::getValueAsDouble( "std_maxObstacleDist" );
  keepAwayDist               = Config::getValueAsDouble( "std_keepAwayDist" );
  goalieMovePositions        = Config::getValueAsInt( "std_goalieMovePositions" );
  goalieDistToFrontOfPenArea = Config::getValueAsDouble( "std_goalieDistToFrontOfPenArea" );
  goalieDistToSideOfPenArea  = Config::getValueAsDouble( "std_goalieDistToSideOfPenArea" );

  // read player types and count for stdsits
  players[0].defenders       = Config::getValueAsInt( "std_backField_defenders" );
  players[0].midfield        = Config::getValueAsInt( "std_backField_midfield" );
  players[0].offenders       = Config::getValueAsInt( "std_backField_offenders" );

  players[1].defenders       = Config::getValueAsInt( "std_middleField_defenders" );
  players[1].midfield        = Config::getValueAsInt( "std_middleField_midfield" );
  players[1].offenders       = Config::getValueAsInt( "std_middleField_offenders" );

  players[2].defenders       = Config::getValueAsInt( "std_frontField_defenders" );
  players[2].midfield        = Config::getValueAsInt( "std_frontField_midfield" );
  players[2].offenders       = Config::getValueAsInt( "std_frontField_offenders" );

  goalieMovesInCurrentSituation = 0;
  goalieDoesFreeKick = false;
  goalieMayMove = false;
  weAreKickingOff = false;
  actedThisCycle = false;
  wantToTurnHead = false;
  timeLeft = DROP_BALL_TIME - 10; //[C]
  stdSitFinished = false;

  iAmGoalie = isGoalie;
  maxAge = 20;
  waitingTime = -1;
}


bool 
StandardSituationModule::feedbackAudio( AudioMsg &msg )
{
  if( !msg.getState() || !msg.isCoachMsg() )
    {
      DECISION( "got bad msg" );
      return false;
    }

  Int3B player( 12 );

  msg >> player;

  if( player < 1 || player > 11 )
    {
      DECISION( "got bad player" << player );
      return false;
    }
  else
    kickUsPlayer = int( player );

  DECISION( "according to coach acting Player is " << kickUsPlayer );

  return true;
};


bool 
StandardSituationModule::feedbackWorld()
{
    if (world->playMode == PLAY_MODE_PENALTY_MISS_US && 
	playmode != PLAY_MODE_PENALTY_MISS_US        ||
	world->playMode == PLAY_MODE_PENALTY_SCORE_US && 
	playmode != PLAY_MODE_PENALTY_SCORE_US
	)
    {
	penaltyPlayer = (penaltyPlayer + 1) % 12;
	if (penaltyPlayer == 0) ++penaltyPlayer;
    }

  if( world->playMode == PLAY_MODE_PLAY_ON || playmode != world->playMode )
    {
      goToPhase = 0;
      stdSitFinished = false;
      actPlayer = UNKNOWN;
      kickUsPlayer = UNKNOWN;
      playmode = world->playMode;
    }

  // for drawing free kick circles
  switch( world->playMode )
    {
    case PLAY_MODE_OFFSIDE_THEM:
    case PLAY_MODE_OFFSIDE_US:
    case PLAY_MODE_FREE_KICK_THEM:
    case PLAY_MODE_FREE_KICK_US:
    case PLAY_MODE_CORNER_KICK_US:
    case PLAY_MODE_CORNER_KICK_THEM:
    case PLAY_MODE_KICK_IN_US:
    case PLAY_MODE_KICK_IN_THEM:
      MCMD( CIRCLE(world->ball.pos, FREE_KICK_FORBIDDEN_AREA) );
    default: ;
    }

  return true;
}


Grade 
StandardSituationModule::evaluate()
{
  actedThisCycle = false;
  Grade value;

  if( stdSitFinished )
    {
      value.set_IMPOSSIBLE();
      DECISION( "stdsit finished, returning IMPOSSIBLE" );
      returnGRADE( value );
    }

  // the evaluation
  switch( world->playMode )
    {
    case PLAY_MODE_UNKNOWN:
    case PLAY_MODE_TIME_OVER:
    case PLAY_MODE_OFFSIDE_US :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_GOALIE_CATCH_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
      value.set_IMPOSSIBLE();
      break;

    case PLAY_MODE_TIME_EXTENDED:
      timeLeft = DROP_BALL_TIME - 10; //[C]
      value.set_ABSOLUTE();
      weAreKickingOff = (world->ourSide == SIDE_LEFT);
      if( weAreKickingOff )
	{
	  DECISION( "Extended time, our kick off" );
	}
      else
	{
	  DECISION( "Extended time, other kick off" );
	}
      break;

    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_BACKPASS_THEM :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_FREEKICK_FAULT_THEM :
    case PLAY_MODE_GOALIE_CATCH_US :
      timeLeft--;
      value = evalFreeKickUs();
      break;

    case PLAY_MODE_GOAL_US :
      timeLeft--;
      weAreKickingOff = false;
      value.set_ABSOLUTE();
      DECISION( "We scored a goal" );
      break;

    case PLAY_MODE_GOAL_THEM :
      timeLeft--;
      weAreKickingOff = true;
      value.set_ABSOLUTE();
      DECISION( "The opponent team scored a goal" );
      break;      

    case PLAY_MODE_KICK_OFF_US :
      value.set_ABSOLUTE();
      weAreKickingOff = true;
      break;

    case PLAY_MODE_KICK_OFF_THEM :
      timeLeft = DROP_BALL_TIME - 10; //[C]
      weAreKickingOff = false;
      value.set_ABSOLUTE();
      break;

    case PLAY_MODE_BEFORE_KICK_OFF :
      timeLeft = DROP_BALL_TIME - 10; //[C]
      if( ( world->ourSide == SIDE_LEFT && world->time == 0 ) ||
	  ( world->ourSide == SIDE_RIGHT && world->time == HALF_TIME_IN_CYCLES ) ||
	  ( world->ourSide == SIDE_LEFT && world->time == 2*HALF_TIME_IN_CYCLES ) )
	{
	  // start of one of the half-times or the extended time
	  weAreKickingOff = true;
	}

      value.set_ABSOLUTE();
      if( weAreKickingOff )
	{
	  DECISION( "before our kick-off" );
	}
      else
	{
	  DECISION( "before opponent kick-off" );
	}
      break;

    case PLAY_MODE_GOAL_KICK_US :
      {
	timeLeft--;
	value.set_IMPOSSIBLE();

	Vector2d leftPos(-FIELD_LENGTH/2.+PENALTY_AREA_LENGTH-3., -PENALTY_AREA_WIDTH/2.0+3.),
	  rightPos(-FIELD_LENGTH/2.+PENALTY_AREA_LENGTH-3., +PENALTY_AREA_WIDTH/2.0-3.);
	switch( tact->ourSystem.numDefenders )
	  {
	  case 4:
	    targetPos[2] = leftPos;
	    targetPos[5] = rightPos;
	    if( world->my.tacticalNo == 2 || world->my.tacticalNo == 5 )
	      value.set_ABSOLUTE();
	    break;
	  case 3:
	    targetPos[2] = leftPos;
	    targetPos[4] = rightPos;
	    if( world->my.tacticalNo == 2 || world->my.tacticalNo == 4 )
	      value.set_ABSOLUTE();
	    break;
	  default:
	    targetPos[2] = leftPos;
	    targetPos[3] = rightPos;
	    if( world->my.tacticalNo == 2 || world->my.tacticalNo == 3 )
	      value.set_ABSOLUTE();
	    break;
	  }

	if( world->my.inOurPenaltyArea() )
	  value.set_ABSOLUTE();

	if( !value.is_ABSOLUTE() )
	  if( tact->ourAction.state == TA_PASS &&
	      tact->ourAction.receiver != world->my.number &&
	      tact->ourAction.performer != world->my.number )
	    {
	      stdSitFinished = true;
	      returnGRADE( value );
	    }

	if( iAmGoalie )
	  {
	    if( stdSitFinished )
	      {
		value.set_IMPOSSIBLE();
		break;
	      }
	    if( waitingTime < 0 )
	      waitingTime = standardWaitingTime;
	    value.set_ABSOLUTE();
	  }
	else
	  {
	    if( waitingTime < 0 )
	      waitingTime = 8; //[C]
	  }
	break;
      }

    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_CORNER_KICK_US :
      timeLeft--;
      value = evalRemainingKicksUs();
      break;

    case PLAY_MODE_HALF_TIME :
      timeLeft = DROP_BALL_TIME - 10; //[C]
      value.set_ABSOLUTE();
      weAreKickingOff = ( world->ourSide == SIDE_RIGHT );
      if( weAreKickingOff )
	{
	  DECISION( "Half-Time, our kick-off" );
	}
      else
	{
	  DECISION( "Half-Time, other kick-off" );
	}
      break;

    case PLAY_MODE_PLAY_ON:
      timeLeft = DROP_BALL_TIME - 10; //[C]
      waitingTime = -1;
      weAreKickingOff = false;
      goalieMovesInCurrentSituation = 0;
      goalieMayMove = false;
      value.set_IMPOSSIBLE();
      break;

    case PLAY_MODE_DROP_BALL :
      goalieMovesInCurrentSituation = 0;
      goalieMayMove = false;
      value.set_IMPOSSIBLE();
      break;

    case PLAY_MODE_PENALTY_KICK_US:
      if( world->time == 0 )
        value.set_ABSOLUTE();
      else 
	value.set_IMPOSSIBLE();
      break;

    case PLAY_MODE_PENALTY_KICK_THEM:
      if( world->time == 0 )
	value.set_ABSOLUTE();	
      else
	value.set_IMPOSSIBLE();
      break;
      
    case PLAY_MODE_PENALTY_SETUP_US:
    case PLAY_MODE_PENALTY_SETUP_THEM:
    case PLAY_MODE_PENALTY_READY_THEM:
    case PLAY_MODE_PENALTY_MISS_US:
    case PLAY_MODE_PENALTY_MISS_THEM:
    case PLAY_MODE_PENALTY_SCORE_US:
    case PLAY_MODE_PENALTY_SCORE_THEM:
      value.set_ABSOLUTE();
      break;
      
    case PLAY_MODE_PENALTY_READY_US:
    case PLAY_MODE_PENALTY_TAKEN_US:
      if (PENALTY_ALLOW_MULT_KICKS &&
	  penaltyPlayer == world->my.number)
	value.set_POOR();
      else
	value.set_ABSOLUTE();
      break;

      /* when the other team is going to shoot, make sure
	 our goalie is ready to catch
      */
      
    case PLAY_MODE_PENALTY_TAKEN_THEM:
      if (iAmGoalie)
	value.set_POOR();
      else
	value.set_ABSOLUTE();
      break;
      
    default:
      value.set_IMPOSSIBLE();
      DECISION( "Playmode weird ?!?" );
      break;
    }
  DECISION( "evaluate returns " << value \
	    << ", playmode = " << world->playMode );
  DECISION( "Time left: " << timeLeft );

  returnGRADE( value );
}

TurnHeadWish
StandardSituationModule::turnHeadEvaluate()
{
  if( actedThisCycle && wantToTurnHead )
    thwish.grade.set_ABSOLUTE();
  else
    thwish.grade.set_IMPOSSIBLE();

  return thwish;
}


bool 
StandardSituationModule::act()
{
  actedThisCycle = true;
  wantToTurnHead = false;

  switch( world->playMode )
    {
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_GOAL_THEM :
      return doingKickOff();

    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_FREEKICK_FAULT_THEM :
    case PLAY_MODE_BACKPASS_THEM :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_GOALIE_CATCH_US :
      if( goalieDoesFreeKick )
	return doingGoalieFreeKickUs();   // this method also considers field players!
      // no break here, fall through to doingStdKickUs()

    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_CORNER_KICK_US :
      {
	if( world->my.number == actPlayer )
	  return doingStdKickUs();
	else
	  {
	    DECISION( "Not responsible in " << world->playMode << " situation, doing nothing" );
	    skills->noAction();
	    return false;
	  }
	break;
      }

    case PLAY_MODE_GOAL_KICK_US :
      return ( iAmGoalie ? doingGoalieGoalKickUs() : doingFieldPlayerGoalKickUs() );

    case PLAY_MODE_PENALTY_KICK_US :
      if( world->time == 0 )
	{
	  if( world->my.distTo( penaltyKickFieldplayerPos ) > 0.5 )
	    {
	      DECISION( "Moving to " << penaltyKickFieldplayerPos );
	      skills->basicMove( penaltyKickFieldplayerPos.getX(), penaltyKickFieldplayerPos.getY() );
	      return true;
	    }
	  else
	    {
	      DECISION( "Have already moved, trying to face the goal" );
	      AdvSkills::TurnCommand tcmd = 
		skills->turnTo( world->otherGoal );
	      if( tcmd.state != AdvSkills::RFINISHED )
		skills->exec( tcmd );
	      else
		skills->noAction();
	      thwish.lookAtBodyDir( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
	      wantToTurnHead = true;
	      return true;
	    }
	}
      else
	{
	  DECISION( "I am not responsible for " << world->playMode );
	  skills->noAction();
	  return false;
	}

    case PLAY_MODE_PENALTY_KICK_THEM :
      if( world->time == 0 )
	{
	  if( world->my.distTo( penaltyKickGoaliePos ) > 0.5 )
	    {
	      DECISION( "Moving to " << penaltyKickGoaliePos );
	      skills->basicMove( penaltyKickGoaliePos.getX(), penaltyKickGoaliePos.getY() );
	      return true;
	    }
	  else
	    {
	      DECISION( "Have already moved, trying to face the opponent" );
	      AdvSkills::TurnCommand tcmd = 
		skills->turnTo( world->otherGoal );
	      if( tcmd.state != AdvSkills::RFINISHED )
		skills->exec( tcmd );
	      else
		skills->noAction();
	      thwish.lookAtBodyDir( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
	      wantToTurnHead = true;
	      return true;
	    }
	}
      else
	{
	  DECISION( "I am not responsible for " << world->playMode );
	  skills->noAction();
	  return false;
	}
      break;

    case PLAY_MODE_PENALTY_MISS_THEM:
    case PLAY_MODE_PENALTY_SCORE_THEM:
    case PLAY_MODE_PENALTY_SETUP_US:
	doPenaltyShootOutOurSetup();
	return true;
	break;

    case PLAY_MODE_PENALTY_MISS_US:
    case PLAY_MODE_PENALTY_SCORE_US:
    case PLAY_MODE_PENALTY_SETUP_THEM:
	doPenaltyShootOutTheirSetup();
	return true;
	break;

    case PLAY_MODE_PENALTY_READY_US:
    case PLAY_MODE_PENALTY_TAKEN_US:
	doPenaltyShootOut();
	return true;
	break;

    case PLAY_MODE_PENALTY_TAKEN_THEM:
    case PLAY_MODE_PENALTY_READY_THEM:

    case PLAY_MODE_OFFSIDE_US :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
    case PLAY_MODE_UNKNOWN :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_PLAY_ON :
    default :
      skills->noAction();
      DECISION( "act does not do anything!!!" );
      break;
    }

  return false;
}


bool 
StandardSituationModule::goingTo( Vector2d &target, const double stamina )
{
  AdvSkills::TurtleCommand cmd = skills->goToPosition( target, world->my.size, stamina );

  if( cmd.state != AdvSkills::TFINISHED )
    {
      // we are not at the target yet, so run and look around (avoiding obstacles)
      avoidObstacles( target );
      AdvSkills::TurtleCommand cmd2 = skills->goToPosition( target, world->my.size, stamina );
      // This should not happen but better be careful...
      if( cmd2.state != AdvSkills::TFINISHED )
	skills->exec( cmd2 );
      else
	skills->exec( cmd );
      lookAround();
      DECISION( " Going towards " << target );
      MCMD( POSITION(target) );
      MCMD( LINE(world->my.pos, target) );
      return true;
    }
  else
    {
      // we arrived at the target, so turn to ball
      AdvSkills::TurnCommand tcmd = skills->turnTo( world->ball );
      if( tcmd.state != AdvSkills::RFINISHED )
	{
	  DECISION( "Turning to ball" );
	  skills->exec( tcmd );
	  thwish.lookAtBall( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
	  wantToTurnHead = true;
	  return true;
	}
    }
  // target reached
  return false;
}


void
StandardSituationModule::avoidObstacles( Vector2d &target ) const
{
  WMovedObject closestObstacle;
  double closestDist = maxObstacleDist;
  bool obstacleFound = false;

  // avoid players
  for( int i = 1; i <= 11; i++ )
    {
      if( i != world->my.number && world->my.distTo( world->ourTeam[i] ) < closestDist )
	if( world->ourTeam[i].distToLine( world->my.pos, target ) < world->ourTeam[i].size + world->my.size )
	  {
	    obstacleFound = true;
	    closestDist = world->my.distTo( world->ourTeam[i] );
	    closestObstacle = world->ourTeam[i];
	  }
      if( world->my.distTo( world->otherTeam[i] ) < closestDist )
	if( world->otherTeam[i].distToLine( world->my.pos, target ) < world->otherTeam[i].size + world->my.size )
	  {
	    obstacleFound = true;
	    closestDist = world->my.distTo( world->ourTeam[i] );
	    closestObstacle = world->ourTeam[i];
	  }
    }

  // avoid ball
  if( world->my.distTo( world->ball ) < closestDist )
    if( world->ball.distToLine( world->my.pos, target ) < world->ball.size + world->my.size )
      {
	obstacleFound = true;
	closestDist = world->my.distTo( world->ball );
	closestObstacle = world->ball;
      }

  // no obstacles in the next steps
  if( !obstacleFound )
    return;

  // calculate new target
  Vector2d dir = target - world->my.pos;
  dir.stretch( ( world->my.size + closestObstacle.size + keepAwayDist ) / dir.getLength() );
  dir.perp();
  Vector2d rightPos = closestObstacle.pos + dir;
  Vector2d leftPos = closestObstacle.pos - dir;
  if( norm( rightPos - world->my.pos ) + norm( target - rightPos ) <
      norm( leftPos - world->my.pos ) + norm( target - leftPos ) )
    target = rightPos;
  else
    target = leftPos;
}


void 
StandardSituationModule::lookAround()
{
  wantToTurnHead = true;
  thwish.lookAround( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
}


int
StandardSituationModule::calcClosestDesiredPlayerTo( const Vector2d &target )
{
  int plNo = UNKNOWN;

  int off = 0;
  int mid = 0;
  int def = 0;

  double minDist = NOWHERE_DIST;

  calcDesiredPlayerTypes( world->ball.pos, def, mid, off );

  for( int i = 1; i <= 11; i++ )
    {
      if( i == world->ourGoalieNo )
	continue;
      if( tact->ourTeamDefender( i ) && def == 0 )
	continue;
      if( tact->ourTeamMidfield( i ) && mid == 0 )
	continue;
      if( tact->ourTeamOffender( i ) && off == 0 )
	continue;

      double plDist = norm( world->ourTeam[i].pos - target );
      if( plDist < minDist )
	{
	  minDist = plDist;
	  plNo = i;
	}
    }

  return plNo;
}


void
StandardSituationModule::
calcDesiredPlayerTypes( const Vector2d &target, int &noDef, int &noMid, int &noOff ) const
{
  // calc the indices to access the array with the number of player types
  // for the numbers see agent.conf
  int fieldPart = 0;

  double maxX = positionEval->GetMaxX(), minX = positionEval->GetMinX();
  double d = fabs( maxX - minX ) / 3.0;

  MCMD( VLINE(maxX) );
  MCMD( VLINE(minX) );

  // [C]*
  // if the target is in left third between minX and maxX
  if( target.getX() <= minX + d )
    fieldPart = 0;
  // if the target is in the middle third between minX and maxX
  else if( target.getX() <= minX + 2.0 * d )
    fieldPart = 1;
  // if the target is in the right third between minX and maxX
  else
    fieldPart = 2;

  noDef = players[fieldPart].defenders;
  noMid = players[fieldPart].midfield;
  noOff = players[fieldPart].offenders;

  DECISION( " Desired player types: defenders: " << noDef );
  DECISION( "                        midfield: " << noMid );
  DECISION( "                       offenders: " << noOff );
}
